import{V as r}from"./VegaLite-sOKbiepp.js";import{a as e}from"./VegaLite-sOKbiepp.js";import{R as t}from"./index-FKnJJang.js";import"./time-BB5za7E_.js";import"./timer-CK-RIWfW.js";import"./linear-spCykGIY.js";import"./init-DLRA0X12.js";import"./range-CtcPcB_L.js";import"./zoom-COrs4lFh.js";import"./ordinal-DDUp3AbE.js";import"./colors-bszWmPJw.js";import"./arc-lt4NHaVr.js";import"./step-CNn2V7Th.js";import"./index-D4qOZ_cs.js";function o(){return o=Object.assign||function(r){for(var e=1;e<arguments.length;e++){var t=arguments[e];for(var o in t)Object.prototype.hasOwnProperty.call(t,o)&&(r[o]=t[o])}return r},o.apply(this,arguments)}function i(e){let{mode:i,spec:n}=e;class s extends t.PureComponent{render(){return t.createElement(r,o({mode:i,spec:n},this.props))}}var p,a,m;return m=function(){return n},(a="getSpec")in(p=s)?Object.defineProperty(p,a,{value:m,enumerable:!0,configurable:!0,writable:!0}):p[a]=m,s}export{r as Vega,e as VegaLite,i as createClassFromSpec};
