import{b0 as e,R as t,b1 as a,b2 as s,D as r,G as o,aj as n,b3 as l,K as i,b4 as c,b5 as d,b6 as h,b7 as m,b8 as x,b9 as f,ba as p,a7 as u,bb as b,bc as y,bd as j,be as g,bf as v,as as k,bg as w,bh as N,bi as _,bj as S,bk as C,ah as z,bl as M,bm as I,bn as D,bo as O,bp as P,bq as A,br as T,bs as R,W as $,bt as B,bu as W,bv as H,bw as L,bx as q,by as E,bz as F,bA as G,bB as V,bC as U,bD as Y,O as J,bE as K,bF as Q,bG as X,a0 as Z,at as ee}from"./index-FKnJJang.js";import{u as te,T as ae,C as se,S as re,M as oe,G as ne,P as le}from"./tree-Dvjb07YW.js";import"./main-CEaoNRsh.js";import"./index-D4qOZ_cs.js";
/**
 * @license lucide-react v0.428.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const ie=e("Activity",[["path",{d:"M22 12h-2.48a2 2 0 0 0-1.93 1.46l-2.35 8.36a.25.25 0 0 1-.48 0L9.24 2.18a.25.25 0 0 0-.48 0l-2.35 8.36A2 2 0 0 1 4.49 12H2",key:"169zse"}]]),ce=e("BookOpen",[["path",{d:"M2 3h6a4 4 0 0 1 4 4v14a3 3 0 0 0-3-3H2z",key:"vv98re"}],["path",{d:"M22 3h-6a4 4 0 0 0-4 4v14a3 3 0 0 1 3-3h7z",key:"1cyq3y"}]]),de=e("BookText",[["path",{d:"M4 19.5v-15A2.5 2.5 0 0 1 6.5 2H19a1 1 0 0 1 1 1v18a1 1 0 0 1-1 1H6.5a1 1 0 0 1 0-5H20",key:"k3hazp"}],["path",{d:"M8 11h8",key:"vwpz6n"}],["path",{d:"M8 7h6",key:"1f0q6e"}]]),he=e("ChartNoAxesColumn",[["line",{x1:"18",x2:"18",y1:"20",y2:"10",key:"1xfpm4"}],["line",{x1:"12",x2:"12",y1:"20",y2:"4",key:"be30l9"}],["line",{x1:"6",x2:"6",y1:"20",y2:"14",key:"1r4le6"}]]),me=e("CirclePlay",[["circle",{cx:"12",cy:"12",r:"10",key:"1mglay"}],["polygon",{points:"10 8 16 12 10 16 10 8",key:"1cimsy"}]]),xe=e("GraduationCap",[["path",{d:"M21.42 10.922a1 1 0 0 0-.019-1.838L12.83 5.18a2 2 0 0 0-1.66 0L2.6 9.08a1 1 0 0 0 0 1.832l8.57 3.908a2 2 0 0 0 1.66 0z",key:"j76jl0"}],["path",{d:"M22 10v6",key:"1lu8f3"}],["path",{d:"M6 12.5V16a6 3 0 0 0 12 0v-3.5",key:"1r8lef"}]]),fe=e("Orbit",[["circle",{cx:"12",cy:"12",r:"3",key:"1v7zrd"}],["circle",{cx:"19",cy:"5",r:"2",key:"mhkx31"}],["circle",{cx:"5",cy:"19",r:"2",key:"v8kfzx"}],["path",{d:"M10.4 21.9a10 10 0 0 0 9.941-15.416",key:"eohfx2"}],["path",{d:"M13.5 2.1a10 10 0 0 0-9.841 15.416",key:"19pvbm"}]]),pe=e("PanelsTopLeft",[["rect",{width:"18",height:"18",x:"3",y:"3",rx:"2",key:"afitv7"}],["path",{d:"M3 9h18",key:"1pudct"}],["path",{d:"M9 21V9",key:"1oto5p"}]]),ue=t.createContext({runningNotebooks:new Map,setRunningNotebooks:a.NOOP}),be=t.createContext(""),ye=s("marimo:home:include-markdown",!1,void 0,{getOnInit:!0}),je=s("marimo:home:expanded-folders",{},void 0,{getOnInit:!0});
/**
 * @license lucide-react v0.428.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */function ge(e){return`${J()}-${encodeURIComponent(e)}`}const ve=()=>{const e=r(51),[t,a]=o.useState(0);let s;e[0]===Symbol.for("react.memo_cache_sentinel")?(s=[],e[0]=s):s=e[0];const h=n(Ae,s);let m,f,p;e[1]===Symbol.for("react.memo_cache_sentinel")?(m=()=>{a((e=>e+1))},f={delayMs:1e4,whenVisible:!0},e[1]=m,e[2]=f):(m=e[1],f=e[2]),te(m,f),e[3]!==t?(p=[t],e[3]=t,e[4]=p):p=e[4];const u=n(Re,p);let b,y,j,g,v,k,w,N,_,S,C,z,M,I,D,O,P,A,T,R;if(e[5]!==h||e[6]!==u){z=Symbol.for("react.early_return_sentinel");e:{const t=l(h,u);if(t.error)throw t.error;const a=t.data;if(!a){let t;e[19]===Symbol.for("react.memo_cache_sentinel")?(t=i.jsx(x,{centered:!0,size:"xlarge"}),e[19]=t):t=e[19],z=t;break e}const[s,r]=a;let n,c;g=s,C=o.Suspense,N=ue.Provider,_={runningNotebooks:r,setRunningNotebooks:u.setData},e[20]===Symbol.for("react.memo_cache_sentinel")?(n=i.jsx(Pe,{}),c=i.jsx(se,{showAppConfig:!1}),e[20]=n,e[21]=c):(n=e[20],c=e[21]);const d=`This will shutdown the notebook server and terminate all running notebooks (${r.size}). You'll lose all data that's in memory.`;e[22]!==d?(S=i.jsxs("div",{className:"absolute top-3 right-5 flex gap-2 z-50",children:[n,c,i.jsx(re,{description:d})]}),e[22]=d,e[23]=S):S=e[23],v="flex flex-col gap-6 max-w-5xl container pt-5 pb-20 z-10",e[24]===Symbol.for("react.memo_cache_sentinel")?(k=i.jsx("img",{src:"logo.png",alt:"marimo logo",className:"w-64 mb-4"}),w=i.jsx(De,{}),e[24]=k,e[25]=w):(k=e[24],w=e[25]),y=Ce,e[26]===Symbol.for("react.memo_cache_sentinel")?(j=i.jsx(ze,{Icon:me,children:"Running notebooks"}),e[26]=j):j=e[26],b=r.values()}e[5]=h,e[6]=u,e[7]=b,e[8]=y,e[9]=j,e[10]=g,e[11]=v,e[12]=k,e[13]=w,e[14]=N,e[15]=_,e[16]=S,e[17]=C,e[18]=z}else b=e[7],y=e[8],j=e[9],g=e[10],v=e[11],k=e[12],w=e[13],N=e[14],_=e[15],S=e[16],C=e[17],z=e[18];return z!==Symbol.for("react.early_return_sentinel")?z:(e[27]!==b?(M=[...b],e[27]=b,e[28]=M):M=e[28],e[29]!==y||e[30]!==j||e[31]!==M?(I=i.jsx(y,{header:j,files:M}),e[29]=y,e[30]=j,e[31]=M,e[32]=I):I=e[32],e[33]===Symbol.for("react.memo_cache_sentinel")?(D=i.jsx(ze,{Icon:c,children:"Recent notebooks"}),e[33]=D):D=e[33],e[34]!==g.files?(O=i.jsx(Ce,{header:D,files:g.files}),e[34]=g.files,e[35]=O):O=e[35],e[36]===Symbol.for("react.memo_cache_sentinel")?(P=i.jsx(d,{children:i.jsx(ke,{})}),e[36]=P):P=e[36],e[37]!==v||e[38]!==k||e[39]!==w||e[40]!==I||e[41]!==O?(A=i.jsxs("div",{className:v,children:[k,w,I,O,P]}),e[37]=v,e[38]=k,e[39]=w,e[40]=I,e[41]=O,e[42]=A):A=e[42],e[43]!==N||e[44]!==_||e[45]!==S||e[46]!==A?(T=i.jsxs(N,{value:_,children:[S,A]}),e[43]=N,e[44]=_,e[45]=S,e[46]=A,e[47]=T):T=e[47],e[48]!==C||e[49]!==T?(R=i.jsx(C,{children:T}),e[48]=C,e[49]=T,e[50]=R):R=e[50],R)},ke=()=>{const e=r(39),[t,a]=h(ye),[s,l]=o.useState("");let c,d;e[0]!==t?(c=()=>M({includeMarkdown:t}),d=[t],e[0]=t,e[1]=c,e[2]=d):(c=e[1],d=e[2]);const f=n(c,d);if(f.error){let t,a;return e[3]!==f.error?(t=m(f.error),e[3]=f.error,e[4]=t):t=e[4],e[5]!==t?(a=i.jsx(I,{kind:"danger",className:"rounded p-4",children:t}),e[5]=t,e[6]=a):a=e[6],a}if(f.loading||!f.data){let t;return e[7]===Symbol.for("react.memo_cache_sentinel")?(t=i.jsx(x,{centered:!0,size:"xlarge",className:"mt-6"}),e[7]=t):t=e[7],t}const p=f.data;let u,b,y,j,g,v,k,w,N,_,S,C,z,R;return e[8]===Symbol.for("react.memo_cache_sentinel")?(u=i.jsx(D,{size:13}),b=e=>l(e.target.value),e[8]=u,e[9]=b):(u=e[8],b=e[9]),e[10]!==s?(y=i.jsx(O,{id:"search",value:s,icon:u,onChange:b,placeholder:"Search",className:"mb-0 border-border"}),e[10]=s,e[11]=y):y=e[11],e[12]===Symbol.for("react.memo_cache_sentinel")?(j=i.jsx(we,{}),e[12]=j):j=e[12],e[13]!==a?(g=e=>a(Boolean(e)),e[13]=a,e[14]=g):g=e[14],e[15]!==t||e[16]!==g?(v=i.jsx(P,{"data-testid":"include-markdown-checkbox",id:"include-markdown",checked:t,onCheckedChange:g}),e[15]=t,e[16]=g,e[17]=v):v=e[17],e[18]===Symbol.for("react.memo_cache_sentinel")?(k=i.jsx(A,{htmlFor:"include-markdown",children:"Include markdown"}),e[18]=k):k=e[18],e[19]!==y||e[20]!==v?(w=i.jsxs("div",{className:"flex items-center gap-2",children:[y,j,v,k]}),e[19]=y,e[20]=v,e[21]=w):w=e[21],e[22]!==f?(N=i.jsx(T,{className:"w-4 h-4 ml-1 cursor-pointer opacity-70 hover:opacity-100",onClick:()=>f.reload()}),e[22]=f,e[23]=N):N=e[23],e[24]!==f.loading?(_=f.loading&&i.jsx(x,{size:"small"}),e[24]=f.loading,e[25]=_):_=e[25],e[26]!==w||e[27]!==N||e[28]!==_?(S=i.jsxs(ze,{Icon:de,control:w,children:["Workspace",N,_]}),e[26]=w,e[27]=N,e[28]=_,e[29]=S):S=e[29],e[30]!==s||e[31]!==p.files?(C=i.jsx("div",{className:"flex flex-col divide-y divide-[var(--slate-3)] border rounded overflow-hidden max-h-[48rem] overflow-y-auto shadow-sm bg-background",children:i.jsx(Ne,{searchText:s,files:p.files})}),e[30]=s,e[31]=p.files,e[32]=C):C=e[32],e[33]!==S||e[34]!==C?(z=i.jsxs("div",{className:"flex flex-col gap-2",children:[S,C]}),e[33]=S,e[34]=C,e[35]=z):z=e[35],e[36]!==p.root||e[37]!==z?(R=i.jsx(be.Provider,{value:p.root,children:z}),e[36]=p.root,e[37]=z,e[38]=R):R=e[38],R},we=()=>{const e=r(5),t=f(je);let a,s,o;return e[0]!==t?(a=()=>{t({})},e[0]=t,e[1]=a):a=e[1],e[2]===Symbol.for("react.memo_cache_sentinel")?(s=i.jsx(R,{className:"w-4 h-4 mr-1"}),e[2]=s):s=e[2],e[3]!==a?(o=i.jsxs($,{variant:"text",size:"sm",className:"h-fit",onClick:a,children:[s,"Collapse all"]}),e[3]=a,e[4]=o):o=e[4],o},Ne=e=>{const t=r(12),{files:a,searchText:s}=e,[n,l]=h(je),c=0===Object.keys(n).length,d=o.useRef();let m,x,f,p;if(t[0]!==c?(m=()=>{var e;c&&(null==(e=d.current)||e.closeAll())},x=[c],t[0]=c,t[1]=m,t[2]=x):(m=t[1],x=t[2]),o.useEffect(m,x),0===a.length){let e;return t[3]===Symbol.for("react.memo_cache_sentinel")?(e=i.jsx("div",{className:"flex flex-col px-5 py-10 items-center justify-center",children:i.jsx("p",{className:"text-center text-muted-foreground",children:"No files in this workspace"})}),t[3]=e):e=t[3],e}return t[4]!==n||t[5]!==l?(f=async e=>{const t=n[e]??!1;l({...n,[e]:!t})},t[4]=n,t[5]=l,t[6]=f):f=t[6],t[7]!==s||t[8]!==a||t[9]!==n||t[10]!==f?(p=i.jsx(ae,{ref:d,width:"100%",height:500,searchTerm:s,className:"h-full",idAccessor:$e,data:a,openByDefault:!1,initialOpenState:n,onToggle:f,padding:5,rowHeight:35,indent:15,overscanCount:1e3,renderCursor:Be,disableDrop:!0,disableDrag:!0,disableEdit:!0,disableMultiSelection:!0,children:_e}),t[7]=s,t[8]=a,t[9]=n,t[10]=f,t[11]=p):p=t[11],p},_e=e=>{const t=r(19),{node:a,style:s}=e,n=a.data.isDirectory?"directory":p(a.data.name),l=B[n];let c;t[0]!==l?(c=i.jsx(l,{className:"w-5 h-5 flex-shrink-0",strokeWidth:1.5}),t[0]=l,t[1]=c):c=t[1];const d=c,h=o.useContext(be);let m;t[2]!==a.data.isDirectory||t[3]!==a.data.name||t[4]!==a.data.path||t[5]!==d||t[6]!==h?(m=()=>{if(a.data.isDirectory)return i.jsxs("span",{className:"flex items-center pl-1 cursor-pointer hover:bg-accent/50 hover:text-accent-foreground rounded-l flex-1 overflow-hidden h-full pr-3 gap-2",children:[d,a.data.name]});const e=a.data.path.startsWith(h)&&W.isAbsolute(a.data.path)?W.rest(a.data.path,h):a.data.path,t=e.endsWith(".md");return i.jsxs("a",{className:"flex items-center pl-1 cursor-pointer hover:bg-accent/50 hover:text-accent-foreground rounded-l flex-1 overflow-hidden h-full pr-3 gap-2",href:z(`?file=${e}`).toString(),target:ge(e),children:[d,i.jsxs("span",{className:"flex-1 overflow-hidden text-ellipsis",children:[a.data.name,t&&i.jsx(oe,{className:"ml-2 inline opacity-80"})]}),i.jsx(Ie,{filePath:e}),i.jsx(H,{size:20,className:"group-hover:opacity-100 opacity-0 text-primary"})]})},t[2]=a.data.isDirectory,t[3]=a.data.name,t[4]=a.data.path,t[5]=d,t[6]=h,t[7]=m):m=t[7];const x=m;let f,b,y,j,g;return t[8]===Symbol.for("react.memo_cache_sentinel")?(f=u("flex items-center cursor-pointer ml-1 text-muted-foreground whitespace-nowrap group h-full"),t[8]=f):f=t[8],t[9]!==a?(b=e=>{e.stopPropagation(),a.data.isDirectory&&a.toggle()},y=i.jsx(Se,{node:a}),t[9]=a,t[10]=b,t[11]=y):(b=t[10],y=t[11]),t[12]!==x?(j=x(),t[12]=x,t[13]=j):j=t[13],t[14]!==s||t[15]!==b||t[16]!==y||t[17]!==j?(g=i.jsxs("div",{style:s,className:f,onClick:b,children:[y,j]}),t[14]=s,t[15]=b,t[16]=y,t[17]=j,t[18]=g):g=t[18],g},Se=e=>{const t=r(3),{node:a}=e;if(!a.data.isDirectory){let e;return t[0]===Symbol.for("react.memo_cache_sentinel")?(e=i.jsx("span",{className:"w-5 h-5 flex-shrink-0"}),t[0]=e):e=t[0],e}let s;return t[1]!==a.isOpen?(s=a.isOpen?i.jsx(b,{className:"w-5 h-5 flex-shrink-0"}):i.jsx(y,{className:"w-5 h-5 flex-shrink-0"}),t[1]=a.isOpen,t[2]=s):s=t[2],s},Ce=e=>{const t=r(7),{header:a,files:s}=e;if(0===s.length)return null;let o,n,l;return t[0]!==s?(o=s.map(We),t[0]=s,t[1]=o):o=t[1],t[2]!==o?(n=i.jsx("div",{className:"flex flex-col divide-y divide-[var(--slate-3)] border rounded overflow-hidden max-h-[48rem] overflow-y-auto shadow-sm bg-background",children:o}),t[2]=o,t[3]=n):n=t[3],t[4]!==a||t[5]!==n?(l=i.jsxs("div",{className:"flex flex-col gap-2",children:[a,n]}),t[4]=a,t[5]=n,t[6]=l):l=t[6],l},ze=e=>{const t=r(8),{Icon:a,control:s,children:o}=e;let n,l,c;return t[0]!==a?(n=i.jsx(a,{className:"h-5 w-5"}),t[0]=a,t[1]=n):n=t[1],t[2]!==n||t[3]!==o?(l=i.jsxs("h2",{className:"flex items-center gap-2 text-xl font-semibold text-muted-foreground select-none",children:[n,o]}),t[2]=n,t[3]=o,t[4]=l):l=t[4],t[5]!==l||t[6]!==s?(c=i.jsxs("div",{className:"flex items-center justify-between gap-2",children:[l,s]}),t[5]=l,t[6]=s,t[7]=c):c=t[7],c},Me=e=>{const t=r(39),{file:a}=e;let s;t[0]!==a.path?(s=L(a.path),t[0]=a.path,t[1]=s):s=t[1];const o=s;let n,l,c,d;if(t[2]!==o||t[3]!==a){const e=z(o?`?file=${a.initializationId}&session_id=${a.path}`:`?file=${a.path}`);let s;t[8]!==a.path?(s=a.path.endsWith(".md"),t[8]=a.path,t[9]=s):s=t[9],n=s,l="py-1.5 px-4 hover:bg-[var(--blue-2)] hover:text-primary transition-all duration-300 cursor-pointer group relative flex gap-4 items-center",c=a.path,d=e.toString(),t[2]=o,t[3]=a,t[4]=n,t[5]=l,t[6]=c,t[7]=d}else n=t[4],l=t[5],c=t[6],d=t[7];const h=a.initializationId||a.path;let m,x,f,p,u,b,y,j,g,v,k;return t[10]!==h?(m=ge(h),t[10]=h,t[11]=m):m=t[11],t[12]!==n?(x=n&&i.jsx("span",{className:"opacity-80",children:i.jsx(oe,{})}),t[12]=n,t[13]=x):x=t[13],t[14]!==a.name||t[15]!==x?(f=i.jsxs("span",{className:"flex items-center gap-2",children:[a.name,x]}),t[14]=a.name,t[15]=x,t[16]=f):f=t[16],t[17]!==a.path?(p=i.jsx("p",{title:a.path,className:"text-sm text-muted-foreground overflow-hidden whitespace-nowrap overflow-ellipsis",children:a.path}),t[17]=a.path,t[18]=p):p=t[18],t[19]!==f||t[20]!==p?(u=i.jsxs("div",{className:"flex flex-col justify-between flex-1",children:[f,p]}),t[19]=f,t[20]=p,t[21]=u):u=t[21],t[22]!==a.path?(b=i.jsx("div",{children:i.jsx(Ie,{filePath:a.path})}),t[22]=a.path,t[23]=b):b=t[23],t[24]===Symbol.for("react.memo_cache_sentinel")?(y=i.jsx(H,{size:20,className:"group-hover:opacity-100 opacity-0 transition-all duration-300 text-primary"}),t[24]=y):y=t[24],t[25]!==b?(j=i.jsxs("div",{className:"flex gap-3 items-center",children:[b,y]}),t[25]=b,t[26]=j):j=t[26],t[27]!==a.lastModified?(g=!!a.lastModified&&i.jsx("div",{className:"text-xs text-muted-foreground opacity-80",children:q(1e3*a.lastModified)}),t[27]=a.lastModified,t[28]=g):g=t[28],t[29]!==j||t[30]!==g?(v=i.jsxs("div",{className:"flex flex-col gap-1 items-end",children:[j,g]}),t[29]=j,t[30]=g,t[31]=v):v=t[31],t[32]!==l||t[33]!==c||t[34]!==d||t[35]!==m||t[36]!==u||t[37]!==v?(k=i.jsxs("a",{className:l,href:d,target:m,children:[u,v]},c),t[32]=l,t[33]=c,t[34]=d,t[35]=m,t[36]=u,t[37]=v,t[38]=k):k=t[38],k},Ie=e=>{const t=r(9),{filePath:a}=e,{openConfirm:s,closeModal:n}=E(),{runningNotebooks:l,setRunningNotebooks:c}=o.useContext(ue);if(!l.has(a))return null;let d,h,m;return t[0]!==s||t[1]!==l||t[2]!==a||t[3]!==c||t[4]!==n?(d=e=>{e.stopPropagation(),e.preventDefault(),s({title:"Shutdown",description:"This will terminate the Python kernel. You'll lose all data that's in memory.",variant:"destructive",confirmAction:i.jsx(K,{onClick:e=>{const t=l.get(a);Q(t),X({sessionId:t.sessionId}).then((e=>{c(_.keyBy(e.files,(e=>e.path)))})),n(),Z({description:"Notebook has been shutdown."})},"aria-label":"Confirm Shutdown",children:"Shutdown"})})},t[0]=s,t[1]=l,t[2]=a,t[3]=c,t[4]=n,t[5]=d):d=t[5],t[6]===Symbol.for("react.memo_cache_sentinel")?(h=i.jsx(le,{size:14}),t[6]=h):h=t[6],t[7]!==d?(m=i.jsx(ee,{content:"Shutdown",children:i.jsx($,{size:"icon",variant:"outline",className:"opacity-80 hover:opacity-100 hover:bg-accent text-destructive border-destructive hover:border-destructive hover:text-destructive bg-background hover:bg-[var(--red-1)]",onClick:d,children:h})}),t[7]=d,t[8]=m):m=t[8],m},De=()=>{const e=r(3);let t;e[0]===Symbol.for("react.memo_cache_sentinel")?(t=j(),e[0]=t):t=e[0];const a=t;let s,o;return e[1]===Symbol.for("react.memo_cache_sentinel")?(s=i.jsx("h2",{className:"text-lg font-semibold",children:"Create a new notebook"}),e[1]=s):s=e[1],e[2]===Symbol.for("react.memo_cache_sentinel")?(o=i.jsxs("a",{className:"relative rounded-lg p-6 group\n      text-primary hover:bg-[var(--blue-2)] shadow-mdSolid shadow-accent border\n      transition-all duration-300 cursor-pointer\n      ",href:a,target:"_blank",rel:"noreferrer",children:[s,i.jsx("div",{className:"group-hover:opacity-100 opacity-0 absolute right-5 top-0 bottom-0 rounded-lg flex items-center justify-center transition-all duration-300",children:i.jsx(H,{size:24})})]}),e[2]=o):o=e[2],o},Oe={intro:["Introduction",ce],dataflow:["Dataflow",ie],ui:["UI Elements",pe],markdown:["Markdown",F],plots:["Plots",he],sql:["SQL",G],layout:["Layout",ne],fileformat:["File format",V],"for-jupyter-users":["For Jupyter users",fe],"markdown-format":["Markdown format",oe]},Pe=()=>{const e=r(3);let t,a,s;return e[0]===Symbol.for("react.memo_cache_sentinel")?(t=i.jsx(xe,{className:"w-4 h-4 mr-2"}),e[0]=t):t=e[0],e[1]===Symbol.for("react.memo_cache_sentinel")?(a=i.jsx(U,{asChild:!0,children:i.jsxs($,{"data-testid":"open-tutorial-button",size:"xs",variant:"outline",children:[t,"Tutorials",i.jsx(Y,{className:"w-3 h-3 ml-1"})]})}),e[1]=a):a=e[1],e[2]===Symbol.for("react.memo_cache_sentinel")?(s=i.jsxs(g,{children:[a,i.jsx(v,{side:"bottom",className:"no-print",children:k.entries(Oe).map(He)})]}),e[2]=s):s=e[2],s};function Ae(){return w()}function Te(e){return e.path}async function Re(){const e=await N();return _.keyBy(e.files,Te)}function $e(e){return e.path}function Be(){return null}function We(e){return i.jsx(Me,{file:e},e.path)}function He(e){const[t,a]=e,[s,r]=a;return i.jsxs(S,{onSelect:async()=>{const e=await C({tutorialId:t});e&&window.open(z(`?file=${e.path}`).toString(),"_blank")},children:[i.jsx(r,{strokeWidth:1.5,className:"w-4 h-4 mr-2"}),s]},t)}export{ve as default};
