import{b as a}from"./transform-0ILWlX8T.js";import{af as r,aU as n,n as t,B as o,aV as s,ab as e,aa as i,a9 as f,ad as u}from"./index-FKnJJang.js";import{an as v}from"./mermaid-C2iOgJeu.js";import{a as l}from"./_baseEach-DIpKRMNf.js";import{d as m}from"./_baseUniq-DmSpOwc-.js";import{b as c}from"./_baseMap-C7H3vESs.js";var d=Object.prototype,p=d.hasOwnProperty,b=a((function(a,t){a=Object(a);var o=-1,s=t.length,e=s>2?t[2]:void 0;for(e&&r(t[0],t[1],e)&&(s=1);++o<s;)for(var i=t[o],f=v(i),u=-1,l=f.length;++u<l;){var m=f[u],c=a[m];(void 0===c||n(c,d[m])&&!p.call(a,m))&&(a[m]=i[m])}return a}));function h(a){var r=null==a?0:a.length;return r?a[r-1]:void 0}var j=Math.max;var g,O=(g=function(a,r,n){var t=null==a?0:a.length;if(!t)return-1;var o=null==n?0:s(n);return o<0&&(o=j(t+o,0)),m(a,l(r),o)},function(a,r,n){var s=Object(a);if(!t(a)){var e=l(r);a=o(a),r=function(a){return e(s[a],a,s)}}var i=g(a,r,n);return i>-1?s[e?a[i]:i]:void 0});function x(a,r){return(e(a)?i:c)(a,l(r))}function _(a,r){return a<r}function y(a,r,n){for(var t=-1,o=a.length;++t<o;){var s=a[t],e=r(s);if(null!=e&&(void 0===i?e==e&&!f(e):n(e,i)))var i=e,u=s}return u}function M(a){return a&&a.length?y(a,u,_):void 0}export{_ as a,y as b,M as c,b as d,O as f,h as l,x as m};
