.mo-button.gray{background-color:var(--slate-1);border-color:var(--slate-6);color:var(--slate-11)}.mo-button.gray:hover{background-color:var(--slate-3);border-color:var(--slate-8)}.mo-button.gray.selected .mo-button.gray:active{background-color:var(--slate-4);border-color:var(--slate-7)}.mo-button.white{--tw-bg-opacity:1;background-color:color-mix(in srgb,var(--background),transparent calc((1 - var(--tw-bg-opacity))*100%));border-color:var(--slate-7);color:var(--slate-11)}.mo-button.white:hover{background-color:var(--slate-1);border-color:var(--slate-8)}.mo-button.white.selected .mo-button.white:active{background-color:var(--slate-2);border-color:var(--slate-7)}.mo-button.disabled{--tw-bg-opacity:1;background-color:color-mix(in srgb,var(--background),transparent calc((1 - var(--tw-bg-opacity))*100%));border-color:var(--slate-6);color:var(--slate-7)}.mo-button.green{background-color:var(--grass-3);border-color:var(--grass-6);color:var(--grass-11)}.mo-button.green:hover{background-color:var(--grass-2);border-color:var(--grass-7)}.mo-button.green.selected .mo-button.green:active{background-color:var(--grass-4);border-color:var(--grass-7)}.mo-button.hint-green{--tw-bg-opacity:1;background-color:color-mix(in srgb,var(--background),transparent calc((1 - var(--tw-bg-opacity))*100%));border-color:var(--slate-7);color:var(--slate-11)}.mo-button.hint-green:hover{background-color:var(--grass-2);border-color:var(--grass-7);color:var(--grass-11)}.mo-button.hint-green.selected .mo-button.hint-green:active{background-color:var(--grass-4);border-color:var(--grass-7);color:var(--slate-11)}.mo-button.red{background-color:var(--red-3);border-color:var(--red-6);color:var(--red-11)}.mo-button.red:hover{background-color:var(--red-4);border-color:var(--red-8)}.mo-button.red.selected .mo-button.red:active{background-color:var(--red-5);border-color:var(--red-7)}.mo-button.yellow{background-color:var(--yellow-3);border-color:var(--yellow-6);color:var(--yellow-11)}.mo-button.yellow:hover{background-color:var(--yellow-4);border-color:var(--yellow-8)}.mo-button.yellow.selected .mo-button.yellow:active{background-color:var(--yellow-5);border-color:var(--yellow-7)}
