import{_ as r,l as s,G as t,j as a,H as e}from"./mermaid-C2iOgJeu.js";import{p as i}from"./gitGraph-YCYPL57B-uDbxXOCi.js";import"./index-FKnJJang.js";import"./transform-0ILWlX8T.js";import"./timer-CK-RIWfW.js";import"./step-CNn2V7Th.js";import"./main-CEaoNRsh.js";import"./_baseEach-DIpKRMNf.js";import"./_baseUniq-DmSpOwc-.js";import"./min-BcCd9LO5.js";import"./_baseMap-C7H3vESs.js";import"./clone-CV2-s9tE.js";var o={version:e},n={parser:{parse:r((async r=>{const t=await i("info",r);s.debug(t)}),"parse")},db:{getVersion:r((()=>o.version),"getVersion")},renderer:{draw:r(((r,e,i)=>{s.debug("rendering info diagram\n"+r);const o=t(e);a(o,100,400,!0);o.append("g").append("text").attr("x",100).attr("y",40).attr("class","version").attr("font-size",32).style("text-anchor","middle").text(`v${i}`)}),"draw")}};export{n as diagram};
