import{bZ as n,b_ as r,b$ as t,B as u,c0 as e,c1 as a,c2 as i,c3 as o,ac as f,ad as c,ab as s,n as v}from"./index-FKnJJang.js";import{a as l}from"./transform-0ILWlX8T.js";function b(n){return n==n&&!t(n)}function d(n,r){return function(t){return null!=t&&(t[n]===r&&(void 0!==r||n in Object(t)))}}function j(t){var e=function(n){for(var r=u(n),t=r.length;t--;){var e=r[t],a=n[e];r[t]=[e,a,b(a)]}return r}(t);return 1==e.length&&e[0][2]?d(e[0][0],e[0][1]):function(u){return u===t||function(t,u,e,a){var i=e.length,o=i;if(null==t)return!o;for(t=Object(t);i--;){var f=e[i];if(f[2]?f[1]!==t[f[0]]:!(f[0]in t))return!1}for(;++i<o;){var c=(f=e[i])[0],s=t[c],v=f[1];if(f[2]){if(void 0===s&&!(c in t))return!1}else{var l=new n;if(!r(v,s,3,a,l))return!1}}return!0}(u,0,e)}}function m(n){return e(n)?(r=a(n),function(n){return null==n?void 0:n[r]}):function(n){return function(r){return f(r,n)}}(n);var r}function p(n){return"function"==typeof n?n:null==n?c:"object"==typeof n?s(n)?(t=n[0],u=n[1],e(t)&&b(u)?d(a(t),u):function(n){var e=i(n,t);return void 0===e&&e===u?o(n,t):r(u,e,3)}):j(n):m(n);var t,u}function g(n,r){return n&&l(n,r,u)}var h,O=(h=g,function(n,r){if(null==n)return n;if(!v(n))return h(n,r);for(var t=n.length,u=-1,e=Object(n);++u<t&&!1!==r(e[u],u,e););return n});export{p as a,O as b,g as c};
