import{b0 as e,b$ as t,am as n,dE as r,R as o,D as i,es as s,K as a,by as c,bE as l,cZ as u,at as d,et as f,G as h,cd as p,a7 as g,b2 as m,dX as v,dr as y,eu as b,ev as w,ew as x,av as T,cQ as D,ex as S,ey as O,ez as R,eA as j,eB as C,eC as N,eD as k,ar as P,eE as E,b6 as I,bq as _,cV as M,W as q,dO as A,a1 as L,$ as F,a0 as H,eF as W,eG as U,eH as B,eI as z,eJ as $,eK as V,eL as K,eM as G,d6 as J,e7 as Y,eN as X,eO as Z,eP as Q,eQ as ee,H as te,b9 as ne,al as re,eR as oe,b1 as ie,eS as se,as as ae,dj as ce,eT as le,bo as ue,eU as de,eV as fe,eW as he,U as pe,eX as ge,eY as me,X as ve,Y as ye,Z as be,eZ as we,e_ as xe,e$ as Te,dl as De,f0 as Se,f1 as Oe,f2 as Re,f3 as je,f4 as Ce,f5 as Ne,f6 as ke,f7 as Pe,bp as Ee,f8 as Ie,f9 as _e,fa as Me,fb as qe,fc as Ae,fd as Le,cS as Fe,fe as He,ff as We,d$ as Ue,fg as Be,fh as ze,fi as $e,fj as Ve,V as Ke,fk as Ge,fl as Je,fm as Ye,fn as Xe,fo as Ze}from"./index-FKnJJang.js";import{r as Qe,a as et,b as tt,c as nt,m as rt}from"./main-CEaoNRsh.js";import{e as ot}from"./index-D4qOZ_cs.js";
/**
 * @license lucide-react v0.428.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */const it=e("Brain",[["path",{d:"M12 5a3 3 0 1 0-5.997.125 4 4 0 0 0-2.526 5.77 4 4 0 0 0 .556 6.588A4 4 0 1 0 12 18Z",key:"l5xja"}],["path",{d:"M12 5a3 3 0 1 1 5.997.125 4 4 0 0 1 2.526 5.77 4 4 0 0 1-.556 6.588A4 4 0 1 1 12 18Z",key:"ep3f8r"}],["path",{d:"M15 13a4.5 4.5 0 0 1-3-4 4.5 4.5 0 0 1-3 4",key:"1p4c4q"}],["path",{d:"M17.599 6.5a3 3 0 0 0 .399-1.375",key:"tmeiqw"}],["path",{d:"M6.003 5.125A3 3 0 0 0 6.401 6.5",key:"105sqy"}],["path",{d:"M3.477 10.896a4 4 0 0 1 .585-.396",key:"ql3yin"}],["path",{d:"M19.938 10.5a4 4 0 0 1 .585.396",key:"1qfode"}],["path",{d:"M6 18a4 4 0 0 1-1.967-.516",key:"2e4loj"}],["path",{d:"M19.967 17.484A4 4 0 0 1 18 18",key:"159ez6"}]]),st=e("Cpu",[["rect",{width:"16",height:"16",x:"4",y:"4",rx:"2",key:"14l7u7"}],["rect",{width:"6",height:"6",x:"9",y:"9",rx:"1",key:"5aljv4"}],["path",{d:"M15 2v2",key:"13l42r"}],["path",{d:"M15 20v2",key:"15mkzm"}],["path",{d:"M2 15h2",key:"1gxd5l"}],["path",{d:"M2 9h2",key:"1bbxkp"}],["path",{d:"M20 15h2",key:"19e6y8"}],["path",{d:"M20 9h2",key:"19tzq7"}],["path",{d:"M9 2v2",key:"165o2o"}],["path",{d:"M9 20v2",key:"i2bqo8"}]]),at=e("FlaskConical",[["path",{d:"M10 2v7.527a2 2 0 0 1-.211.896L4.72 20.55a1 1 0 0 0 .9 1.45h12.76a1 1 0 0 0 .9-1.45l-5.069-10.127A2 2 0 0 1 14 9.527V2",key:"pzvekw"}],["path",{d:"M8.5 2h7",key:"csnxdl"}],["path",{d:"M7 16h10",key:"wp8him"}]]),ct=e("Grid3x3",[["rect",{width:"18",height:"18",x:"3",y:"3",rx:"2",key:"afitv7"}],["path",{d:"M3 9h18",key:"1pudct"}],["path",{d:"M3 15h18",key:"5xshup"}],["path",{d:"M9 3v18",key:"fh3hqa"}],["path",{d:"M15 3v18",key:"14nvp0"}]]),lt=e("Monitor",[["rect",{width:"20",height:"14",x:"2",y:"3",rx:"2",key:"48i651"}],["line",{x1:"8",x2:"16",y1:"21",y2:"21",key:"1svkeh"}],["line",{x1:"12",x2:"12",y1:"17",y2:"21",key:"vw1qmm"}]]),ut=e("Package",[["path",{d:"m7.5 4.27 9 5.15",key:"1c824w"}],["path",{d:"M21 8a2 2 0 0 0-1-1.73l-7-4a2 2 0 0 0-2 0l-7 4A2 2 0 0 0 3 8v8a2 2 0 0 0 1 1.73l7 4a2 2 0 0 0 2 0l7-4A2 2 0 0 0 21 16Z",key:"hh9hay"}],["path",{d:"m3.3 7 8.7 5 8.7-5",key:"g66t2b"}],["path",{d:"M12 22V12",key:"d0xqtd"}]]),dt=e("PowerOff",[["path",{d:"M18.36 6.64A9 9 0 0 1 20.77 15",key:"dxknvb"}],["path",{d:"M6.16 6.16a9 9 0 1 0 12.68 12.68",key:"1x7qb5"}],["path",{d:"M12 2v4",key:"3427ic"}],["path",{d:"m2 2 20 20",key:"1ooewy"}]]);
/**
 * @license lucide-react v0.428.0 - ISC
 *
 * This source code is licensed under the ISC license.
 * See the LICENSE file in the root directory of this source tree.
 */function ft(e,t){return(ft=Object.setPrototypeOf?Object.setPrototypeOf.bind():function(e,t){return e.__proto__=t,e})(e,t)}function ht(e){if(void 0===e)throw new ReferenceError("this hasn't been initialised - super() hasn't been called");return e}const pt=r("flex items-center justify-center m-0 leading-none font-medium border border-foreground/10 shadow-xsSolid hover:cursor-pointer active:shadow-none dark:border-border text-sm",{variants:{color:{gray:"mo-button gray",white:"mo-button white",green:"mo-button green",red:"mo-button red",yellow:"mo-button yellow","hint-green":"mo-button hint-green",disabled:"mo-button disabled"},shape:{rectangle:"rounded",circle:"rounded-full"},size:{small:"",medium:""}},compoundVariants:[{size:"small",shape:"circle",class:"h-[24px] w-[24px] px-[5.5px] py-[5.5px]"},{size:"medium",shape:"circle",class:"px-2 py-2"},{size:"small",shape:"rectangle",class:"px-1 py-1 h-[24px] w-[24px]"},{size:"medium",shape:"rectangle",class:"px-3 py-2"}],defaultVariants:{color:"gray",size:"medium",shape:"rectangle"}}),gt=r("font-mono w-full flex-1 inline-flex items-center justify-center rounded px-2.5 text-foreground/60 h-[36px] hover:shadow-md hover:cursor-pointer focus:shadow-md focus:outline-none text-[hsl(0, 0%, 43.5%)] bg-transparent hover:bg-background focus:bg-background"),mt=o.forwardRef(((e,t)=>{const n=i(15);let r,o,c,l,u,d,f;return n[0]!==e?(({color:r,shape:o,size:c,className:l,...u}=e),n[0]=e,n[1]=r,n[2]=o,n[3]=c,n[4]=l,n[5]=u):(r=n[1],o=n[2],c=n[3],l=n[4],u=n[5]),n[6]!==r||n[7]!==o||n[8]!==c||n[9]!==l?(d=s(pt({color:r,shape:o,size:c}),l),n[6]=r,n[7]=o,n[8]=c,n[9]=l,n[10]=d):d=n[10],n[11]!==t||n[12]!==d||n[13]!==u?(f=a.jsx("button",{ref:t,className:d,...u,children:u.children}),n[11]=t,n[12]=d,n[13]=u,n[14]=f):f=n[14],f}));mt.displayName="Button";o.forwardRef(((e,t)=>{const n=i(9);let r,o,c,l;return n[0]!==e?(({className:r,...o}=e),n[0]=e,n[1]=r,n[2]=o):(r=n[1],o=n[2]),n[3]!==r?(c=s(gt(),r),n[3]=r,n[4]=c):c=n[4],n[5]!==t||n[6]!==c||n[7]!==o?(l=a.jsx("input",{ref:t,className:c,...o}),n[5]=t,n[6]=c,n[7]=o,n[8]=l):l=n[8],l})).displayName="Input";const vt=e=>{const t=i(7),{openConfirm:n,closeModal:r}=c(),o=bt;let s,f,h;return t[0]!==n||t[1]!==e.description||t[2]!==r?(s=t=>{t.stopPropagation(),n({title:"Shutdown",description:e.description,variant:"destructive",confirmAction:a.jsx(l,{onClick:e=>{o(),r()},"aria-label":"Confirm Shutdown",children:"Shutdown"})})},t[0]=n,t[1]=e.description,t[2]=r,t[3]=s):s=t[3],t[4]===Symbol.for("react.memo_cache_sentinel")?(f=a.jsx(u,{strokeWidth:1}),t[4]=f):f=t[4],t[5]!==s?(h=a.jsx(d,{content:"Shutdown",children:a.jsx(mt,{"aria-label":"Shutdown","data-testid":"shutdown-button",shape:"circle",size:"small",color:"red",className:"h-[27px] w-[27px]",onClick:s,children:f})}),t[5]=s,t[6]=h):h=t[6],h};function yt(){window.close()}function bt(){f(),setTimeout(yt,200)}function wt(e,t){const n=i(11),{delayMs:r,whenVisible:o,disabled:s}=t,a=void 0!==s&&s,c=h.useRef();let l,u,d,f,g;return n[0]!==e?(l=()=>{c.current=e},u=[e],n[0]=e,n[1]=l,n[2]=u):(l=n[1],u=n[2]),h.useEffect(l,u),n[3]!==r||n[4]!==a||n[5]!==o?(d=()=>{if(null===r||a)return;const e=setInterval((()=>{var e;o&&"visible"!==document.visibilityState||null==(e=c.current)||e.call(c)}),r);return()=>clearInterval(e)},f=[r,o,a],n[3]=r,n[4]=a,n[5]=o,n[6]=d,n[7]=f):(d=n[6],f=n[7]),h.useEffect(d,f),n[8]!==o||n[9]!==a?(g=()=>{var e;"visible"===document.visibilityState&&o&&!a&&(null==(e=c.current)||e.call(c))},n[8]=o,n[9]=a,n[10]=g):g=n[10],p(document,"visibilitychange",g),null}const xt=e=>{const t=i(2),{children:n}=e;let r;return t[0]!==n?(r=a.jsx("div",{className:"text-md font-semibold text-muted-foreground uppercase tracking-wide  mb-1",children:n}),t[0]=n,t[1]=r):r=t[1],r},Tt=e=>{const t=i(10);let n,r,o,s,c;return t[0]!==e?(({children:o,className:n,...r}=e),t[0]=e,t[1]=n,t[2]=r,t[3]=o):(n=t[1],r=t[2],o=t[3]),t[4]!==n?(s=g("text-sm font-semibold underline-offset-2 text-accent-foreground uppercase tracking-wide",n),t[4]=n,t[5]=s):s=t[5],t[6]!==r||t[7]!==s||t[8]!==o?(c=a.jsx("div",{...r,className:s,children:o}),t[6]=r,t[7]=s,t[8]=o,t[9]=c):c=t[9],c},Dt=e=>{const t=i(2),{children:n}=e;let r;return t[0]!==n?(r=a.jsx("p",{className:"text-sm text-muted-foreground",children:n}),t[0]=n,t[1]=r):r=t[1],r},St="marimo:copilot:signedIn",Ot=m(St,null,void 0,{getOnInit:!0}),Rt=v(null);function jt(){const e="true"===localStorage.getItem(St),t=b();return e&&"github"===t.completion.copilot}var Ct,Nt={},kt={},Pt={exports:{}},Et="object"==typeof Reflect?Reflect:null,It=Et&&"function"==typeof Et.apply?Et.apply:function(e,t,n){return Function.prototype.apply.call(e,t,n)};Ct=Et&&"function"==typeof Et.ownKeys?Et.ownKeys:Object.getOwnPropertySymbols?function(e){return Object.getOwnPropertyNames(e).concat(Object.getOwnPropertySymbols(e))}:function(e){return Object.getOwnPropertyNames(e)};var _t=Number.isNaN||function(e){return e!=e};function Mt(){Mt.init.call(this)}Pt.exports=Mt,Pt.exports.once=function(e,t){return new Promise((function(n,r){function o(n){e.removeListener(t,i),r(n)}function i(){"function"==typeof e.removeListener&&e.removeListener("error",o),n([].slice.call(arguments))}$t(e,t,i,{once:!0}),"error"!==t&&function(e,t,n){"function"==typeof e.on&&$t(e,"error",t,n)}(e,o,{once:!0})}))},Mt.EventEmitter=Mt,Mt.prototype._events=void 0,Mt.prototype._eventsCount=0,Mt.prototype._maxListeners=void 0;var qt=10;function At(e){if("function"!=typeof e)throw new TypeError('The "listener" argument must be of type Function. Received type '+typeof e)}function Lt(e){return void 0===e._maxListeners?Mt.defaultMaxListeners:e._maxListeners}function Ft(e,t,n,r){var o,i,s,a;if(At(n),void 0===(i=e._events)?(i=e._events=Object.create(null),e._eventsCount=0):(void 0!==i.newListener&&(e.emit("newListener",t,n.listener?n.listener:n),i=e._events),s=i[t]),void 0===s)s=i[t]=n,++e._eventsCount;else if("function"==typeof s?s=i[t]=r?[n,s]:[s,n]:r?s.unshift(n):s.push(n),(o=Lt(e))>0&&s.length>o&&!s.warned){s.warned=!0;var c=new Error("Possible EventEmitter memory leak detected. "+s.length+" "+String(t)+" listeners added. Use emitter.setMaxListeners() to increase limit");c.name="MaxListenersExceededWarning",c.emitter=e,c.type=t,c.count=s.length,a=c,console&&console.warn&&console.warn(a)}return e}function Ht(){if(!this.fired)return this.target.removeListener(this.type,this.wrapFn),this.fired=!0,0===arguments.length?this.listener.call(this.target):this.listener.apply(this.target,arguments)}function Wt(e,t,n){var r={fired:!1,wrapFn:void 0,target:e,type:t,listener:n},o=Ht.bind(r);return o.listener=n,r.wrapFn=o,o}function Ut(e,t,n){var r=e._events;if(void 0===r)return[];var o=r[t];return void 0===o?[]:"function"==typeof o?n?[o.listener||o]:[o]:n?function(e){for(var t=new Array(e.length),n=0;n<t.length;++n)t[n]=e[n].listener||e[n];return t}(o):zt(o,o.length)}function Bt(e){var t=this._events;if(void 0!==t){var n=t[e];if("function"==typeof n)return 1;if(void 0!==n)return n.length}return 0}function zt(e,t){for(var n=new Array(t),r=0;r<t;++r)n[r]=e[r];return n}function $t(e,t,n,r){if("function"==typeof e.on)r.once?e.once(t,n):e.on(t,n);else{if("function"!=typeof e.addEventListener)throw new TypeError('The "emitter" argument must be of type EventEmitter. Received type '+typeof e);e.addEventListener(t,(function o(i){r.once&&e.removeEventListener(t,o),n(i)}))}}Object.defineProperty(Mt,"defaultMaxListeners",{enumerable:!0,get:function(){return qt},set:function(e){if("number"!=typeof e||e<0||_t(e))throw new RangeError('The value of "defaultMaxListeners" is out of range. It must be a non-negative number. Received '+e+".");qt=e}}),Mt.init=function(){void 0!==this._events&&this._events!==Object.getPrototypeOf(this)._events||(this._events=Object.create(null),this._eventsCount=0),this._maxListeners=this._maxListeners||void 0},Mt.prototype.setMaxListeners=function(e){if("number"!=typeof e||e<0||_t(e))throw new RangeError('The value of "n" is out of range. It must be a non-negative number. Received '+e+".");return this._maxListeners=e,this},Mt.prototype.getMaxListeners=function(){return Lt(this)},Mt.prototype.emit=function(e){for(var t=[],n=1;n<arguments.length;n++)t.push(arguments[n]);var r="error"===e,o=this._events;if(void 0!==o)r=r&&void 0===o.error;else if(!r)return!1;if(r){var i;if(t.length>0&&(i=t[0]),i instanceof Error)throw i;var s=new Error("Unhandled error."+(i?" ("+i.message+")":""));throw s.context=i,s}var a=o[e];if(void 0===a)return!1;if("function"==typeof a)It(a,this,t);else{var c=a.length,l=zt(a,c);for(n=0;n<c;++n)It(l[n],this,t)}return!0},Mt.prototype.addListener=function(e,t){return Ft(this,e,t,!1)},Mt.prototype.on=Mt.prototype.addListener,Mt.prototype.prependListener=function(e,t){return Ft(this,e,t,!0)},Mt.prototype.once=function(e,t){return At(t),this.on(e,Wt(this,e,t)),this},Mt.prototype.prependOnceListener=function(e,t){return At(t),this.prependListener(e,Wt(this,e,t)),this},Mt.prototype.removeListener=function(e,t){var n,r,o,i,s;if(At(t),void 0===(r=this._events))return this;if(void 0===(n=r[e]))return this;if(n===t||n.listener===t)0==--this._eventsCount?this._events=Object.create(null):(delete r[e],r.removeListener&&this.emit("removeListener",e,n.listener||t));else if("function"!=typeof n){for(o=-1,i=n.length-1;i>=0;i--)if(n[i]===t||n[i].listener===t){s=n[i].listener,o=i;break}if(o<0)return this;0===o?n.shift():function(e,t){for(;t+1<e.length;t++)e[t]=e[t+1];e.pop()}(n,o),1===n.length&&(r[e]=n[0]),void 0!==r.removeListener&&this.emit("removeListener",e,s||t)}return this},Mt.prototype.off=Mt.prototype.removeListener,Mt.prototype.removeAllListeners=function(e){var t,n,r;if(void 0===(n=this._events))return this;if(void 0===n.removeListener)return 0===arguments.length?(this._events=Object.create(null),this._eventsCount=0):void 0!==n[e]&&(0==--this._eventsCount?this._events=Object.create(null):delete n[e]),this;if(0===arguments.length){var o,i=Object.keys(n);for(r=0;r<i.length;++r)"removeListener"!==(o=i[r])&&this.removeAllListeners(o);return this.removeAllListeners("removeListener"),this._events=Object.create(null),this._eventsCount=0,this}if("function"==typeof(t=n[e]))this.removeListener(e,t);else if(void 0!==t)for(r=t.length-1;r>=0;r--)this.removeListener(e,t[r]);return this},Mt.prototype.listeners=function(e){return Ut(this,e,!0)},Mt.prototype.rawListeners=function(e){return Ut(this,e,!1)},Mt.listenerCount=function(e,t){return"function"==typeof e.listenerCount?e.listenerCount(t):Bt.call(e,t)},Mt.prototype.listenerCount=Bt,Mt.prototype.eventNames=function(){return this._eventsCount>0?Ct(this._events):[]};var Vt=Pt.exports;!function(e){var t=T&&T.__awaiter||function(e,t,n,r){return new(n||(n=Promise))((function(o,i){function s(e){try{c(r.next(e))}catch(t){i(t)}}function a(e){try{c(r.throw(e))}catch(t){i(t)}}function c(e){var t;e.done?o(e.value):(t=e.value,t instanceof n?t:new n((function(e){e(t)}))).then(s,a)}c((r=r.apply(e,t||[])).next())}))},n=T&&T.__generator||function(e,t){var n,r,o,i,s={label:0,sent:function(){if(1&o[0])throw o[1];return o[1]},trys:[],ops:[]};return i={next:a(0),throw:a(1),return:a(2)},"function"==typeof Symbol&&(i[Symbol.iterator]=function(){return this}),i;function a(i){return function(a){return function(i){if(n)throw new TypeError("Generator is already executing.");for(;s;)try{if(n=1,r&&(o=2&i[0]?r.return:i[0]?r.throw||((o=r.return)&&o.call(r),0):r.next)&&!(o=o.call(r,i[1])).done)return o;switch(r=0,o&&(i=[2&i[0],o.value]),i[0]){case 0:case 1:o=i;break;case 4:return s.label++,{value:i[1],done:!1};case 5:s.label++,r=i[1],i=[0];continue;case 7:i=s.ops.pop(),s.trys.pop();continue;default:if(!(o=s.trys,(o=o.length>0&&o[o.length-1])||6!==i[0]&&2!==i[0])){s=0;continue}if(3===i[0]&&(!o||i[1]>o[0]&&i[1]<o[3])){s.label=i[1];break}if(6===i[0]&&s.label<o[1]){s.label=o[1],o=i;break}if(o&&s.label<o[2]){s.label=o[2],s.ops.push(i);break}o[2]&&s.ops.pop(),s.trys.pop();continue}i=t.call(e,s)}catch(a){i=[6,a],r=0}finally{n=o=0}if(5&i[0])throw i[1];return{value:i[0]?i[1]:void 0,done:!0}}([i,a])}}};Object.defineProperty(e,"__esModule",{value:!0}),e.defaultNextRequest=void 0;var r=Vt;e.defaultNextRequest=function(){var e=-1;return function(){return++e}};var o=function(){function o(t,n){void 0===n&&(n=e.defaultNextRequest()),this.batch=[],this.batchStarted=!1,this.lastId=-1,this.transports=t,this.requests={},this.connectPromise=this.connect(),this.requestChannel=new r.EventEmitter,this.nextID=n}return o.prototype.connect=function(){var e=this;return Promise.all(this.transports.map((function(r){return t(e,void 0,void 0,(function(){return n(this,(function(e){switch(e.label){case 0:return r.subscribe("error",this.handleError.bind(this)),r.subscribe("notification",this.handleNotification.bind(this)),[4,r.connect()];case 1:return e.sent(),[2]}}))}))})))},o.prototype.getPrimaryTransport=function(){return this.transports[0]},o.prototype.request=function(e,r,o){return void 0===r&&(r=!1),t(this,void 0,void 0,(function(){var t,i,s,a=this;return n(this,(function(n){return t=this.nextID().toString(),i=r?null:t,s={request:this.makeRequest(e.method,e.params||[],i),internalID:t},this.batchStarted?[2,new Promise((function(e,t){a.batch.push({resolve:e,reject:t,request:s})}))]:[2,this.getPrimaryTransport().sendData(s,o)]}))}))},o.prototype.close=function(){this.requestChannel.removeAllListeners(),this.transports.forEach((function(e){e.unsubscribe(),e.close()}))},o.prototype.startBatch=function(){this.batchStarted=!0},o.prototype.stopBatch=function(){if(!1===this.batchStarted)throw new Error("cannot end that which has never started");0!==this.batch.length?(this.getPrimaryTransport().sendData(this.batch),this.batch=[],this.batchStarted=!1):this.batchStarted=!1},o.prototype.makeRequest=function(e,t,n){return n?{jsonrpc:"2.0",id:n,method:e,params:t}:{jsonrpc:"2.0",method:e,params:t}},o.prototype.handleError=function(e){this.requestChannel.emit("error",e)},o.prototype.handleNotification=function(e){this.requestChannel.emit("notification",e)},o}();e.default=o}(kt);var Kt={},Gt={},Jt={},Yt={};!function(e){var t=T&&T.__extends||function(){var e=function(t,n){return(e=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,t){e.__proto__=t}||function(e,t){for(var n in t)t.hasOwnProperty(n)&&(e[n]=t[n])})(t,n)};return function(t,n){function r(){this.constructor=t}e(t,n),t.prototype=null===n?Object.create(n):(r.prototype=n.prototype,new r)}}();Object.defineProperty(e,"__esModule",{value:!0}),e.convertJSONToRPCError=e.JSONRPCError=e.ERR_UNKNOWN=e.ERR_MISSIING_ID=e.ERR_TIMEOUT=void 0,e.ERR_TIMEOUT=7777,e.ERR_MISSIING_ID=7878,e.ERR_UNKNOWN=7979;var n=function(e){function n(t,n,r){var o=this.constructor,i=e.call(this,t)||this;return i.message=t,i.code=n,i.data=r,Object.setPrototypeOf(i,o.prototype),i}return t(n,e),n}(Error);e.JSONRPCError=n,e.convertJSONToRPCError=function(t){if(t.error){var r=t.error,o=r.message,i=r.code,s=r.data;return new n(o,i,s)}return new n("Unknown error",e.ERR_UNKNOWN,t)}}(Yt),Object.defineProperty(Jt,"__esModule",{value:!0}),Jt.TransportRequestManager=void 0;var Xt=Vt,Zt=Yt,Qt=function(){function e(){this.pendingRequest={},this.pendingBatchRequest={},this.transportEventChannel=new Xt.EventEmitter}return e.prototype.addRequest=function(e,t){return this.transportEventChannel.emit("pending",e),e instanceof Array?(this.addBatchReq(e,t),Promise.resolve()):this.addReq(e.internalID,t)},e.prototype.settlePendingRequest=function(e,t){var n=this;e.forEach((function(e){var r=n.pendingRequest[e.internalID];delete n.pendingBatchRequest[e.internalID],void 0!==r&&(t?r.reject(t):(r.resolve(),null!==e.request.id&&void 0!==e.request.id||delete n.pendingRequest[e.internalID]))}))},e.prototype.isPendingRequest=function(e){return this.pendingRequest.hasOwnProperty(e)},e.prototype.resolveResponse=function(e,t){void 0===t&&(t=!0);var n=e;try{if(n=JSON.parse(e),!1===this.checkJSONRPC(n))return;return n instanceof Array?this.resolveBatch(n,t):this.resolveRes(n,t)}catch(o){var r=new Zt.JSONRPCError("Bad response format",Zt.ERR_UNKNOWN,e);return t&&this.transportEventChannel.emit("error",r),r}},e.prototype.addBatchReq=function(e,t){var n=this;return e.forEach((function(e){var t=e.resolve,r=e.reject,o=e.request.internalID;n.pendingBatchRequest[o]=!0,n.pendingRequest[o]={resolve:t,reject:r}})),Promise.resolve()},e.prototype.addReq=function(e,t){var n=this;return new Promise((function(r,o){null!==t&&t&&n.setRequestTimeout(e,t,o),n.pendingRequest[e]={resolve:r,reject:o}}))},e.prototype.checkJSONRPC=function(e){var t=[e];return e instanceof Array&&(t=e),t.every((function(e){return void 0!==e.result||void 0!==e.error||void 0!==e.method}))},e.prototype.processResult=function(e,t){if(e.error){var n=Zt.convertJSONToRPCError(e);t.reject(n)}else t.resolve(e.result)},e.prototype.resolveBatch=function(e,t){var n=this,r=e.map((function(e){return n.resolveRes(e,t)})).filter((function(e){return e}));if(r.length>0)return r[0]},e.prototype.resolveRes=function(e,t){var n,r=e.id,o=e.error,i=this.pendingRequest[r];return i?(delete this.pendingRequest[r],this.processResult(e,i),void this.transportEventChannel.emit("response",e)):void 0!==r||void 0!==o?(o&&(n=Zt.convertJSONToRPCError(e)),t&&o&&n&&this.transportEventChannel.emit("error",n),n):void this.transportEventChannel.emit("notification",e)},e.prototype.setRequestTimeout=function(e,t,n){var r=this;setTimeout((function(){delete r.pendingRequest[e],n(new Zt.JSONRPCError("Request timeout request took longer than "+t+" ms to resolve",Zt.ERR_TIMEOUT))}),t)},e}();Jt.TransportRequestManager=Qt,Object.defineProperty(Gt,"__esModule",{value:!0});var en=Gt.Transport=void 0,tn=Jt,nn=function(){function e(){this.transportRequestManager=new tn.TransportRequestManager,this.transportRequestManager.transportEventChannel.on("error",(function(){}))}return e.prototype.subscribe=function(e,t){this.transportRequestManager.transportEventChannel.addListener(e,t)},e.prototype.unsubscribe=function(e,t){if(!e)return this.transportRequestManager.transportEventChannel.removeAllListeners();e&&t&&this.transportRequestManager.transportEventChannel.removeListener(e,t)},e.prototype.parseData=function(e){return e instanceof Array?e.map((function(e){return e.request.request})):e.request},e}();en=Gt.Transport=nn;var rn,on={};rn=on,Object.defineProperty(rn,"__esModule",{value:!0}),rn.getNotifications=rn.getBatchRequests=rn.isNotification=void 0,rn.isNotification=function(e){return void 0===e.request.id||null===e.request.id},rn.getBatchRequests=function(e){return e instanceof Array?e.filter((function(e){var t=e.request.request.id;return null!=t})).map((function(e){return e.request})):[]},rn.getNotifications=function(e){return e instanceof Array?e.filter((function(e){return rn.isNotification(e.request)})).map((function(e){return e.request})):rn.isNotification(e)?[e]:[]};var sn=T&&T.__extends||function(){var e=function(t,n){return(e=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,t){e.__proto__=t}||function(e,t){for(var n in t)t.hasOwnProperty(n)&&(e[n]=t[n])})(t,n)};return function(t,n){function r(){this.constructor=t}e(t,n),t.prototype=null===n?Object.create(n):(r.prototype=n.prototype,new r)}}();Object.defineProperty(Kt,"__esModule",{value:!0});var an=on,cn=Yt,ln=function(e){function t(t,n,r){var o=e.call(this)||this;return o.connection=t,o.reqUri=n,o.resUri=r,o}return sn(t,e),t.prototype.connect=function(){var e=this;return this.connection.on(this.resUri,(function(t){e.transportRequestManager.resolveResponse(t)})),Promise.resolve()},t.prototype.sendData=function(e,t){void 0===t&&(t=null);var n=this.transportRequestManager.addRequest(e,t),r=an.getNotifications(e),o=this.parseData(e);try{return this.connection.emit(this.reqUri,o),this.transportRequestManager.settlePendingRequest(r),n}catch(s){var i=new cn.JSONRPCError(s.message,cn.ERR_UNKNOWN,s);return this.transportRequestManager.settlePendingRequest(r,i),Promise.reject(i)}},t.prototype.close=function(){this.connection.removeAllListeners()},t}(Gt.Transport);Kt.default=ln;var un={},dn="undefined"!=typeof globalThis&&globalThis||"undefined"!=typeof self&&self||"undefined"!=typeof global&&global||{},fn="URLSearchParams"in dn,hn="Symbol"in dn&&"iterator"in Symbol,pn="FileReader"in dn&&"Blob"in dn&&function(){try{return new Blob,!0}catch(e){return!1}}(),gn="FormData"in dn,mn="ArrayBuffer"in dn;if(mn)var vn=["[object Int8Array]","[object Uint8Array]","[object Uint8ClampedArray]","[object Int16Array]","[object Uint16Array]","[object Int32Array]","[object Uint32Array]","[object Float32Array]","[object Float64Array]"],yn=ArrayBuffer.isView||function(e){return e&&vn.indexOf(Object.prototype.toString.call(e))>-1};function bn(e){if("string"!=typeof e&&(e=String(e)),/[^a-z0-9\-#$%&'*+.^_`|~!]/i.test(e)||""===e)throw new TypeError('Invalid character in header field name: "'+e+'"');return e.toLowerCase()}function wn(e){return"string"!=typeof e&&(e=String(e)),e}function xn(e){var t={next:function(){var t=e.shift();return{done:void 0===t,value:t}}};return hn&&(t[Symbol.iterator]=function(){return t}),t}function Tn(e){this.map={},e instanceof Tn?e.forEach((function(e,t){this.append(t,e)}),this):Array.isArray(e)?e.forEach((function(e){if(2!=e.length)throw new TypeError("Headers constructor: expected name/value pair to be length 2, found"+e.length);this.append(e[0],e[1])}),this):e&&Object.getOwnPropertyNames(e).forEach((function(t){this.append(t,e[t])}),this)}function Dn(e){if(!e._noBody)return e.bodyUsed?Promise.reject(new TypeError("Already read")):void(e.bodyUsed=!0)}function Sn(e){return new Promise((function(t,n){e.onload=function(){t(e.result)},e.onerror=function(){n(e.error)}}))}function On(e){var t=new FileReader,n=Sn(t);return t.readAsArrayBuffer(e),n}function Rn(e){if(e.slice)return e.slice(0);var t=new Uint8Array(e.byteLength);return t.set(new Uint8Array(e)),t.buffer}function jn(){return this.bodyUsed=!1,this._initBody=function(e){var t;this.bodyUsed=this.bodyUsed,this._bodyInit=e,e?"string"==typeof e?this._bodyText=e:pn&&Blob.prototype.isPrototypeOf(e)?this._bodyBlob=e:gn&&FormData.prototype.isPrototypeOf(e)?this._bodyFormData=e:fn&&URLSearchParams.prototype.isPrototypeOf(e)?this._bodyText=e.toString():mn&&pn&&((t=e)&&DataView.prototype.isPrototypeOf(t))?(this._bodyArrayBuffer=Rn(e.buffer),this._bodyInit=new Blob([this._bodyArrayBuffer])):mn&&(ArrayBuffer.prototype.isPrototypeOf(e)||yn(e))?this._bodyArrayBuffer=Rn(e):this._bodyText=e=Object.prototype.toString.call(e):(this._noBody=!0,this._bodyText=""),this.headers.get("content-type")||("string"==typeof e?this.headers.set("content-type","text/plain;charset=UTF-8"):this._bodyBlob&&this._bodyBlob.type?this.headers.set("content-type",this._bodyBlob.type):fn&&URLSearchParams.prototype.isPrototypeOf(e)&&this.headers.set("content-type","application/x-www-form-urlencoded;charset=UTF-8"))},pn&&(this.blob=function(){var e=Dn(this);if(e)return e;if(this._bodyBlob)return Promise.resolve(this._bodyBlob);if(this._bodyArrayBuffer)return Promise.resolve(new Blob([this._bodyArrayBuffer]));if(this._bodyFormData)throw new Error("could not read FormData body as blob");return Promise.resolve(new Blob([this._bodyText]))}),this.arrayBuffer=function(){if(this._bodyArrayBuffer){var e=Dn(this);return e||(ArrayBuffer.isView(this._bodyArrayBuffer)?Promise.resolve(this._bodyArrayBuffer.buffer.slice(this._bodyArrayBuffer.byteOffset,this._bodyArrayBuffer.byteOffset+this._bodyArrayBuffer.byteLength)):Promise.resolve(this._bodyArrayBuffer))}if(pn)return this.blob().then(On);throw new Error("could not read as ArrayBuffer")},this.text=function(){var e,t,n,r,o,i=Dn(this);if(i)return i;if(this._bodyBlob)return e=this._bodyBlob,t=new FileReader,n=Sn(t),r=/charset=([A-Za-z0-9_-]+)/.exec(e.type),o=r?r[1]:"utf-8",t.readAsText(e,o),n;if(this._bodyArrayBuffer)return Promise.resolve(function(e){for(var t=new Uint8Array(e),n=new Array(t.length),r=0;r<t.length;r++)n[r]=String.fromCharCode(t[r]);return n.join("")}(this._bodyArrayBuffer));if(this._bodyFormData)throw new Error("could not read FormData body as text");return Promise.resolve(this._bodyText)},gn&&(this.formData=function(){return this.text().then(kn)}),this.json=function(){return this.text().then(JSON.parse)},this}Tn.prototype.append=function(e,t){e=bn(e),t=wn(t);var n=this.map[e];this.map[e]=n?n+", "+t:t},Tn.prototype.delete=function(e){delete this.map[bn(e)]},Tn.prototype.get=function(e){return e=bn(e),this.has(e)?this.map[e]:null},Tn.prototype.has=function(e){return this.map.hasOwnProperty(bn(e))},Tn.prototype.set=function(e,t){this.map[bn(e)]=wn(t)},Tn.prototype.forEach=function(e,t){for(var n in this.map)this.map.hasOwnProperty(n)&&e.call(t,this.map[n],n,this)},Tn.prototype.keys=function(){var e=[];return this.forEach((function(t,n){e.push(n)})),xn(e)},Tn.prototype.values=function(){var e=[];return this.forEach((function(t){e.push(t)})),xn(e)},Tn.prototype.entries=function(){var e=[];return this.forEach((function(t,n){e.push([n,t])})),xn(e)},hn&&(Tn.prototype[Symbol.iterator]=Tn.prototype.entries);var Cn=["CONNECT","DELETE","GET","HEAD","OPTIONS","PATCH","POST","PUT","TRACE"];function Nn(e,t){if(!(this instanceof Nn))throw new TypeError('Please use the "new" operator, this DOM object constructor cannot be called as a function.');var n,r,o=(t=t||{}).body;if(e instanceof Nn){if(e.bodyUsed)throw new TypeError("Already read");this.url=e.url,this.credentials=e.credentials,t.headers||(this.headers=new Tn(e.headers)),this.method=e.method,this.mode=e.mode,this.signal=e.signal,o||null==e._bodyInit||(o=e._bodyInit,e.bodyUsed=!0)}else this.url=String(e);if(this.credentials=t.credentials||this.credentials||"same-origin",!t.headers&&this.headers||(this.headers=new Tn(t.headers)),this.method=(n=t.method||this.method||"GET",r=n.toUpperCase(),Cn.indexOf(r)>-1?r:n),this.mode=t.mode||this.mode||null,this.signal=t.signal||this.signal||function(){if("AbortController"in dn)return(new AbortController).signal}(),this.referrer=null,("GET"===this.method||"HEAD"===this.method)&&o)throw new TypeError("Body not allowed for GET or HEAD requests");if(this._initBody(o),!("GET"!==this.method&&"HEAD"!==this.method||"no-store"!==t.cache&&"no-cache"!==t.cache)){var i=/([?&])_=[^&]*/;if(i.test(this.url))this.url=this.url.replace(i,"$1_="+(new Date).getTime());else{this.url+=(/\?/.test(this.url)?"&":"?")+"_="+(new Date).getTime()}}}function kn(e){var t=new FormData;return e.trim().split("&").forEach((function(e){if(e){var n=e.split("="),r=n.shift().replace(/\+/g," "),o=n.join("=").replace(/\+/g," ");t.append(decodeURIComponent(r),decodeURIComponent(o))}})),t}function Pn(e,t){if(!(this instanceof Pn))throw new TypeError('Please use the "new" operator, this DOM object constructor cannot be called as a function.');if(t||(t={}),this.type="default",this.status=void 0===t.status?200:t.status,this.status<200||this.status>599)throw new RangeError("Failed to construct 'Response': The status provided (0) is outside the range [200, 599].");this.ok=this.status>=200&&this.status<300,this.statusText=void 0===t.statusText?"":""+t.statusText,this.headers=new Tn(t.headers),this.url=t.url||"",this._initBody(e)}Nn.prototype.clone=function(){return new Nn(this,{body:this._bodyInit})},jn.call(Nn.prototype),jn.call(Pn.prototype),Pn.prototype.clone=function(){return new Pn(this._bodyInit,{status:this.status,statusText:this.statusText,headers:new Tn(this.headers),url:this.url})},Pn.error=function(){var e=new Pn(null,{status:200,statusText:""});return e.status=0,e.type="error",e};var En=[301,302,303,307,308];Pn.redirect=function(e,t){if(-1===En.indexOf(t))throw new RangeError("Invalid status code");return new Pn(null,{status:t,headers:{location:e}})};var In=dn.DOMException;try{new In}catch(Ih){(In=function(e,t){this.message=e,this.name=t;var n=Error(e);this.stack=n.stack}).prototype=Object.create(Error.prototype),In.prototype.constructor=In}function _n(e,t){return new Promise((function(n,r){var o=new Nn(e,t);if(o.signal&&o.signal.aborted)return r(new In("Aborted","AbortError"));var i=new XMLHttpRequest;function s(){i.abort()}if(i.onload=function(){var e,t,r={status:i.status,statusText:i.statusText,headers:(e=i.getAllResponseHeaders()||"",t=new Tn,e.replace(/\r?\n[\t ]+/g," ").split("\r").map((function(e){return 0===e.indexOf("\n")?e.substr(1,e.length):e})).forEach((function(e){var n=e.split(":"),r=n.shift().trim();if(r){var o=n.join(":").trim();try{t.append(r,o)}catch(i){console.warn("Response "+i.message)}}})),t)};r.url="responseURL"in i?i.responseURL:r.headers.get("X-Request-URL");var o="response"in i?i.response:i.responseText;setTimeout((function(){n(new Pn(o,r))}),0)},i.onerror=function(){setTimeout((function(){r(new TypeError("Network request failed"))}),0)},i.ontimeout=function(){setTimeout((function(){r(new TypeError("Network request failed"))}),0)},i.onabort=function(){setTimeout((function(){r(new In("Aborted","AbortError"))}),0)},i.open(o.method,function(e){try{return""===e&&dn.location.href?dn.location.href:e}catch(t){return e}}(o.url),!0),"include"===o.credentials?i.withCredentials=!0:"omit"===o.credentials&&(i.withCredentials=!1),"responseType"in i&&(pn?i.responseType="blob":mn&&(i.responseType="arraybuffer")),t&&"object"==typeof t.headers&&!(t.headers instanceof Tn||dn.Headers&&t.headers instanceof dn.Headers)){var a=[];Object.getOwnPropertyNames(t.headers).forEach((function(e){a.push(bn(e)),i.setRequestHeader(e,wn(t.headers[e]))})),o.headers.forEach((function(e,t){-1===a.indexOf(t)&&i.setRequestHeader(t,e)}))}else o.headers.forEach((function(e,t){i.setRequestHeader(t,e)}));o.signal&&(o.signal.addEventListener("abort",s),i.onreadystatechange=function(){4===i.readyState&&o.signal.removeEventListener("abort",s)}),i.send(void 0===o._bodyInit?null:o._bodyInit)}))}_n.polyfill=!0,dn.fetch||(dn.fetch=_n,dn.Headers=Tn,dn.Request=Nn,dn.Response=Pn);var Mn=self.fetch.bind(self),qn=T&&T.__extends||function(){var e=function(t,n){return(e=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,t){e.__proto__=t}||function(e,t){for(var n in t)t.hasOwnProperty(n)&&(e[n]=t[n])})(t,n)};return function(t,n){function r(){this.constructor=t}e(t,n),t.prototype=null===n?Object.create(n):(r.prototype=n.prototype,new r)}}(),An=T&&T.__awaiter||function(e,t,n,r){return new(n||(n=Promise))((function(o,i){function s(e){try{c(r.next(e))}catch(t){i(t)}}function a(e){try{c(r.throw(e))}catch(t){i(t)}}function c(e){var t;e.done?o(e.value):(t=e.value,t instanceof n?t:new n((function(e){e(t)}))).then(s,a)}c((r=r.apply(e,t||[])).next())}))},Ln=T&&T.__generator||function(e,t){var n,r,o,i,s={label:0,sent:function(){if(1&o[0])throw o[1];return o[1]},trys:[],ops:[]};return i={next:a(0),throw:a(1),return:a(2)},"function"==typeof Symbol&&(i[Symbol.iterator]=function(){return this}),i;function a(i){return function(a){return function(i){if(n)throw new TypeError("Generator is already executing.");for(;s;)try{if(n=1,r&&(o=2&i[0]?r.return:i[0]?r.throw||((o=r.return)&&o.call(r),0):r.next)&&!(o=o.call(r,i[1])).done)return o;switch(r=0,o&&(i=[2&i[0],o.value]),i[0]){case 0:case 1:o=i;break;case 4:return s.label++,{value:i[1],done:!1};case 5:s.label++,r=i[1],i=[0];continue;case 7:i=s.ops.pop(),s.trys.pop();continue;default:if(!(o=s.trys,(o=o.length>0&&o[o.length-1])||6!==i[0]&&2!==i[0])){s=0;continue}if(3===i[0]&&(!o||i[1]>o[0]&&i[1]<o[3])){s.label=i[1];break}if(6===i[0]&&s.label<o[1]){s.label=o[1],o=i;break}if(o&&s.label<o[2]){s.label=o[2],s.ops.push(i);break}o[2]&&s.ops.pop(),s.trys.pop();continue}i=t.call(e,s)}catch(a){i=[6,a],r=0}finally{n=o=0}if(5&i[0])throw i[1];return{value:i[0]?i[1]:void 0,done:!0}}([i,a])}}},Fn=T&&T.__importDefault||function(e){return e&&e.__esModule?e:{default:e}};Object.defineProperty(un,"__esModule",{value:!0}),un.HTTPTransport=void 0;var Hn=Fn(Mn),Wn=on,Un=Yt,Bn=function(e){function t(n,r){var o=e.call(this)||this;return o.onlyNotifications=function(e){return e instanceof Array?e.every((function(e){return null===e.request.request.id||void 0===e.request.request.id})):null===e.request.id||void 0===e.request.id},o.uri=n,o.credentials=r&&r.credentials,o.headers=t.setupHeaders(r&&r.headers),o.injectedFetcher=null==r?void 0:r.fetcher,o}return qn(t,e),t.prototype.connect=function(){return Promise.resolve()},t.prototype.sendData=function(e,t){return void 0===t&&(t=null),An(this,void 0,void 0,(function(){var n,r,o,i,s,a,c,l;return Ln(this,(function(u){switch(u.label){case 0:n=this.transportRequestManager.addRequest(e,t),r=Wn.getNotifications(e),o=Wn.getBatchRequests(e),i=this.injectedFetcher||Hn.default,u.label=1;case 1:return u.trys.push([1,4,,5]),[4,i(this.uri,{method:"POST",headers:this.headers,body:JSON.stringify(this.parseData(e)),credentials:this.credentials})];case 2:return s=u.sent(),this.transportRequestManager.settlePendingRequest(r),this.onlyNotifications(e)?[2,Promise.resolve()]:[4,s.text()];case 3:return a=u.sent(),(l=this.transportRequestManager.resolveResponse(a))?(this.transportRequestManager.settlePendingRequest(o,l),[2,Promise.reject(l)]):[3,5];case 4:return c=u.sent(),l=new Un.JSONRPCError(c.message,Un.ERR_UNKNOWN,c),this.transportRequestManager.settlePendingRequest(r,l),this.transportRequestManager.settlePendingRequest(Wn.getBatchRequests(e),l),[2,Promise.reject(l)];case 5:return[2,n]}}))}))},t.prototype.close=function(){},t.setupHeaders=function(e){var t=new Headers(e);return t.set("Content-Type","application/json"),t},t}(Gt.Transport);un.HTTPTransport=Bn,un.default=Bn;var zn={},$n=null;"undefined"!=typeof WebSocket?$n=WebSocket:"undefined"!=typeof MozWebSocket?$n=MozWebSocket:"undefined"!=typeof global?$n=global.WebSocket||global.MozWebSocket:"undefined"!=typeof window?$n=window.WebSocket||window.MozWebSocket:"undefined"!=typeof self&&($n=self.WebSocket||self.MozWebSocket);const Vn=$n,Kn=D(Object.freeze(Object.defineProperty({__proto__:null,default:Vn},Symbol.toStringTag,{value:"Module"})));var Gn=T&&T.__extends||function(){var e=function(t,n){return(e=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,t){e.__proto__=t}||function(e,t){for(var n in t)t.hasOwnProperty(n)&&(e[n]=t[n])})(t,n)};return function(t,n){function r(){this.constructor=t}e(t,n),t.prototype=null===n?Object.create(n):(r.prototype=n.prototype,new r)}}(),Jn=T&&T.__awaiter||function(e,t,n,r){return new(n||(n=Promise))((function(o,i){function s(e){try{c(r.next(e))}catch(t){i(t)}}function a(e){try{c(r.throw(e))}catch(t){i(t)}}function c(e){var t;e.done?o(e.value):(t=e.value,t instanceof n?t:new n((function(e){e(t)}))).then(s,a)}c((r=r.apply(e,t||[])).next())}))},Yn=T&&T.__generator||function(e,t){var n,r,o,i,s={label:0,sent:function(){if(1&o[0])throw o[1];return o[1]},trys:[],ops:[]};return i={next:a(0),throw:a(1),return:a(2)},"function"==typeof Symbol&&(i[Symbol.iterator]=function(){return this}),i;function a(i){return function(a){return function(i){if(n)throw new TypeError("Generator is already executing.");for(;s;)try{if(n=1,r&&(o=2&i[0]?r.return:i[0]?r.throw||((o=r.return)&&o.call(r),0):r.next)&&!(o=o.call(r,i[1])).done)return o;switch(r=0,o&&(i=[2&i[0],o.value]),i[0]){case 0:case 1:o=i;break;case 4:return s.label++,{value:i[1],done:!1};case 5:s.label++,r=i[1],i=[0];continue;case 7:i=s.ops.pop(),s.trys.pop();continue;default:if(!(o=s.trys,(o=o.length>0&&o[o.length-1])||6!==i[0]&&2!==i[0])){s=0;continue}if(3===i[0]&&(!o||i[1]>o[0]&&i[1]<o[3])){s.label=i[1];break}if(6===i[0]&&s.label<o[1]){s.label=o[1],o=i;break}if(o&&s.label<o[2]){s.label=o[2],s.ops.push(i);break}o[2]&&s.ops.pop(),s.trys.pop();continue}i=t.call(e,s)}catch(a){i=[6,a],r=0}finally{n=o=0}if(5&i[0])throw i[1];return{value:i[0]?i[1]:void 0,done:!0}}([i,a])}}},Xn=T&&T.__importDefault||function(e){return e&&e.__esModule?e:{default:e}};Object.defineProperty(zn,"__esModule",{value:!0});var Zn=Xn(Kn),Qn=on,er=Yt,tr=function(e){function t(t){var n=e.call(this)||this;return n.uri=t,n.connection=new Zn.default(t),n}return Gn(t,e),t.prototype.connect=function(){var e=this;return new Promise((function(t,n){var r=function(){e.connection.removeEventListener("open",r),t()};e.connection.addEventListener("open",r),e.connection.addEventListener("message",(function(t){var n=t.data;e.transportRequestManager.resolveResponse(n)}))}))},t.prototype.sendData=function(e,t){return void 0===t&&(t=5e3),Jn(this,void 0,void 0,(function(){var n,r,o;return Yn(this,(function(i){n=this.transportRequestManager.addRequest(e,t),r=Qn.getNotifications(e);try{this.connection.send(JSON.stringify(this.parseData(e))),this.transportRequestManager.settlePendingRequest(r)}catch(Ih){o=new er.JSONRPCError(Ih.message,er.ERR_UNKNOWN,Ih),this.transportRequestManager.settlePendingRequest(r,o),this.transportRequestManager.settlePendingRequest(Qn.getBatchRequests(e),o),n=Promise.reject(o)}return[2,n]}))}))},t.prototype.close=function(){this.connection.close()},t}(Gt.Transport);zn.default=tr;var nr={},rr=T&&T.__extends||function(){var e=function(t,n){return(e=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,t){e.__proto__=t}||function(e,t){for(var n in t)t.hasOwnProperty(n)&&(e[n]=t[n])})(t,n)};return function(t,n){function r(){this.constructor=t}e(t,n),t.prototype=null===n?Object.create(n):(r.prototype=n.prototype,new r)}}(),or=T&&T.__awaiter||function(e,t,n,r){return new(n||(n=Promise))((function(o,i){function s(e){try{c(r.next(e))}catch(t){i(t)}}function a(e){try{c(r.throw(e))}catch(t){i(t)}}function c(e){var t;e.done?o(e.value):(t=e.value,t instanceof n?t:new n((function(e){e(t)}))).then(s,a)}c((r=r.apply(e,t||[])).next())}))},ir=T&&T.__generator||function(e,t){var n,r,o,i,s={label:0,sent:function(){if(1&o[0])throw o[1];return o[1]},trys:[],ops:[]};return i={next:a(0),throw:a(1),return:a(2)},"function"==typeof Symbol&&(i[Symbol.iterator]=function(){return this}),i;function a(i){return function(a){return function(i){if(n)throw new TypeError("Generator is already executing.");for(;s;)try{if(n=1,r&&(o=2&i[0]?r.return:i[0]?r.throw||((o=r.return)&&o.call(r),0):r.next)&&!(o=o.call(r,i[1])).done)return o;switch(r=0,o&&(i=[2&i[0],o.value]),i[0]){case 0:case 1:o=i;break;case 4:return s.label++,{value:i[1],done:!1};case 5:s.label++,r=i[1],i=[0];continue;case 7:i=s.ops.pop(),s.trys.pop();continue;default:if(!(o=s.trys,(o=o.length>0&&o[o.length-1])||6!==i[0]&&2!==i[0])){s=0;continue}if(3===i[0]&&(!o||i[1]>o[0]&&i[1]<o[3])){s.label=i[1];break}if(6===i[0]&&s.label<o[1]){s.label=o[1],o=i;break}if(o&&s.label<o[2]){s.label=o[2],s.ops.push(i);break}o[2]&&s.ops.pop(),s.trys.pop();continue}i=t.call(e,s)}catch(a){i=[6,a],r=0}finally{n=o=0}if(5&i[0])throw i[1];return{value:i[0]?i[1]:void 0,done:!0}}([i,a])}}};Object.defineProperty(nr,"__esModule",{value:!0});var sr=on,ar=function(e){function t(t){var n=e.call(this)||this;return n.messageHandler=function(e){n.transportRequestManager.resolveResponse(JSON.stringify(e.data))},n.uri=t,n.postMessageID="post-message-transport-"+Math.random(),n}return rr(t,e),t.prototype.createWindow=function(e){return new Promise((function(t,n){var r,o,i;o=e,i=window.screen.height,r=window.open(o,"inspector:popup","left=0,top=0,width=400,height="+i+",resizable,scrollbars=yes,status=1"),setTimeout((function(){t(r)}),3e3)}))},t.prototype.connect=function(){var e=this,t=/^(http|https):\/\/.*$/;return new Promise((function(n,r){return or(e,void 0,void 0,(function(){var e;return ir(this,(function(o){switch(o.label){case 0:return t.test(this.uri)||r(new Error("Bad URI")),e=this,[4,this.createWindow(this.uri)];case 1:return e.frame=o.sent(),window.addEventListener("message",this.messageHandler),n(),[2]}}))}))}))},t.prototype.sendData=function(e,t){return or(this,void 0,void 0,(function(){var t,n;return ir(this,(function(r){return t=this.transportRequestManager.addRequest(e,null),n=sr.getNotifications(e),this.frame&&(this.frame.postMessage(e.request,this.uri),this.transportRequestManager.settlePendingRequest(n)),[2,t]}))}))},t.prototype.close=function(){this.frame&&(window.removeEventListener("message",this.messageHandler),this.frame.close())},t}(Gt.Transport);nr.default=ar;var cr={},lr=T&&T.__extends||function(){var e=function(t,n){return(e=Object.setPrototypeOf||{__proto__:[]}instanceof Array&&function(e,t){e.__proto__=t}||function(e,t){for(var n in t)t.hasOwnProperty(n)&&(e[n]=t[n])})(t,n)};return function(t,n){function r(){this.constructor=t}e(t,n),t.prototype=null===n?Object.create(n):(r.prototype=n.prototype,new r)}}(),ur=T&&T.__awaiter||function(e,t,n,r){return new(n||(n=Promise))((function(o,i){function s(e){try{c(r.next(e))}catch(t){i(t)}}function a(e){try{c(r.throw(e))}catch(t){i(t)}}function c(e){var t;e.done?o(e.value):(t=e.value,t instanceof n?t:new n((function(e){e(t)}))).then(s,a)}c((r=r.apply(e,t||[])).next())}))},dr=T&&T.__generator||function(e,t){var n,r,o,i,s={label:0,sent:function(){if(1&o[0])throw o[1];return o[1]},trys:[],ops:[]};return i={next:a(0),throw:a(1),return:a(2)},"function"==typeof Symbol&&(i[Symbol.iterator]=function(){return this}),i;function a(i){return function(a){return function(i){if(n)throw new TypeError("Generator is already executing.");for(;s;)try{if(n=1,r&&(o=2&i[0]?r.return:i[0]?r.throw||((o=r.return)&&o.call(r),0):r.next)&&!(o=o.call(r,i[1])).done)return o;switch(r=0,o&&(i=[2&i[0],o.value]),i[0]){case 0:case 1:o=i;break;case 4:return s.label++,{value:i[1],done:!1};case 5:s.label++,r=i[1],i=[0];continue;case 7:i=s.ops.pop(),s.trys.pop();continue;default:if(!(o=s.trys,(o=o.length>0&&o[o.length-1])||6!==i[0]&&2!==i[0])){s=0;continue}if(3===i[0]&&(!o||i[1]>o[0]&&i[1]<o[3])){s.label=i[1];break}if(6===i[0]&&s.label<o[1]){s.label=o[1],o=i;break}if(o&&s.label<o[2]){s.label=o[2],s.ops.push(i);break}o[2]&&s.ops.pop(),s.trys.pop();continue}i=t.call(e,s)}catch(a){i=[6,a],r=0}finally{n=o=0}if(5&i[0])throw i[1];return{value:i[0]?i[1]:void 0,done:!0}}([i,a])}}};Object.defineProperty(cr,"__esModule",{value:!0});var fr=on,hr=function(e){function t(t){var n=e.call(this)||this;return n.messageHandler=function(e){n.transportRequestManager.resolveResponse(JSON.stringify(e.data))},n.uri=t,n.postMessageID="post-message-transport-"+Math.random(),n}return lr(t,e),t.prototype.createWindow=function(e){var t=this;return new Promise((function(n,r){var o,i=document.createElement("iframe");i.setAttribute("id",t.postMessageID),i.setAttribute("width","0px"),i.setAttribute("height","0px"),i.setAttribute("style","visiblity:hidden;border:none;outline:none;"),i.addEventListener("load",(function(){n(o)})),i.setAttribute("src",e),window.document.body.appendChild(i),o=i.contentWindow}))},t.prototype.connect=function(){var e=this,t=/^(http|https):\/\/.*$/;return new Promise((function(n,r){return ur(e,void 0,void 0,(function(){var e;return dr(this,(function(o){switch(o.label){case 0:return t.test(this.uri)||r(new Error("Bad URI")),e=this,[4,this.createWindow(this.uri)];case 1:return e.frame=o.sent(),window.addEventListener("message",this.messageHandler),n(),[2]}}))}))}))},t.prototype.sendData=function(e,t){return ur(this,void 0,void 0,(function(){var t,n;return dr(this,(function(r){return t=this.transportRequestManager.addRequest(e,null),n=fr.getNotifications(e),this.frame&&(this.frame.postMessage(e.request,"*"),this.transportRequestManager.settlePendingRequest(n)),[2,t]}))}))},t.prototype.close=function(){var e=document.getElementById(this.postMessageID);null==e||e.remove(),window.removeEventListener("message",this.messageHandler)},t}(Gt.Transport);cr.default=hr;var pr={},gr=T&&T.__awaiter||function(e,t,n,r){return new(n||(n=Promise))((function(o,i){function s(e){try{c(r.next(e))}catch(t){i(t)}}function a(e){try{c(r.throw(e))}catch(t){i(t)}}function c(e){var t;e.done?o(e.value):(t=e.value,t instanceof n?t:new n((function(e){e(t)}))).then(s,a)}c((r=r.apply(e,t||[])).next())}))},mr=T&&T.__generator||function(e,t){var n,r,o,i,s={label:0,sent:function(){if(1&o[0])throw o[1];return o[1]},trys:[],ops:[]};return i={next:a(0),throw:a(1),return:a(2)},"function"==typeof Symbol&&(i[Symbol.iterator]=function(){return this}),i;function a(i){return function(a){return function(i){if(n)throw new TypeError("Generator is already executing.");for(;s;)try{if(n=1,r&&(o=2&i[0]?r.return:i[0]?r.throw||((o=r.return)&&o.call(r),0):r.next)&&!(o=o.call(r,i[1])).done)return o;switch(r=0,o&&(i=[2&i[0],o.value]),i[0]){case 0:case 1:o=i;break;case 4:return s.label++,{value:i[1],done:!1};case 5:s.label++,r=i[1],i=[0];continue;case 7:i=s.ops.pop(),s.trys.pop();continue;default:if(!(o=s.trys,(o=o.length>0&&o[o.length-1])||6!==i[0]&&2!==i[0])){s=0;continue}if(3===i[0]&&(!o||i[1]>o[0]&&i[1]<o[3])){s.label=i[1];break}if(6===i[0]&&s.label<o[1]){s.label=o[1],o=i;break}if(o&&s.label<o[2]){s.label=o[2],s.ops.push(i);break}o[2]&&s.ops.pop(),s.trys.pop();continue}i=t.call(e,s)}catch(a){i=[6,a],r=0}finally{n=o=0}if(5&i[0])throw i[1];return{value:i[0]?i[1]:void 0,done:!0}}([i,a])}}};Object.defineProperty(pr,"__esModule",{value:!0});var vr=function(){function e(e){this.requestManager=e}return e.prototype.startBatch=function(){return this.requestManager.startBatch()},e.prototype.stopBatch=function(){return this.requestManager.stopBatch()},e.prototype.request=function(e,t){return gr(this,void 0,void 0,(function(){return mr(this,(function(n){switch(n.label){case 0:return this.requestManager.connectPromise?[4,this.requestManager.connectPromise]:[3,2];case 1:n.sent(),n.label=2;case 2:return[2,this.requestManager.request(e,!1,t)]}}))}))},e.prototype.notify=function(e){return gr(this,void 0,void 0,(function(){return mr(this,(function(t){switch(t.label){case 0:return this.requestManager.connectPromise?[4,this.requestManager.connectPromise]:[3,2];case 1:t.sent(),t.label=2;case 2:return[2,this.requestManager.request(e,!0,null)]}}))}))},e.prototype.onNotification=function(e){this.requestManager.requestChannel.addListener("notification",e)},e.prototype.onError=function(e){this.requestManager.requestChannel.addListener("error",e)},e.prototype.close=function(){this.requestManager.close()},e}();pr.default=vr,function(e){var t=T&&T.__importDefault||function(e){return e&&e.__esModule?e:{default:e}};Object.defineProperty(e,"__esModule",{value:!0}),e.PostMessageIframeTransport=e.PostMessageWindowTransport=e.JSONRPCError=e.WebSocketTransport=e.EventEmitterTransport=e.HTTPTransport=e.RequestManager=e.Client=void 0;var n=t(kt);e.RequestManager=n.default;var r=t(Kt);e.EventEmitterTransport=r.default;var o=t(un);e.HTTPTransport=o.default;var i=t(zn);e.WebSocketTransport=i.default;var s=t(nr);e.PostMessageWindowTransport=s.default;var a=t(cr);e.PostMessageIframeTransport=a.default;var c=Yt;Object.defineProperty(e,"JSONRPCError",{enumerable:!0,get:function(){return c.JSONRPCError}});var l=t(pr);e.Client=l.default,e.default=l.default}(Nt);var yr,br={},wr={},xr={},Tr={},Dr={};function Sr(){if(yr)return Dr;yr=1,Object.defineProperty(Dr,"__esModule",{value:!0}),Dr.Message=Dr.NotificationType9=Dr.NotificationType8=Dr.NotificationType7=Dr.NotificationType6=Dr.NotificationType5=Dr.NotificationType4=Dr.NotificationType3=Dr.NotificationType2=Dr.NotificationType1=Dr.NotificationType0=Dr.NotificationType=Dr.RequestType9=Dr.RequestType8=Dr.RequestType7=Dr.RequestType6=Dr.RequestType5=Dr.RequestType4=Dr.RequestType3=Dr.RequestType2=Dr.RequestType1=Dr.RequestType=Dr.RequestType0=Dr.AbstractMessageSignature=Dr.ParameterStructures=Dr.ResponseError=Dr.ErrorCodes=void 0;const e=Qe();var t,n,r,o;(n=t||(Dr.ErrorCodes=t={})).ParseError=-32700,n.InvalidRequest=-32600,n.MethodNotFound=-32601,n.InvalidParams=-32602,n.InternalError=-32603,n.jsonrpcReservedErrorRangeStart=-32099,n.serverErrorStart=-32099,n.MessageWriteError=-32099,n.MessageReadError=-32098,n.PendingResponseRejected=-32097,n.ConnectionInactive=-32096,n.ServerNotInitialized=-32002,n.UnknownErrorCode=-32001,n.jsonrpcReservedErrorRangeEnd=-32e3,n.serverErrorEnd=-32e3;class i extends Error{constructor(n,r,o){super(r),this.code=e.number(n)?n:t.UnknownErrorCode,this.data=o,Object.setPrototypeOf(this,i.prototype)}toJson(){const e={code:this.code,message:this.message};return void 0!==this.data&&(e.data=this.data),e}}Dr.ResponseError=i;class s{constructor(e){this.kind=e}static is(e){return e===s.auto||e===s.byName||e===s.byPosition}toString(){return this.kind}}Dr.ParameterStructures=s,s.auto=new s("auto"),s.byPosition=new s("byPosition"),s.byName=new s("byName");class a{constructor(e,t){this.method=e,this.numberOfParams=t}get parameterStructures(){return s.auto}}Dr.AbstractMessageSignature=a;Dr.RequestType0=class extends a{constructor(e){super(e,0)}};Dr.RequestType=class extends a{constructor(e,t=s.auto){super(e,1),this._parameterStructures=t}get parameterStructures(){return this._parameterStructures}};Dr.RequestType1=class extends a{constructor(e,t=s.auto){super(e,1),this._parameterStructures=t}get parameterStructures(){return this._parameterStructures}};Dr.RequestType2=class extends a{constructor(e){super(e,2)}};Dr.RequestType3=class extends a{constructor(e){super(e,3)}};Dr.RequestType4=class extends a{constructor(e){super(e,4)}};Dr.RequestType5=class extends a{constructor(e){super(e,5)}};Dr.RequestType6=class extends a{constructor(e){super(e,6)}};Dr.RequestType7=class extends a{constructor(e){super(e,7)}};Dr.RequestType8=class extends a{constructor(e){super(e,8)}};Dr.RequestType9=class extends a{constructor(e){super(e,9)}};Dr.NotificationType=class extends a{constructor(e,t=s.auto){super(e,1),this._parameterStructures=t}get parameterStructures(){return this._parameterStructures}};Dr.NotificationType0=class extends a{constructor(e){super(e,0)}};Dr.NotificationType1=class extends a{constructor(e,t=s.auto){super(e,1),this._parameterStructures=t}get parameterStructures(){return this._parameterStructures}};Dr.NotificationType2=class extends a{constructor(e){super(e,2)}};Dr.NotificationType3=class extends a{constructor(e){super(e,3)}};Dr.NotificationType4=class extends a{constructor(e){super(e,4)}};Dr.NotificationType5=class extends a{constructor(e){super(e,5)}};Dr.NotificationType6=class extends a{constructor(e){super(e,6)}};Dr.NotificationType7=class extends a{constructor(e){super(e,7)}};Dr.NotificationType8=class extends a{constructor(e){super(e,8)}};return Dr.NotificationType9=class extends a{constructor(e){super(e,9)}},(o=r||(Dr.Message=r={})).isRequest=function(t){const n=t;return n&&e.string(n.method)&&(e.string(n.id)||e.number(n.id))},o.isNotification=function(t){const n=t;return n&&e.string(n.method)&&void 0===t.id},o.isResponse=function(t){const n=t;return n&&(void 0!==n.result||!!n.error)&&(e.string(n.id)||e.number(n.id)||null===n.id)},Dr}var Or,Rr={};function jr(){if(Or)return Rr;var e,t,n;Or=1,Object.defineProperty(Rr,"__esModule",{value:!0}),Rr.LRUCache=Rr.LinkedMap=Rr.Touch=void 0,(n=t||(Rr.Touch=t={})).None=0,n.First=1,n.AsOld=n.First,n.Last=2,n.AsNew=n.Last;class r{constructor(){this[e]="LinkedMap",this._map=new Map,this._head=void 0,this._tail=void 0,this._size=0,this._state=0}clear(){this._map.clear(),this._head=void 0,this._tail=void 0,this._size=0,this._state++}isEmpty(){return!this._head&&!this._tail}get size(){return this._size}get first(){var e;return null==(e=this._head)?void 0:e.value}get last(){var e;return null==(e=this._tail)?void 0:e.value}has(e){return this._map.has(e)}get(e,n=t.None){const r=this._map.get(e);if(r)return n!==t.None&&this.touch(r,n),r.value}set(e,n,r=t.None){let o=this._map.get(e);if(o)o.value=n,r!==t.None&&this.touch(o,r);else{switch(o={key:e,value:n,next:void 0,previous:void 0},r){case t.None:this.addItemLast(o);break;case t.First:this.addItemFirst(o);break;case t.Last:default:this.addItemLast(o)}this._map.set(e,o),this._size++}return this}delete(e){return!!this.remove(e)}remove(e){const t=this._map.get(e);if(t)return this._map.delete(e),this.removeItem(t),this._size--,t.value}shift(){if(!this._head&&!this._tail)return;if(!this._head||!this._tail)throw new Error("Invalid list");const e=this._head;return this._map.delete(e.key),this.removeItem(e),this._size--,e.value}forEach(e,t){const n=this._state;let r=this._head;for(;r;){if(t?e.bind(t)(r.value,r.key,this):e(r.value,r.key,this),this._state!==n)throw new Error("LinkedMap got modified during iteration.");r=r.next}}keys(){const e=this._state;let t=this._head;const n={[Symbol.iterator]:()=>n,next:()=>{if(this._state!==e)throw new Error("LinkedMap got modified during iteration.");if(t){const e={value:t.key,done:!1};return t=t.next,e}return{value:void 0,done:!0}}};return n}values(){const e=this._state;let t=this._head;const n={[Symbol.iterator]:()=>n,next:()=>{if(this._state!==e)throw new Error("LinkedMap got modified during iteration.");if(t){const e={value:t.value,done:!1};return t=t.next,e}return{value:void 0,done:!0}}};return n}entries(){const e=this._state;let t=this._head;const n={[Symbol.iterator]:()=>n,next:()=>{if(this._state!==e)throw new Error("LinkedMap got modified during iteration.");if(t){const e={value:[t.key,t.value],done:!1};return t=t.next,e}return{value:void 0,done:!0}}};return n}[(e=Symbol.toStringTag,Symbol.iterator)](){return this.entries()}trimOld(e){if(e>=this.size)return;if(0===e)return void this.clear();let t=this._head,n=this.size;for(;t&&n>e;)this._map.delete(t.key),t=t.next,n--;this._head=t,this._size=n,t&&(t.previous=void 0),this._state++}addItemFirst(e){if(this._head||this._tail){if(!this._head)throw new Error("Invalid list");e.next=this._head,this._head.previous=e}else this._tail=e;this._head=e,this._state++}addItemLast(e){if(this._head||this._tail){if(!this._tail)throw new Error("Invalid list");e.previous=this._tail,this._tail.next=e}else this._head=e;this._tail=e,this._state++}removeItem(e){if(e===this._head&&e===this._tail)this._head=void 0,this._tail=void 0;else if(e===this._head){if(!e.next)throw new Error("Invalid list");e.next.previous=void 0,this._head=e.next}else if(e===this._tail){if(!e.previous)throw new Error("Invalid list");e.previous.next=void 0,this._tail=e.previous}else{const t=e.next,n=e.previous;if(!t||!n)throw new Error("Invalid list");t.previous=n,n.next=t}e.next=void 0,e.previous=void 0,this._state++}touch(e,n){if(!this._head||!this._tail)throw new Error("Invalid list");if(n===t.First||n===t.Last)if(n===t.First){if(e===this._head)return;const t=e.next,n=e.previous;e===this._tail?(n.next=void 0,this._tail=n):(t.previous=n,n.next=t),e.previous=void 0,e.next=this._head,this._head.previous=e,this._head=e,this._state++}else if(n===t.Last){if(e===this._tail)return;const t=e.next,n=e.previous;e===this._head?(t.previous=void 0,this._head=t):(t.previous=n,n.next=t),e.next=void 0,e.previous=this._tail,this._tail.next=e,this._tail=e,this._state++}}toJSON(){const e=[];return this.forEach(((t,n)=>{e.push([n,t])})),e}fromJSON(e){this.clear();for(const[t,n]of e)this.set(t,n)}}Rr.LinkedMap=r;return Rr.LRUCache=class extends r{constructor(e,t=1){super(),this._limit=e,this._ratio=Math.min(Math.max(0,t),1)}get limit(){return this._limit}set limit(e){this._limit=e,this.checkTrim()}get ratio(){return this._ratio}set ratio(e){this._ratio=Math.min(Math.max(0,e),1),this.checkTrim()}get(e,n=t.AsNew){return super.get(e,n)}peek(e){return super.get(e,t.None)}set(e,n){return super.set(e,n,t.Last),this.checkTrim(),this}checkTrim(){this.size>this._limit&&this.trimOld(Math.round(this._limit*this._ratio))}},Rr}var Cr,Nr={};var kr,Pr={};var Er,Ir,_r={},Mr={};function qr(){if(Er)return Mr;Er=1,Object.defineProperty(Mr,"__esModule",{value:!0}),Mr.Semaphore=void 0;const e=tt();return Mr.Semaphore=class{constructor(e=1){if(e<=0)throw new Error("Capacity must be greater than 0");this._capacity=e,this._active=0,this._waiting=[]}lock(e){return new Promise(((t,n)=>{this._waiting.push({thunk:e,resolve:t,reject:n}),this.runNext()}))}get active(){return this._active}runNext(){0!==this._waiting.length&&this._active!==this._capacity&&(0,e.default)().timer.setImmediate((()=>this.doRunNext()))}doRunNext(){if(0===this._waiting.length||this._active===this._capacity)return;const e=this._waiting.shift();if(this._active++,this._active>this._capacity)throw new Error("To many thunks active");try{const t=e.thunk();t instanceof Promise?t.then((t=>{this._active--,e.resolve(t),this.runNext()}),(t=>{this._active--,e.reject(t),this.runNext()})):(this._active--,e.resolve(t),this.runNext())}catch(Ih){this._active--,e.reject(Ih),this.runNext()}}},Mr}var Ar,Lr={};var Fr,Hr={};var Wr,Ur,Br={};function zr(){return Wr||(Wr=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.createMessageConnection=e.ConnectionOptions=e.MessageStrategy=e.CancellationStrategy=e.CancellationSenderStrategy=e.CancellationReceiverStrategy=e.RequestCancellationReceiverStrategy=e.IdCancellationReceiverStrategy=e.ConnectionStrategy=e.ConnectionError=e.ConnectionErrors=e.LogTraceNotification=e.SetTraceNotification=e.TraceFormat=e.TraceValues=e.Trace=e.NullLogger=e.ProgressType=e.ProgressToken=void 0;const t=tt(),n=Qe(),r=Sr(),o=jr(),i=nt(),s=et();var a,c,l,u,d,f,h,p,g,m,v,y,b,w,x,T,D,S,O,R,j,C,N,k,P,E,I;(a||(a={})).type=new r.NotificationType("$/cancelRequest"),(c||(e.ProgressToken=c={})).is=function(e){return"string"==typeof e||"number"==typeof e},(l||(l={})).type=new r.NotificationType("$/progress");e.ProgressType=class{constructor(){}},(u||(u={})).is=function(e){return n.func(e)},e.NullLogger=Object.freeze({error:()=>{},warn:()=>{},info:()=>{},log:()=>{}}),(f=d||(e.Trace=d={}))[f.Off=0]="Off",f[f.Messages=1]="Messages",f[f.Compact=2]="Compact",f[f.Verbose=3]="Verbose",(p=h||(e.TraceValues=h={})).Off="off",p.Messages="messages",p.Compact="compact",p.Verbose="verbose",function(e){e.fromString=function(t){if(!n.string(t))return e.Off;switch(t=t.toLowerCase()){case"off":default:return e.Off;case"messages":return e.Messages;case"compact":return e.Compact;case"verbose":return e.Verbose}},e.toString=function(t){switch(t){case e.Off:return"off";case e.Messages:return"messages";case e.Compact:return"compact";case e.Verbose:return"verbose";default:return"off"}}}(d||(e.Trace=d={})),(m=g||(e.TraceFormat=g={})).Text="text",m.JSON="json",function(e){e.fromString=function(t){return n.string(t)&&"json"===(t=t.toLowerCase())?e.JSON:e.Text}}(g||(e.TraceFormat=g={})),(v||(e.SetTraceNotification=v={})).type=new r.NotificationType("$/setTrace"),(y||(e.LogTraceNotification=y={})).type=new r.NotificationType("$/logTrace"),(w=b||(e.ConnectionErrors=b={}))[w.Closed=1]="Closed",w[w.Disposed=2]="Disposed",w[w.AlreadyListening=3]="AlreadyListening";class _ extends Error{constructor(e,t){super(t),this.code=e,Object.setPrototypeOf(this,_.prototype)}}e.ConnectionError=_,(x||(e.ConnectionStrategy=x={})).is=function(e){const t=e;return t&&n.func(t.cancelUndispatched)},(T||(e.IdCancellationReceiverStrategy=T={})).is=function(e){const t=e;return t&&(void 0===t.kind||"id"===t.kind)&&n.func(t.createCancellationTokenSource)&&(void 0===t.dispose||n.func(t.dispose))},(D||(e.RequestCancellationReceiverStrategy=D={})).is=function(e){const t=e;return t&&"request"===t.kind&&n.func(t.createCancellationTokenSource)&&(void 0===t.dispose||n.func(t.dispose))},(O=S||(e.CancellationReceiverStrategy=S={})).Message=Object.freeze({createCancellationTokenSource:e=>new s.CancellationTokenSource}),O.is=function(e){return T.is(e)||D.is(e)},(j=R||(e.CancellationSenderStrategy=R={})).Message=Object.freeze({sendCancellation:(e,t)=>e.sendNotification(a.type,{id:t}),cleanup(e){}}),j.is=function(e){const t=e;return t&&n.func(t.sendCancellation)&&n.func(t.cleanup)},(N=C||(e.CancellationStrategy=C={})).Message=Object.freeze({receiver:S.Message,sender:R.Message}),N.is=function(e){const t=e;return t&&S.is(t.receiver)&&R.is(t.sender)},(k||(e.MessageStrategy=k={})).is=function(e){const t=e;return t&&n.func(t.handleMessage)},(P||(e.ConnectionOptions=P={})).is=function(e){const t=e;return t&&(C.is(t.cancellationStrategy)||x.is(t.connectionStrategy)||k.is(t.messageStrategy))},(I=E||(E={}))[I.New=1]="New",I[I.Listening=2]="Listening",I[I.Closed=3]="Closed",I[I.Disposed=4]="Disposed",e.createMessageConnection=function(f,h,p,m){const w=void 0!==p?p:e.NullLogger;let x=0,D=0,S=0;const O="2.0";let R;const j=new Map;let N;const P=new Map,I=new Map;let M,q,A=new o.LinkedMap,L=new Map,F=new Set,H=new Map,W=d.Off,U=g.Text,B=E.New;const z=new i.Emitter,$=new i.Emitter,V=new i.Emitter,K=new i.Emitter,G=new i.Emitter,J=m&&m.cancellationStrategy?m.cancellationStrategy:C.Message;function Y(e){if(null===e)throw new Error("Can't send requests with id null since the response can't be correlated.");return"req-"+e.toString()}function X(e,t){var n;r.Message.isRequest(t)?e.set(Y(t.id),t):r.Message.isResponse(t)?e.set(null===(n=t.id)?"res-unknown-"+(++S).toString():"res-"+n.toString(),t):e.set("not-"+(++D).toString(),t)}function Z(e){}function Q(){return B===E.Listening}function ee(){return B===E.Closed}function te(){return B===E.Disposed}function ne(){B!==E.New&&B!==E.Listening||(B=E.Closed,$.fire(void 0))}function re(){M||0===A.size||(M=(0,t.default)().timer.setImmediate((()=>{M=void 0,function(){if(0===A.size)return;const e=A.shift();try{const t=null==m?void 0:m.messageStrategy;k.is(t)?t.handleMessage(e,oe):oe(e)}finally{re()}}()})))}function oe(e){r.Message.isRequest(e)?function(e){if(te())return;function t(t,n,o){const i={jsonrpc:O,id:e.id};t instanceof r.ResponseError?i.error=t.toJson():i.result=void 0===t?null:t,ae(i,n,o),h.write(i).catch((()=>w.error("Sending response failed.")))}function o(t,n,r){const o={jsonrpc:O,id:e.id,error:t.toJson()};ae(o,n,r),h.write(o).catch((()=>w.error("Sending response failed.")))}function i(t,n,r){void 0===t&&(t=null);const o={jsonrpc:O,id:e.id,result:t};ae(o,n,r),h.write(o).catch((()=>w.error("Sending response failed.")))}!function(e){if(W===d.Off||!q)return;if(U===g.Text){let t;W!==d.Verbose&&W!==d.Compact||!e.params||(t=`Params: ${se(e.params)}\n\n`),q.log(`Received request '${e.method} - (${e.id})'.`,t)}else le("receive-request",e)}(e);const s=j.get(e.method);let a,c;s&&(a=s.type,c=s.handler);const l=Date.now();if(c||R){const s=e.id??String(Date.now()),d=T.is(J.receiver)?J.receiver.createCancellationTokenSource(s):J.receiver.createCancellationTokenSource(e);null!==e.id&&F.has(e.id)&&d.cancel(),null!==e.id&&H.set(s,d);try{let u;if(c)if(void 0===e.params){if(void 0!==a&&0!==a.numberOfParams)return void o(new r.ResponseError(r.ErrorCodes.InvalidParams,`Request ${e.method} defines ${a.numberOfParams} params but received none.`),e.method,l);u=c(d.token)}else if(Array.isArray(e.params)){if(void 0!==a&&a.parameterStructures===r.ParameterStructures.byName)return void o(new r.ResponseError(r.ErrorCodes.InvalidParams,`Request ${e.method} defines parameters by name but received parameters by position`),e.method,l);u=c(...e.params,d.token)}else{if(void 0!==a&&a.parameterStructures===r.ParameterStructures.byPosition)return void o(new r.ResponseError(r.ErrorCodes.InvalidParams,`Request ${e.method} defines parameters by position but received parameters by name`),e.method,l);u=c(e.params,d.token)}else R&&(u=R(e.method,e.params,d.token));const f=u;u?f.then?f.then((n=>{H.delete(s),t(n,e.method,l)}),(t=>{H.delete(s),t instanceof r.ResponseError?o(t,e.method,l):t&&n.string(t.message)?o(new r.ResponseError(r.ErrorCodes.InternalError,`Request ${e.method} failed with message: ${t.message}`),e.method,l):o(new r.ResponseError(r.ErrorCodes.InternalError,`Request ${e.method} failed unexpectedly without providing any details.`),e.method,l)})):(H.delete(s),t(u,e.method,l)):(H.delete(s),i(u,e.method,l))}catch(u){H.delete(s),u instanceof r.ResponseError?t(u,e.method,l):u&&n.string(u.message)?o(new r.ResponseError(r.ErrorCodes.InternalError,`Request ${e.method} failed with message: ${u.message}`),e.method,l):o(new r.ResponseError(r.ErrorCodes.InternalError,`Request ${e.method} failed unexpectedly without providing any details.`),e.method,l)}}else o(new r.ResponseError(r.ErrorCodes.MethodNotFound,`Unhandled method ${e.method}`),e.method,l)}(e):r.Message.isNotification(e)?function(e){if(te())return;let t,n;if(e.method===a.type.method){const t=e.params.id;return F.delete(t),void ce(e)}{const r=P.get(e.method);r&&(n=r.handler,t=r.type)}if(n||N)try{if(ce(e),n)if(void 0===e.params)void 0!==t&&0!==t.numberOfParams&&t.parameterStructures!==r.ParameterStructures.byName&&w.error(`Notification ${e.method} defines ${t.numberOfParams} params but received none.`),n();else if(Array.isArray(e.params)){const o=e.params;e.method===l.type.method&&2===o.length&&c.is(o[0])?n({token:o[0],value:o[1]}):(void 0!==t&&(t.parameterStructures===r.ParameterStructures.byName&&w.error(`Notification ${e.method} defines parameters by name but received parameters by position`),t.numberOfParams!==e.params.length&&w.error(`Notification ${e.method} defines ${t.numberOfParams} params but received ${o.length} arguments`)),n(...o))}else void 0!==t&&t.parameterStructures===r.ParameterStructures.byPosition&&w.error(`Notification ${e.method} defines parameters by position but received parameters by name`),n(e.params);else N&&N(e.method,e.params)}catch(o){o.message?w.error(`Notification handler '${e.method}' failed with message: ${o.message}`):w.error(`Notification handler '${e.method}' failed unexpectedly.`)}else V.fire(e)}(e):r.Message.isResponse(e)?function(e){if(te())return;if(null===e.id)e.error?w.error(`Received response message without id: Error is: \n${JSON.stringify(e.error,void 0,4)}`):w.error("Received response message without id. No further error information provided.");else{const n=e.id,o=L.get(n);if(function(e,t){if(W===d.Off||!q)return;if(U===g.Text){let n;if(W!==d.Verbose&&W!==d.Compact||(e.error&&e.error.data?n=`Error data: ${se(e.error.data)}\n\n`:e.result?n=`Result: ${se(e.result)}\n\n`:void 0===e.error&&(n="No result returned.\n\n")),t){const r=e.error?` Request failed: ${e.error.message} (${e.error.code}).`:"";q.log(`Received response '${t.method} - (${e.id})' in ${Date.now()-t.timerStart}ms.${r}`,n)}else q.log(`Received response ${e.id} without active response promise.`,n)}else le("receive-response",e)}(e,o),void 0!==o){L.delete(n);try{if(e.error){const t=e.error;o.reject(new r.ResponseError(t.code,t.message,t.data))}else{if(void 0===e.result)throw new Error("Should never happen.");o.resolve(e.result)}}catch(t){t.message?w.error(`Response handler '${o.method}' failed with message: ${t.message}`):w.error(`Response handler '${o.method}' failed unexpectedly.`)}}}}(e):function(e){if(!e)return void w.error("Received empty message.");w.error(`Received message which is neither a response nor a notification message:\n${JSON.stringify(e,null,4)}`);const t=e;if(n.string(t.id)||n.number(t.id)){const e=t.id,n=L.get(e);n&&n.reject(new Error("The received response has neither a result nor an error property."))}}(e)}f.onClose(ne),f.onError((function(e){z.fire([e,void 0,void 0])})),h.onClose(ne),h.onError((function(e){z.fire(e)}));const ie=e=>{try{if(r.Message.isNotification(e)&&e.method===a.type.method){const t=e.params.id,n=Y(t),o=A.get(n);if(r.Message.isRequest(o)){const r=null==m?void 0:m.connectionStrategy,i=r&&r.cancelUndispatched?r.cancelUndispatched(o,Z):void 0;if(i&&(void 0!==i.error||void 0!==i.result))return A.delete(n),H.delete(t),i.id=o.id,ae(i,e.method,Date.now()),void h.write(i).catch((()=>w.error("Sending response for canceled message failed.")))}const i=H.get(t);if(void 0!==i)return i.cancel(),void ce(e);F.add(t)}X(A,e)}finally{re()}};function se(e){if(null!=e)switch(W){case d.Verbose:return JSON.stringify(e,null,4);case d.Compact:return JSON.stringify(e);default:return}}function ae(e,t,n){if(W!==d.Off&&q)if(U===g.Text){let r;W!==d.Verbose&&W!==d.Compact||(e.error&&e.error.data?r=`Error data: ${se(e.error.data)}\n\n`:e.result?r=`Result: ${se(e.result)}\n\n`:void 0===e.error&&(r="No result returned.\n\n")),q.log(`Sending response '${t} - (${e.id})'. Processing request took ${Date.now()-n}ms`,r)}else le("send-response",e)}function ce(e){if(W!==d.Off&&q&&e.method!==y.type.method)if(U===g.Text){let t;W!==d.Verbose&&W!==d.Compact||(t=e.params?`Params: ${se(e.params)}\n\n`:"No parameters provided.\n\n"),q.log(`Received notification '${e.method}'.`,t)}else le("receive-notification",e)}function le(e,t){if(!q||W===d.Off)return;const n={isLSPMessage:!0,type:e,message:t,timestamp:Date.now()};q.log(n)}function ue(){if(ee())throw new _(b.Closed,"Connection is closed.");if(te())throw new _(b.Disposed,"Connection is disposed.")}function de(e){return void 0===e?null:e}function fe(e){return null===e?void 0:e}function he(e){return null!=e&&!Array.isArray(e)&&"object"==typeof e}function pe(e,t){switch(e){case r.ParameterStructures.auto:return he(t)?fe(t):[de(t)];case r.ParameterStructures.byName:if(!he(t))throw new Error("Received parameters by name but param is not an object literal.");return fe(t);case r.ParameterStructures.byPosition:return[de(t)];default:throw new Error(`Unknown parameter structure ${e.toString()}`)}}function ge(e,t){let n;const r=e.numberOfParams;switch(r){case 0:n=void 0;break;case 1:n=pe(e.parameterStructures,t[0]);break;default:n=[];for(let e=0;e<t.length&&e<r;e++)n.push(de(t[e]));if(t.length<r)for(let e=t.length;e<r;e++)n.push(null)}return n}const me={sendNotification:(e,...t)=>{let o,i;if(ue(),n.string(e)){o=e;const n=t[0];let s=0,a=r.ParameterStructures.auto;r.ParameterStructures.is(n)&&(s=1,a=n);let c=t.length;const l=c-s;switch(l){case 0:i=void 0;break;case 1:i=pe(a,t[s]);break;default:if(a===r.ParameterStructures.byName)throw new Error(`Received ${l} parameters for 'by Name' notification parameter structure.`);i=t.slice(s,c).map((e=>de(e)))}}else{const n=t;o=e.method,i=ge(e,n)}const s={jsonrpc:O,method:o,params:i};return function(e){if(W!==d.Off&&q)if(U===g.Text){let t;W!==d.Verbose&&W!==d.Compact||(t=e.params?`Params: ${se(e.params)}\n\n`:"No parameters provided.\n\n"),q.log(`Sending notification '${e.method}'.`,t)}else le("send-notification",e)}(s),h.write(s).catch((e=>{throw w.error("Sending notification failed."),e}))},onNotification:(e,t)=>{let r;return ue(),n.func(e)?N=e:t&&(n.string(e)?(r=e,P.set(e,{type:void 0,handler:t})):(r=e.method,P.set(e.method,{type:e,handler:t}))),{dispose:()=>{void 0!==r?P.delete(r):N=void 0}}},onProgress:(e,t,n)=>{if(I.has(t))throw new Error(`Progress handler for token ${t} already registered`);return I.set(t,n),{dispose:()=>{I.delete(t)}}},sendProgress:(e,t,n)=>me.sendNotification(l.type,{token:t,value:n}),onUnhandledProgress:K.event,sendRequest:(e,...t)=>{let o,i,a;if(ue(),function(){if(!Q())throw new Error("Call listen() first.")}(),n.string(e)){o=e;const n=t[0],c=t[t.length-1];let l=0,u=r.ParameterStructures.auto;r.ParameterStructures.is(n)&&(l=1,u=n);let d=t.length;s.CancellationToken.is(c)&&(d-=1,a=c);const f=d-l;switch(f){case 0:i=void 0;break;case 1:i=pe(u,t[l]);break;default:if(u===r.ParameterStructures.byName)throw new Error(`Received ${f} parameters for 'by Name' request parameter structure.`);i=t.slice(l,d).map((e=>de(e)))}}else{const n=t;o=e.method,i=ge(e,n);const r=e.numberOfParams;a=s.CancellationToken.is(n[r])?n[r]:void 0}const c=x++;let l;a&&(l=a.onCancellationRequested((()=>{const e=J.sender.sendCancellation(me,c);return void 0===e?(w.log(`Received no promise from cancellation strategy when cancelling id ${c}`),Promise.resolve()):e.catch((()=>{w.log(`Sending cancellation messages for id ${c} failed`)}))})));const u={jsonrpc:O,id:c,method:o,params:i};return function(e){if(W!==d.Off&&q)if(U===g.Text){let t;W!==d.Verbose&&W!==d.Compact||!e.params||(t=`Params: ${se(e.params)}\n\n`),q.log(`Sending request '${e.method} - (${e.id})'.`,t)}else le("send-request",e)}(u),"function"==typeof J.sender.enableCancellation&&J.sender.enableCancellation(u),new Promise((async(e,t)=>{const n={method:o,timerStart:Date.now(),resolve:t=>{e(t),J.sender.cleanup(c),null==l||l.dispose()},reject:e=>{t(e),J.sender.cleanup(c),null==l||l.dispose()}};try{await h.write(u),L.set(c,n)}catch(i){throw w.error("Sending request failed."),n.reject(new r.ResponseError(r.ErrorCodes.MessageWriteError,i.message?i.message:"Unknown reason")),i}}))},onRequest:(e,t)=>{ue();let r=null;return u.is(e)?(r=void 0,R=e):n.string(e)?(r=null,void 0!==t&&(r=e,j.set(e,{handler:t,type:void 0}))):void 0!==t&&(r=e.method,j.set(e.method,{type:e,handler:t})),{dispose:()=>{null!==r&&(void 0!==r?j.delete(r):R=void 0)}}},hasPendingResponse:()=>L.size>0,trace:async(e,t,r)=>{let o=!1,i=g.Text;void 0!==r&&(n.boolean(r)?o=r:(o=r.sendNotification||!1,i=r.traceFormat||g.Text)),W=e,U=i,q=W===d.Off?void 0:t,!o||ee()||te()||await me.sendNotification(v.type,{value:d.toString(e)})},onError:z.event,onClose:$.event,onUnhandledNotification:V.event,onDispose:G.event,end:()=>{h.end()},dispose:()=>{if(te())return;B=E.Disposed,G.fire(void 0);const e=new r.ResponseError(r.ErrorCodes.PendingResponseRejected,"Pending response rejected since connection got disposed");for(const t of L.values())t.reject(e);L=new Map,H=new Map,F=new Set,A=new o.LinkedMap,n.func(h.dispose)&&h.dispose(),n.func(f.dispose)&&f.dispose()},listen:()=>{ue(),function(){if(Q())throw new _(b.AlreadyListening,"Connection is already listening")}(),B=E.Listening,f.listen(ie)},inspect:()=>{(0,t.default)().console.log("inspect")}};return me.onNotification(y.type,(e=>{if(W===d.Off||!q)return;const t=W===d.Verbose||W===d.Compact;q.log(e.message,t?e.verbose:void 0)})),me.onNotification(l.type,(e=>{const t=I.get(e.token);t?t(e.value):K.fire(e)})),me}}(Br)),Br}function $r(){return Ur||(Ur=1,function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.ProgressType=e.ProgressToken=e.createMessageConnection=e.NullLogger=e.ConnectionOptions=e.ConnectionStrategy=e.AbstractMessageBuffer=e.WriteableStreamMessageWriter=e.AbstractMessageWriter=e.MessageWriter=e.ReadableStreamMessageReader=e.AbstractMessageReader=e.MessageReader=e.SharedArrayReceiverStrategy=e.SharedArraySenderStrategy=e.CancellationToken=e.CancellationTokenSource=e.Emitter=e.Event=e.Disposable=e.LRUCache=e.Touch=e.LinkedMap=e.ParameterStructures=e.NotificationType9=e.NotificationType8=e.NotificationType7=e.NotificationType6=e.NotificationType5=e.NotificationType4=e.NotificationType3=e.NotificationType2=e.NotificationType1=e.NotificationType0=e.NotificationType=e.ErrorCodes=e.ResponseError=e.RequestType9=e.RequestType8=e.RequestType7=e.RequestType6=e.RequestType5=e.RequestType4=e.RequestType3=e.RequestType2=e.RequestType1=e.RequestType0=e.RequestType=e.Message=e.RAL=void 0,e.MessageStrategy=e.CancellationStrategy=e.CancellationSenderStrategy=e.CancellationReceiverStrategy=e.ConnectionError=e.ConnectionErrors=e.LogTraceNotification=e.SetTraceNotification=e.TraceFormat=e.TraceValues=e.Trace=void 0;const t=Sr();Object.defineProperty(e,"Message",{enumerable:!0,get:function(){return t.Message}}),Object.defineProperty(e,"RequestType",{enumerable:!0,get:function(){return t.RequestType}}),Object.defineProperty(e,"RequestType0",{enumerable:!0,get:function(){return t.RequestType0}}),Object.defineProperty(e,"RequestType1",{enumerable:!0,get:function(){return t.RequestType1}}),Object.defineProperty(e,"RequestType2",{enumerable:!0,get:function(){return t.RequestType2}}),Object.defineProperty(e,"RequestType3",{enumerable:!0,get:function(){return t.RequestType3}}),Object.defineProperty(e,"RequestType4",{enumerable:!0,get:function(){return t.RequestType4}}),Object.defineProperty(e,"RequestType5",{enumerable:!0,get:function(){return t.RequestType5}}),Object.defineProperty(e,"RequestType6",{enumerable:!0,get:function(){return t.RequestType6}}),Object.defineProperty(e,"RequestType7",{enumerable:!0,get:function(){return t.RequestType7}}),Object.defineProperty(e,"RequestType8",{enumerable:!0,get:function(){return t.RequestType8}}),Object.defineProperty(e,"RequestType9",{enumerable:!0,get:function(){return t.RequestType9}}),Object.defineProperty(e,"ResponseError",{enumerable:!0,get:function(){return t.ResponseError}}),Object.defineProperty(e,"ErrorCodes",{enumerable:!0,get:function(){return t.ErrorCodes}}),Object.defineProperty(e,"NotificationType",{enumerable:!0,get:function(){return t.NotificationType}}),Object.defineProperty(e,"NotificationType0",{enumerable:!0,get:function(){return t.NotificationType0}}),Object.defineProperty(e,"NotificationType1",{enumerable:!0,get:function(){return t.NotificationType1}}),Object.defineProperty(e,"NotificationType2",{enumerable:!0,get:function(){return t.NotificationType2}}),Object.defineProperty(e,"NotificationType3",{enumerable:!0,get:function(){return t.NotificationType3}}),Object.defineProperty(e,"NotificationType4",{enumerable:!0,get:function(){return t.NotificationType4}}),Object.defineProperty(e,"NotificationType5",{enumerable:!0,get:function(){return t.NotificationType5}}),Object.defineProperty(e,"NotificationType6",{enumerable:!0,get:function(){return t.NotificationType6}}),Object.defineProperty(e,"NotificationType7",{enumerable:!0,get:function(){return t.NotificationType7}}),Object.defineProperty(e,"NotificationType8",{enumerable:!0,get:function(){return t.NotificationType8}}),Object.defineProperty(e,"NotificationType9",{enumerable:!0,get:function(){return t.NotificationType9}}),Object.defineProperty(e,"ParameterStructures",{enumerable:!0,get:function(){return t.ParameterStructures}});const n=jr();Object.defineProperty(e,"LinkedMap",{enumerable:!0,get:function(){return n.LinkedMap}}),Object.defineProperty(e,"LRUCache",{enumerable:!0,get:function(){return n.LRUCache}}),Object.defineProperty(e,"Touch",{enumerable:!0,get:function(){return n.Touch}});const r=(Cr||(Cr=1,Object.defineProperty(Nr,"__esModule",{value:!0}),Nr.Disposable=void 0,(o||(Nr.Disposable=o={})).create=function(e){return{dispose:e}}),Nr);var o;Object.defineProperty(e,"Disposable",{enumerable:!0,get:function(){return r.Disposable}});const i=nt();Object.defineProperty(e,"Event",{enumerable:!0,get:function(){return i.Event}}),Object.defineProperty(e,"Emitter",{enumerable:!0,get:function(){return i.Emitter}});const s=et();Object.defineProperty(e,"CancellationTokenSource",{enumerable:!0,get:function(){return s.CancellationTokenSource}}),Object.defineProperty(e,"CancellationToken",{enumerable:!0,get:function(){return s.CancellationToken}});const a=function(){if(kr)return Pr;kr=1,Object.defineProperty(Pr,"__esModule",{value:!0}),Pr.SharedArrayReceiverStrategy=Pr.SharedArraySenderStrategy=void 0;const e=et();var t,n;(n=t||(t={})).Continue=0,n.Cancelled=1,Pr.SharedArraySenderStrategy=class{constructor(){this.buffers=new Map}enableCancellation(e){if(null===e.id)return;const n=new SharedArrayBuffer(4);new Int32Array(n,0,1)[0]=t.Continue,this.buffers.set(e.id,n),e.$cancellationData=n}async sendCancellation(e,n){const r=this.buffers.get(n);if(void 0===r)return;const o=new Int32Array(r,0,1);Atomics.store(o,0,t.Cancelled)}cleanup(e){this.buffers.delete(e)}dispose(){this.buffers.clear()}};class r{constructor(e){this.data=new Int32Array(e,0,1)}get isCancellationRequested(){return Atomics.load(this.data,0)===t.Cancelled}get onCancellationRequested(){throw new Error("Cancellation over SharedArrayBuffer doesn't support cancellation events")}}class o{constructor(e){this.token=new r(e)}cancel(){}dispose(){}}return Pr.SharedArrayReceiverStrategy=class{constructor(){this.kind="request"}createCancellationTokenSource(t){const n=t.$cancellationData;return void 0===n?new e.CancellationTokenSource:new o(n)}},Pr}();Object.defineProperty(e,"SharedArraySenderStrategy",{enumerable:!0,get:function(){return a.SharedArraySenderStrategy}}),Object.defineProperty(e,"SharedArrayReceiverStrategy",{enumerable:!0,get:function(){return a.SharedArrayReceiverStrategy}});const c=function(){if(Ir)return _r;Ir=1,Object.defineProperty(_r,"__esModule",{value:!0}),_r.ReadableStreamMessageReader=_r.AbstractMessageReader=_r.MessageReader=void 0;const e=tt(),t=Qe(),n=nt(),r=qr();var o,i;(o||(_r.MessageReader=o={})).is=function(e){let n=e;return n&&t.func(n.listen)&&t.func(n.dispose)&&t.func(n.onError)&&t.func(n.onClose)&&t.func(n.onPartialMessage)};class s{constructor(){this.errorEmitter=new n.Emitter,this.closeEmitter=new n.Emitter,this.partialMessageEmitter=new n.Emitter}dispose(){this.errorEmitter.dispose(),this.closeEmitter.dispose()}get onError(){return this.errorEmitter.event}fireError(e){this.errorEmitter.fire(this.asError(e))}get onClose(){return this.closeEmitter.event}fireClose(){this.closeEmitter.fire(void 0)}get onPartialMessage(){return this.partialMessageEmitter.event}firePartialMessage(e){this.partialMessageEmitter.fire(e)}asError(e){return e instanceof Error?e:new Error(`Reader received error. Reason: ${t.string(e.message)?e.message:"unknown"}`)}}return _r.AbstractMessageReader=s,(i||(i={})).fromOptions=function(t){let n,r;const o=new Map;let i;const s=new Map;if(void 0===t||"string"==typeof t)n=t??"utf-8";else{if(n=t.charset??"utf-8",void 0!==t.contentDecoder&&(r=t.contentDecoder,o.set(r.name,r)),void 0!==t.contentDecoders)for(const e of t.contentDecoders)o.set(e.name,e);if(void 0!==t.contentTypeDecoder&&(i=t.contentTypeDecoder,s.set(i.name,i)),void 0!==t.contentTypeDecoders)for(const e of t.contentTypeDecoders)s.set(e.name,e)}return void 0===i&&(i=(0,e.default)().applicationJson.decoder,s.set(i.name,i)),{charset:n,contentDecoder:r,contentDecoders:o,contentTypeDecoder:i,contentTypeDecoders:s}},_r.ReadableStreamMessageReader=class extends s{constructor(t,n){super(),this.readable=t,this.options=i.fromOptions(n),this.buffer=(0,e.default)().messageBuffer.create(this.options.charset),this._partialMessageTimeout=1e4,this.nextMessageLength=-1,this.messageToken=0,this.readSemaphore=new r.Semaphore(1)}set partialMessageTimeout(e){this._partialMessageTimeout=e}get partialMessageTimeout(){return this._partialMessageTimeout}listen(e){this.nextMessageLength=-1,this.messageToken=0,this.partialMessageTimer=void 0,this.callback=e;const t=this.readable.onData((e=>{this.onData(e)}));return this.readable.onError((e=>this.fireError(e))),this.readable.onClose((()=>this.fireClose())),t}onData(e){try{for(this.buffer.append(e);;){if(-1===this.nextMessageLength){const e=this.buffer.tryReadHeaders(!0);if(!e)return;const t=e.get("content-length");if(!t)return void this.fireError(new Error(`Header must provide a Content-Length property.\n${JSON.stringify(Object.fromEntries(e))}`));const n=parseInt(t);if(isNaN(n))return void this.fireError(new Error(`Content-Length value must be a number. Got ${t}`));this.nextMessageLength=n}const e=this.buffer.tryReadBody(this.nextMessageLength);if(void 0===e)return void this.setPartialMessageTimer();this.clearPartialMessageTimer(),this.nextMessageLength=-1,this.readSemaphore.lock((async()=>{const t=void 0!==this.options.contentDecoder?await this.options.contentDecoder.decode(e):e,n=await this.options.contentTypeDecoder.decode(t,this.options);this.callback(n)})).catch((e=>{this.fireError(e)}))}}catch(t){this.fireError(t)}}clearPartialMessageTimer(){this.partialMessageTimer&&(this.partialMessageTimer.dispose(),this.partialMessageTimer=void 0)}setPartialMessageTimer(){this.clearPartialMessageTimer(),this._partialMessageTimeout<=0||(this.partialMessageTimer=(0,e.default)().timer.setTimeout(((e,t)=>{this.partialMessageTimer=void 0,e===this.messageToken&&(this.firePartialMessage({messageToken:e,waitingTime:t}),this.setPartialMessageTimer())}),this._partialMessageTimeout,this.messageToken,this._partialMessageTimeout))}},_r}();Object.defineProperty(e,"MessageReader",{enumerable:!0,get:function(){return c.MessageReader}}),Object.defineProperty(e,"AbstractMessageReader",{enumerable:!0,get:function(){return c.AbstractMessageReader}}),Object.defineProperty(e,"ReadableStreamMessageReader",{enumerable:!0,get:function(){return c.ReadableStreamMessageReader}});const l=function(){if(Ar)return Lr;Ar=1,Object.defineProperty(Lr,"__esModule",{value:!0}),Lr.WriteableStreamMessageWriter=Lr.AbstractMessageWriter=Lr.MessageWriter=void 0;const e=tt(),t=Qe(),n=qr(),r=nt();var o,i;(o||(Lr.MessageWriter=o={})).is=function(e){let n=e;return n&&t.func(n.dispose)&&t.func(n.onClose)&&t.func(n.onError)&&t.func(n.write)};class s{constructor(){this.errorEmitter=new r.Emitter,this.closeEmitter=new r.Emitter}dispose(){this.errorEmitter.dispose(),this.closeEmitter.dispose()}get onError(){return this.errorEmitter.event}fireError(e,t,n){this.errorEmitter.fire([this.asError(e),t,n])}get onClose(){return this.closeEmitter.event}fireClose(){this.closeEmitter.fire(void 0)}asError(e){return e instanceof Error?e:new Error(`Writer received error. Reason: ${t.string(e.message)?e.message:"unknown"}`)}}return Lr.AbstractMessageWriter=s,(i||(i={})).fromOptions=function(t){return void 0===t||"string"==typeof t?{charset:t??"utf-8",contentTypeEncoder:(0,e.default)().applicationJson.encoder}:{charset:t.charset??"utf-8",contentEncoder:t.contentEncoder,contentTypeEncoder:t.contentTypeEncoder??(0,e.default)().applicationJson.encoder}},Lr.WriteableStreamMessageWriter=class extends s{constructor(e,t){super(),this.writable=e,this.options=i.fromOptions(t),this.errorCount=0,this.writeSemaphore=new n.Semaphore(1),this.writable.onError((e=>this.fireError(e))),this.writable.onClose((()=>this.fireClose()))}async write(e){return this.writeSemaphore.lock((async()=>this.options.contentTypeEncoder.encode(e,this.options).then((e=>void 0!==this.options.contentEncoder?this.options.contentEncoder.encode(e):e)).then((t=>{const n=[];return n.push("Content-Length: ",t.byteLength.toString(),"\r\n"),n.push("\r\n"),this.doWrite(e,n,t)}),(e=>{throw this.fireError(e),e}))))}async doWrite(e,t,n){try{return await this.writable.write(t.join(""),"ascii"),this.writable.write(n)}catch(r){return this.handleError(r,e),Promise.reject(r)}}handleError(e,t){this.errorCount++,this.fireError(e,t,this.errorCount)}end(){this.writable.end()}},Lr}();Object.defineProperty(e,"MessageWriter",{enumerable:!0,get:function(){return l.MessageWriter}}),Object.defineProperty(e,"AbstractMessageWriter",{enumerable:!0,get:function(){return l.AbstractMessageWriter}}),Object.defineProperty(e,"WriteableStreamMessageWriter",{enumerable:!0,get:function(){return l.WriteableStreamMessageWriter}});const u=(Fr||(Fr=1,Object.defineProperty(Hr,"__esModule",{value:!0}),Hr.AbstractMessageBuffer=void 0,Hr.AbstractMessageBuffer=class{constructor(e="utf-8"){this._encoding=e,this._chunks=[],this._totalLength=0}get encoding(){return this._encoding}append(e){const t="string"==typeof e?this.fromString(e,this._encoding):e;this._chunks.push(t),this._totalLength+=t.byteLength}tryReadHeaders(e=!1){if(0===this._chunks.length)return;let t=0,n=0,r=0,o=0;e:for(;n<this._chunks.length;){const e=this._chunks[n];for(r=0;r<e.length;){switch(e[r]){case 13:switch(t){case 0:t=1;break;case 2:t=3;break;default:t=0}break;case 10:switch(t){case 1:t=2;break;case 3:t=4,r++;break e;default:t=0}break;default:t=0}r++}o+=e.byteLength,n++}if(4!==t)return;const i=this._read(o+r),s=new Map,a=this.toString(i,"ascii").split("\r\n");if(a.length<2)return s;for(let c=0;c<a.length-2;c++){const t=a[c],n=t.indexOf(":");if(-1===n)throw new Error(`Message header must separate key and value using ':'\n${t}`);const r=t.substr(0,n),o=t.substr(n+1).trim();s.set(e?r.toLowerCase():r,o)}return s}tryReadBody(e){if(!(this._totalLength<e))return this._read(e)}get numberOfBytes(){return this._totalLength}_read(e){if(0===e)return this.emptyBuffer();if(e>this._totalLength)throw new Error("Cannot read so many bytes!");if(this._chunks[0].byteLength===e){const t=this._chunks[0];return this._chunks.shift(),this._totalLength-=e,this.asNative(t)}if(this._chunks[0].byteLength>e){const t=this._chunks[0],n=this.asNative(t,e);return this._chunks[0]=t.slice(e),this._totalLength-=e,n}const t=this.allocNative(e);let n=0;for(;e>0;){const r=this._chunks[0];if(r.byteLength>e){const o=r.slice(0,e);t.set(o,n),n+=e,this._chunks[0]=r.slice(e),this._totalLength-=e,e-=e}else t.set(r,n),n+=r.byteLength,this._chunks.shift(),this._totalLength-=r.byteLength,e-=r.byteLength}return t}}),Hr);Object.defineProperty(e,"AbstractMessageBuffer",{enumerable:!0,get:function(){return u.AbstractMessageBuffer}});const d=zr();Object.defineProperty(e,"ConnectionStrategy",{enumerable:!0,get:function(){return d.ConnectionStrategy}}),Object.defineProperty(e,"ConnectionOptions",{enumerable:!0,get:function(){return d.ConnectionOptions}}),Object.defineProperty(e,"NullLogger",{enumerable:!0,get:function(){return d.NullLogger}}),Object.defineProperty(e,"createMessageConnection",{enumerable:!0,get:function(){return d.createMessageConnection}}),Object.defineProperty(e,"ProgressToken",{enumerable:!0,get:function(){return d.ProgressToken}}),Object.defineProperty(e,"ProgressType",{enumerable:!0,get:function(){return d.ProgressType}}),Object.defineProperty(e,"Trace",{enumerable:!0,get:function(){return d.Trace}}),Object.defineProperty(e,"TraceValues",{enumerable:!0,get:function(){return d.TraceValues}}),Object.defineProperty(e,"TraceFormat",{enumerable:!0,get:function(){return d.TraceFormat}}),Object.defineProperty(e,"SetTraceNotification",{enumerable:!0,get:function(){return d.SetTraceNotification}}),Object.defineProperty(e,"LogTraceNotification",{enumerable:!0,get:function(){return d.LogTraceNotification}}),Object.defineProperty(e,"ConnectionErrors",{enumerable:!0,get:function(){return d.ConnectionErrors}}),Object.defineProperty(e,"ConnectionError",{enumerable:!0,get:function(){return d.ConnectionError}}),Object.defineProperty(e,"CancellationReceiverStrategy",{enumerable:!0,get:function(){return d.CancellationReceiverStrategy}}),Object.defineProperty(e,"CancellationSenderStrategy",{enumerable:!0,get:function(){return d.CancellationSenderStrategy}}),Object.defineProperty(e,"CancellationStrategy",{enumerable:!0,get:function(){return d.CancellationStrategy}}),Object.defineProperty(e,"MessageStrategy",{enumerable:!0,get:function(){return d.MessageStrategy}});const f=tt();e.RAL=f.default}(Tr)),Tr}Object.defineProperty(xr,"__esModule",{value:!0});const Vr=$r();class Kr extends Vr.AbstractMessageBuffer{constructor(e="utf-8"){super(e),this.asciiDecoder=new TextDecoder("ascii")}emptyBuffer(){return Kr.emptyBuffer}fromString(e,t){return(new TextEncoder).encode(e)}toString(e,t){return"ascii"===t?this.asciiDecoder.decode(e):new TextDecoder(t).decode(e)}asNative(e,t){return void 0===t?e:e.slice(0,t)}allocNative(e){return new Uint8Array(e)}}Kr.emptyBuffer=new Uint8Array(0);class Gr{constructor(e){this.socket=e,this._onData=new Vr.Emitter,this._messageListener=e=>{e.data.arrayBuffer().then((e=>{this._onData.fire(new Uint8Array(e))}),(()=>{(0,Vr.RAL)().console.error("Converting blob to array buffer failed.")}))},this.socket.addEventListener("message",this._messageListener)}onClose(e){return this.socket.addEventListener("close",e),Vr.Disposable.create((()=>this.socket.removeEventListener("close",e)))}onError(e){return this.socket.addEventListener("error",e),Vr.Disposable.create((()=>this.socket.removeEventListener("error",e)))}onEnd(e){return this.socket.addEventListener("end",e),Vr.Disposable.create((()=>this.socket.removeEventListener("end",e)))}onData(e){return this._onData.event(e)}}class Jr{constructor(e){this.socket=e}onClose(e){return this.socket.addEventListener("close",e),Vr.Disposable.create((()=>this.socket.removeEventListener("close",e)))}onError(e){return this.socket.addEventListener("error",e),Vr.Disposable.create((()=>this.socket.removeEventListener("error",e)))}onEnd(e){return this.socket.addEventListener("end",e),Vr.Disposable.create((()=>this.socket.removeEventListener("end",e)))}write(e,t){if("string"==typeof e){if(void 0!==t&&"utf-8"!==t)throw new Error(`In a Browser environments only utf-8 text encoding is supported. But got encoding: ${t}`);this.socket.send(e)}else this.socket.send(e);return Promise.resolve()}end(){this.socket.close()}}const Yr=new TextEncoder,Xr=Object.freeze({messageBuffer:Object.freeze({create:e=>new Kr(e)}),applicationJson:Object.freeze({encoder:Object.freeze({name:"application/json",encode:(e,t)=>{if("utf-8"!==t.charset)throw new Error(`In a Browser environments only utf-8 text encoding is supported. But got encoding: ${t.charset}`);return Promise.resolve(Yr.encode(JSON.stringify(e,void 0,0)))}}),decoder:Object.freeze({name:"application/json",decode:(e,t)=>{if(!(e instanceof Uint8Array))throw new Error("In a Browser environments only Uint8Arrays are supported.");return Promise.resolve(JSON.parse(new TextDecoder(t.charset).decode(e)))}})}),stream:Object.freeze({asReadableStream:e=>new Gr(e),asWritableStream:e=>new Jr(e)}),console:console,timer:Object.freeze({setTimeout(e,t,...n){const r=setTimeout(e,t,...n);return{dispose:()=>clearTimeout(r)}},setImmediate(e,...t){const n=setTimeout(e,0,...t);return{dispose:()=>clearTimeout(n)}},setInterval(e,t,...n){const r=setInterval(e,t,...n);return{dispose:()=>clearInterval(r)}}})});function Zr(){return Xr}(Zr||(Zr={})).install=function(){Vr.RAL.install(Xr)},xr.default=Zr,function(e){var t=T&&T.__createBinding||(Object.create?function(e,t,n,r){void 0===r&&(r=n);var o=Object.getOwnPropertyDescriptor(t,n);o&&!("get"in o?!t.__esModule:o.writable||o.configurable)||(o={enumerable:!0,get:function(){return t[n]}}),Object.defineProperty(e,r,o)}:function(e,t,n,r){void 0===r&&(r=n),e[r]=t[n]}),n=T&&T.__exportStar||function(e,n){for(var r in e)"default"===r||Object.prototype.hasOwnProperty.call(n,r)||t(n,e,r)};Object.defineProperty(e,"__esModule",{value:!0}),e.createMessageConnection=e.BrowserMessageWriter=e.BrowserMessageReader=void 0;xr.default.install();const r=$r();n($r(),e);class o extends r.AbstractMessageReader{constructor(e){super(),this._onData=new r.Emitter,this._messageListener=e=>{this._onData.fire(e.data)},e.addEventListener("error",(e=>this.fireError(e))),e.onmessage=this._messageListener}listen(e){return this._onData.event(e)}}e.BrowserMessageReader=o;class i extends r.AbstractMessageWriter{constructor(e){super(),this.port=e,this.errorCount=0,e.addEventListener("error",(e=>this.fireError(e)))}write(e){try{return this.port.postMessage(e),Promise.resolve()}catch(t){return this.handleError(t,e),Promise.reject(t)}}handleError(e,t){this.errorCount++,this.fireError(e,t,this.errorCount)}end(){}}e.BrowserMessageWriter=i,e.createMessageConnection=function(e,t,n,o){return void 0===n&&(n=r.NullLogger),r.ConnectionStrategy.is(o)&&(o={connectionStrategy:o}),(0,r.createMessageConnection)(e,t,n,o)}}(wr);var Qr=wr,eo={};const to=D(rt);var no={};Object.defineProperty(no,"__esModule",{value:!0}),no.ProtocolNotificationType=no.ProtocolNotificationType0=no.ProtocolRequestType=no.ProtocolRequestType0=no.RegistrationType=no.MessageDirection=void 0;const ro=wr;var oo,io;(io=oo||(no.MessageDirection=oo={})).clientToServer="clientToServer",io.serverToClient="serverToClient",io.both="both";no.RegistrationType=class{constructor(e){this.method=e}};no.ProtocolRequestType0=class extends ro.RequestType0{constructor(e){super(e)}};no.ProtocolRequestType=class extends ro.RequestType{constructor(e){super(e,ro.ParameterStructures.byName)}};no.ProtocolNotificationType0=class extends ro.NotificationType0{constructor(e){super(e)}};no.ProtocolNotificationType=class extends ro.NotificationType{constructor(e){super(e,ro.ParameterStructures.byName)}};var so={},ao={};function co(e){return"string"==typeof e||e instanceof String}function lo(e){return Array.isArray(e)}Object.defineProperty(ao,"__esModule",{value:!0}),ao.objectLiteral=ao.typedArray=ao.stringArray=ao.array=ao.func=ao.error=ao.number=ao.string=ao.boolean=void 0,ao.boolean=function(e){return!0===e||!1===e},ao.string=co,ao.number=function(e){return"number"==typeof e||e instanceof Number},ao.error=function(e){return e instanceof Error},ao.func=function(e){return"function"==typeof e},ao.array=lo,ao.stringArray=function(e){return lo(e)&&e.every((e=>co(e)))},ao.typedArray=function(e,t){return Array.isArray(e)&&e.every(t)},ao.objectLiteral=function(e){return null!==e&&"object"==typeof e};var uo={};Object.defineProperty(uo,"__esModule",{value:!0}),uo.ImplementationRequest=void 0;const fo=no;var ho,po;(po=ho||(uo.ImplementationRequest=ho={})).method="textDocument/implementation",po.messageDirection=fo.MessageDirection.clientToServer,po.type=new fo.ProtocolRequestType(po.method);var go={};Object.defineProperty(go,"__esModule",{value:!0}),go.TypeDefinitionRequest=void 0;const mo=no;var vo,yo;(yo=vo||(go.TypeDefinitionRequest=vo={})).method="textDocument/typeDefinition",yo.messageDirection=mo.MessageDirection.clientToServer,yo.type=new mo.ProtocolRequestType(yo.method);var bo={};Object.defineProperty(bo,"__esModule",{value:!0}),bo.DidChangeWorkspaceFoldersNotification=bo.WorkspaceFoldersRequest=void 0;const wo=no;var xo,To,Do,So;(To=xo||(bo.WorkspaceFoldersRequest=xo={})).method="workspace/workspaceFolders",To.messageDirection=wo.MessageDirection.serverToClient,To.type=new wo.ProtocolRequestType0(To.method),(So=Do||(bo.DidChangeWorkspaceFoldersNotification=Do={})).method="workspace/didChangeWorkspaceFolders",So.messageDirection=wo.MessageDirection.clientToServer,So.type=new wo.ProtocolNotificationType(So.method);var Oo={};Object.defineProperty(Oo,"__esModule",{value:!0}),Oo.ConfigurationRequest=void 0;const Ro=no;var jo,Co;(Co=jo||(Oo.ConfigurationRequest=jo={})).method="workspace/configuration",Co.messageDirection=Ro.MessageDirection.serverToClient,Co.type=new Ro.ProtocolRequestType(Co.method);var No={};Object.defineProperty(No,"__esModule",{value:!0}),No.ColorPresentationRequest=No.DocumentColorRequest=void 0;const ko=no;var Po,Eo,Io,_o;(Eo=Po||(No.DocumentColorRequest=Po={})).method="textDocument/documentColor",Eo.messageDirection=ko.MessageDirection.clientToServer,Eo.type=new ko.ProtocolRequestType(Eo.method),(_o=Io||(No.ColorPresentationRequest=Io={})).method="textDocument/colorPresentation",_o.messageDirection=ko.MessageDirection.clientToServer,_o.type=new ko.ProtocolRequestType(_o.method);var Mo={};Object.defineProperty(Mo,"__esModule",{value:!0}),Mo.FoldingRangeRefreshRequest=Mo.FoldingRangeRequest=void 0;const qo=no;var Ao,Lo,Fo,Ho;(Lo=Ao||(Mo.FoldingRangeRequest=Ao={})).method="textDocument/foldingRange",Lo.messageDirection=qo.MessageDirection.clientToServer,Lo.type=new qo.ProtocolRequestType(Lo.method),(Ho=Fo||(Mo.FoldingRangeRefreshRequest=Fo={})).method="workspace/foldingRange/refresh",Ho.messageDirection=qo.MessageDirection.serverToClient,Ho.type=new qo.ProtocolRequestType0(Ho.method);var Wo={};Object.defineProperty(Wo,"__esModule",{value:!0}),Wo.DeclarationRequest=void 0;const Uo=no;var Bo,zo;(zo=Bo||(Wo.DeclarationRequest=Bo={})).method="textDocument/declaration",zo.messageDirection=Uo.MessageDirection.clientToServer,zo.type=new Uo.ProtocolRequestType(zo.method);var $o={};Object.defineProperty($o,"__esModule",{value:!0}),$o.SelectionRangeRequest=void 0;const Vo=no;var Ko,Go;(Go=Ko||($o.SelectionRangeRequest=Ko={})).method="textDocument/selectionRange",Go.messageDirection=Vo.MessageDirection.clientToServer,Go.type=new Vo.ProtocolRequestType(Go.method);var Jo={};Object.defineProperty(Jo,"__esModule",{value:!0}),Jo.WorkDoneProgressCancelNotification=Jo.WorkDoneProgressCreateRequest=Jo.WorkDoneProgress=void 0;const Yo=wr,Xo=no;var Zo,Qo,ei,ti,ni,ri;(Qo=Zo||(Jo.WorkDoneProgress=Zo={})).type=new Yo.ProgressType,Qo.is=function(e){return e===Qo.type},(ti=ei||(Jo.WorkDoneProgressCreateRequest=ei={})).method="window/workDoneProgress/create",ti.messageDirection=Xo.MessageDirection.serverToClient,ti.type=new Xo.ProtocolRequestType(ti.method),(ri=ni||(Jo.WorkDoneProgressCancelNotification=ni={})).method="window/workDoneProgress/cancel",ri.messageDirection=Xo.MessageDirection.clientToServer,ri.type=new Xo.ProtocolNotificationType(ri.method);var oi={};Object.defineProperty(oi,"__esModule",{value:!0}),oi.CallHierarchyOutgoingCallsRequest=oi.CallHierarchyIncomingCallsRequest=oi.CallHierarchyPrepareRequest=void 0;const ii=no;var si,ai,ci,li,ui,di;(ai=si||(oi.CallHierarchyPrepareRequest=si={})).method="textDocument/prepareCallHierarchy",ai.messageDirection=ii.MessageDirection.clientToServer,ai.type=new ii.ProtocolRequestType(ai.method),(li=ci||(oi.CallHierarchyIncomingCallsRequest=ci={})).method="callHierarchy/incomingCalls",li.messageDirection=ii.MessageDirection.clientToServer,li.type=new ii.ProtocolRequestType(li.method),(di=ui||(oi.CallHierarchyOutgoingCallsRequest=ui={})).method="callHierarchy/outgoingCalls",di.messageDirection=ii.MessageDirection.clientToServer,di.type=new ii.ProtocolRequestType(di.method);var fi={};Object.defineProperty(fi,"__esModule",{value:!0}),fi.SemanticTokensRefreshRequest=fi.SemanticTokensRangeRequest=fi.SemanticTokensDeltaRequest=fi.SemanticTokensRequest=fi.SemanticTokensRegistrationType=fi.TokenFormat=void 0;const hi=no;var pi,gi,mi,vi,yi,bi,wi,xi,Ti,Di,Si;(pi||(fi.TokenFormat=pi={})).Relative="relative",(mi=gi||(fi.SemanticTokensRegistrationType=gi={})).method="textDocument/semanticTokens",mi.type=new hi.RegistrationType(mi.method),(yi=vi||(fi.SemanticTokensRequest=vi={})).method="textDocument/semanticTokens/full",yi.messageDirection=hi.MessageDirection.clientToServer,yi.type=new hi.ProtocolRequestType(yi.method),yi.registrationMethod=gi.method,(wi=bi||(fi.SemanticTokensDeltaRequest=bi={})).method="textDocument/semanticTokens/full/delta",wi.messageDirection=hi.MessageDirection.clientToServer,wi.type=new hi.ProtocolRequestType(wi.method),wi.registrationMethod=gi.method,(Ti=xi||(fi.SemanticTokensRangeRequest=xi={})).method="textDocument/semanticTokens/range",Ti.messageDirection=hi.MessageDirection.clientToServer,Ti.type=new hi.ProtocolRequestType(Ti.method),Ti.registrationMethod=gi.method,(Si=Di||(fi.SemanticTokensRefreshRequest=Di={})).method="workspace/semanticTokens/refresh",Si.messageDirection=hi.MessageDirection.serverToClient,Si.type=new hi.ProtocolRequestType0(Si.method);var Oi={};Object.defineProperty(Oi,"__esModule",{value:!0}),Oi.ShowDocumentRequest=void 0;const Ri=no;var ji,Ci;(Ci=ji||(Oi.ShowDocumentRequest=ji={})).method="window/showDocument",Ci.messageDirection=Ri.MessageDirection.serverToClient,Ci.type=new Ri.ProtocolRequestType(Ci.method);var Ni={};Object.defineProperty(Ni,"__esModule",{value:!0}),Ni.LinkedEditingRangeRequest=void 0;const ki=no;var Pi,Ei;(Ei=Pi||(Ni.LinkedEditingRangeRequest=Pi={})).method="textDocument/linkedEditingRange",Ei.messageDirection=ki.MessageDirection.clientToServer,Ei.type=new ki.ProtocolRequestType(Ei.method);var Ii={};Object.defineProperty(Ii,"__esModule",{value:!0}),Ii.WillDeleteFilesRequest=Ii.DidDeleteFilesNotification=Ii.DidRenameFilesNotification=Ii.WillRenameFilesRequest=Ii.DidCreateFilesNotification=Ii.WillCreateFilesRequest=Ii.FileOperationPatternKind=void 0;const _i=no;var Mi,qi,Ai,Li,Fi,Hi,Wi,Ui,Bi,zi,$i,Vi,Ki,Gi;(qi=Mi||(Ii.FileOperationPatternKind=Mi={})).file="file",qi.folder="folder",(Li=Ai||(Ii.WillCreateFilesRequest=Ai={})).method="workspace/willCreateFiles",Li.messageDirection=_i.MessageDirection.clientToServer,Li.type=new _i.ProtocolRequestType(Li.method),(Hi=Fi||(Ii.DidCreateFilesNotification=Fi={})).method="workspace/didCreateFiles",Hi.messageDirection=_i.MessageDirection.clientToServer,Hi.type=new _i.ProtocolNotificationType(Hi.method),(Ui=Wi||(Ii.WillRenameFilesRequest=Wi={})).method="workspace/willRenameFiles",Ui.messageDirection=_i.MessageDirection.clientToServer,Ui.type=new _i.ProtocolRequestType(Ui.method),(zi=Bi||(Ii.DidRenameFilesNotification=Bi={})).method="workspace/didRenameFiles",zi.messageDirection=_i.MessageDirection.clientToServer,zi.type=new _i.ProtocolNotificationType(zi.method),(Vi=$i||(Ii.DidDeleteFilesNotification=$i={})).method="workspace/didDeleteFiles",Vi.messageDirection=_i.MessageDirection.clientToServer,Vi.type=new _i.ProtocolNotificationType(Vi.method),(Gi=Ki||(Ii.WillDeleteFilesRequest=Ki={})).method="workspace/willDeleteFiles",Gi.messageDirection=_i.MessageDirection.clientToServer,Gi.type=new _i.ProtocolRequestType(Gi.method);var Ji={};Object.defineProperty(Ji,"__esModule",{value:!0}),Ji.MonikerRequest=Ji.MonikerKind=Ji.UniquenessLevel=void 0;const Yi=no;var Xi,Zi,Qi,es,ts,ns;(Zi=Xi||(Ji.UniquenessLevel=Xi={})).document="document",Zi.project="project",Zi.group="group",Zi.scheme="scheme",Zi.global="global",(es=Qi||(Ji.MonikerKind=Qi={})).$import="import",es.$export="export",es.local="local",(ns=ts||(Ji.MonikerRequest=ts={})).method="textDocument/moniker",ns.messageDirection=Yi.MessageDirection.clientToServer,ns.type=new Yi.ProtocolRequestType(ns.method);var rs={};Object.defineProperty(rs,"__esModule",{value:!0}),rs.TypeHierarchySubtypesRequest=rs.TypeHierarchySupertypesRequest=rs.TypeHierarchyPrepareRequest=void 0;const os=no;var is,ss,as,cs,ls,us;(ss=is||(rs.TypeHierarchyPrepareRequest=is={})).method="textDocument/prepareTypeHierarchy",ss.messageDirection=os.MessageDirection.clientToServer,ss.type=new os.ProtocolRequestType(ss.method),(cs=as||(rs.TypeHierarchySupertypesRequest=as={})).method="typeHierarchy/supertypes",cs.messageDirection=os.MessageDirection.clientToServer,cs.type=new os.ProtocolRequestType(cs.method),(us=ls||(rs.TypeHierarchySubtypesRequest=ls={})).method="typeHierarchy/subtypes",us.messageDirection=os.MessageDirection.clientToServer,us.type=new os.ProtocolRequestType(us.method);var ds={};Object.defineProperty(ds,"__esModule",{value:!0}),ds.InlineValueRefreshRequest=ds.InlineValueRequest=void 0;const fs=no;var hs,ps,gs,ms;(ps=hs||(ds.InlineValueRequest=hs={})).method="textDocument/inlineValue",ps.messageDirection=fs.MessageDirection.clientToServer,ps.type=new fs.ProtocolRequestType(ps.method),(ms=gs||(ds.InlineValueRefreshRequest=gs={})).method="workspace/inlineValue/refresh",ms.messageDirection=fs.MessageDirection.serverToClient,ms.type=new fs.ProtocolRequestType0(ms.method);var vs={};Object.defineProperty(vs,"__esModule",{value:!0}),vs.InlayHintRefreshRequest=vs.InlayHintResolveRequest=vs.InlayHintRequest=void 0;const ys=no;var bs,ws,xs,Ts,Ds,Ss;(ws=bs||(vs.InlayHintRequest=bs={})).method="textDocument/inlayHint",ws.messageDirection=ys.MessageDirection.clientToServer,ws.type=new ys.ProtocolRequestType(ws.method),(Ts=xs||(vs.InlayHintResolveRequest=xs={})).method="inlayHint/resolve",Ts.messageDirection=ys.MessageDirection.clientToServer,Ts.type=new ys.ProtocolRequestType(Ts.method),(Ss=Ds||(vs.InlayHintRefreshRequest=Ds={})).method="workspace/inlayHint/refresh",Ss.messageDirection=ys.MessageDirection.serverToClient,Ss.type=new ys.ProtocolRequestType0(Ss.method);var Os={};Object.defineProperty(Os,"__esModule",{value:!0}),Os.DiagnosticRefreshRequest=Os.WorkspaceDiagnosticRequest=Os.DocumentDiagnosticRequest=Os.DocumentDiagnosticReportKind=Os.DiagnosticServerCancellationData=void 0;const Rs=wr,js=ao,Cs=no;var Ns,ks,Ps,Es,Is,_s,Ms,qs,As;(Ns||(Os.DiagnosticServerCancellationData=Ns={})).is=function(e){const t=e;return t&&js.boolean(t.retriggerRequest)},(Ps=ks||(Os.DocumentDiagnosticReportKind=ks={})).Full="full",Ps.Unchanged="unchanged",(Is=Es||(Os.DocumentDiagnosticRequest=Es={})).method="textDocument/diagnostic",Is.messageDirection=Cs.MessageDirection.clientToServer,Is.type=new Cs.ProtocolRequestType(Is.method),Is.partialResult=new Rs.ProgressType,(Ms=_s||(Os.WorkspaceDiagnosticRequest=_s={})).method="workspace/diagnostic",Ms.messageDirection=Cs.MessageDirection.clientToServer,Ms.type=new Cs.ProtocolRequestType(Ms.method),Ms.partialResult=new Rs.ProgressType,(As=qs||(Os.DiagnosticRefreshRequest=qs={})).method="workspace/diagnostic/refresh",As.messageDirection=Cs.MessageDirection.serverToClient,As.type=new Cs.ProtocolRequestType0(As.method);var Ls={};Object.defineProperty(Ls,"__esModule",{value:!0}),Ls.DidCloseNotebookDocumentNotification=Ls.DidSaveNotebookDocumentNotification=Ls.DidChangeNotebookDocumentNotification=Ls.NotebookCellArrayChange=Ls.DidOpenNotebookDocumentNotification=Ls.NotebookDocumentSyncRegistrationType=Ls.NotebookDocument=Ls.NotebookCell=Ls.ExecutionSummary=Ls.NotebookCellKind=void 0;const Fs=to,Hs=ao,Ws=no;var Us,Bs,zs,$s,Vs,Ks,Gs,Js,Ys,Xs,Zs,Qs,ea,ta,na,ra,oa,ia,sa;(Bs=Us||(Ls.NotebookCellKind=Us={})).Markup=1,Bs.Code=2,Bs.is=function(e){return 1===e||2===e},($s=zs||(Ls.ExecutionSummary=zs={})).create=function(e,t){const n={executionOrder:e};return!0!==t&&!1!==t||(n.success=t),n},$s.is=function(e){const t=e;return Hs.objectLiteral(t)&&Fs.uinteger.is(t.executionOrder)&&(void 0===t.success||Hs.boolean(t.success))},$s.equals=function(e,t){return e===t||null!=e&&null!=t&&e.executionOrder===t.executionOrder&&e.success===t.success},function(e){function t(e,n){if(e===n)return!0;if(null==e||null==n)return!1;if(typeof e!=typeof n)return!1;if("object"!=typeof e)return!1;const r=Array.isArray(e),o=Array.isArray(n);if(r!==o)return!1;if(r&&o){if(e.length!==n.length)return!1;for(let r=0;r<e.length;r++)if(!t(e[r],n[r]))return!1}if(Hs.objectLiteral(e)&&Hs.objectLiteral(n)){const r=Object.keys(e),o=Object.keys(n);if(r.length!==o.length)return!1;if(r.sort(),o.sort(),!t(r,o))return!1;for(let i=0;i<r.length;i++){const o=r[i];if(!t(e[o],n[o]))return!1}}return!0}e.create=function(e,t){return{kind:e,document:t}},e.is=function(e){const t=e;return Hs.objectLiteral(t)&&Us.is(t.kind)&&Fs.DocumentUri.is(t.document)&&(void 0===t.metadata||Hs.objectLiteral(t.metadata))},e.diff=function(e,n){const r=new Set;return e.document!==n.document&&r.add("document"),e.kind!==n.kind&&r.add("kind"),e.executionSummary!==n.executionSummary&&r.add("executionSummary"),void 0===e.metadata&&void 0===n.metadata||t(e.metadata,n.metadata)||r.add("metadata"),void 0===e.executionSummary&&void 0===n.executionSummary||zs.equals(e.executionSummary,n.executionSummary)||r.add("executionSummary"),r}}(Vs||(Ls.NotebookCell=Vs={})),(Gs=Ks||(Ls.NotebookDocument=Ks={})).create=function(e,t,n,r){return{uri:e,notebookType:t,version:n,cells:r}},Gs.is=function(e){const t=e;return Hs.objectLiteral(t)&&Hs.string(t.uri)&&Fs.integer.is(t.version)&&Hs.typedArray(t.cells,Vs.is)},(Ys=Js||(Ls.NotebookDocumentSyncRegistrationType=Js={})).method="notebookDocument/sync",Ys.messageDirection=Ws.MessageDirection.clientToServer,Ys.type=new Ws.RegistrationType(Ys.method),(Zs=Xs||(Ls.DidOpenNotebookDocumentNotification=Xs={})).method="notebookDocument/didOpen",Zs.messageDirection=Ws.MessageDirection.clientToServer,Zs.type=new Ws.ProtocolNotificationType(Zs.method),Zs.registrationMethod=Js.method,(ea=Qs||(Ls.NotebookCellArrayChange=Qs={})).is=function(e){const t=e;return Hs.objectLiteral(t)&&Fs.uinteger.is(t.start)&&Fs.uinteger.is(t.deleteCount)&&(void 0===t.cells||Hs.typedArray(t.cells,Vs.is))},ea.create=function(e,t,n){const r={start:e,deleteCount:t};return void 0!==n&&(r.cells=n),r},(na=ta||(Ls.DidChangeNotebookDocumentNotification=ta={})).method="notebookDocument/didChange",na.messageDirection=Ws.MessageDirection.clientToServer,na.type=new Ws.ProtocolNotificationType(na.method),na.registrationMethod=Js.method,(oa=ra||(Ls.DidSaveNotebookDocumentNotification=ra={})).method="notebookDocument/didSave",oa.messageDirection=Ws.MessageDirection.clientToServer,oa.type=new Ws.ProtocolNotificationType(oa.method),oa.registrationMethod=Js.method,(sa=ia||(Ls.DidCloseNotebookDocumentNotification=ia={})).method="notebookDocument/didClose",sa.messageDirection=Ws.MessageDirection.clientToServer,sa.type=new Ws.ProtocolNotificationType(sa.method),sa.registrationMethod=Js.method;var aa={};Object.defineProperty(aa,"__esModule",{value:!0}),aa.InlineCompletionRequest=void 0;const ca=no;var la,ua;(ua=la||(aa.InlineCompletionRequest=la={})).method="textDocument/inlineCompletion",ua.messageDirection=ca.MessageDirection.clientToServer,ua.type=new ca.ProtocolRequestType(ua.method),function(e){Object.defineProperty(e,"__esModule",{value:!0}),e.WorkspaceSymbolRequest=e.CodeActionResolveRequest=e.CodeActionRequest=e.DocumentSymbolRequest=e.DocumentHighlightRequest=e.ReferencesRequest=e.DefinitionRequest=e.SignatureHelpRequest=e.SignatureHelpTriggerKind=e.HoverRequest=e.CompletionResolveRequest=e.CompletionRequest=e.CompletionTriggerKind=e.PublishDiagnosticsNotification=e.WatchKind=e.RelativePattern=e.FileChangeType=e.DidChangeWatchedFilesNotification=e.WillSaveTextDocumentWaitUntilRequest=e.WillSaveTextDocumentNotification=e.TextDocumentSaveReason=e.DidSaveTextDocumentNotification=e.DidCloseTextDocumentNotification=e.DidChangeTextDocumentNotification=e.TextDocumentContentChangeEvent=e.DidOpenTextDocumentNotification=e.TextDocumentSyncKind=e.TelemetryEventNotification=e.LogMessageNotification=e.ShowMessageRequest=e.ShowMessageNotification=e.MessageType=e.DidChangeConfigurationNotification=e.ExitNotification=e.ShutdownRequest=e.InitializedNotification=e.InitializeErrorCodes=e.InitializeRequest=e.WorkDoneProgressOptions=e.TextDocumentRegistrationOptions=e.StaticRegistrationOptions=e.PositionEncodingKind=e.FailureHandlingKind=e.ResourceOperationKind=e.UnregistrationRequest=e.RegistrationRequest=e.DocumentSelector=e.NotebookCellTextDocumentFilter=e.NotebookDocumentFilter=e.TextDocumentFilter=void 0,e.MonikerRequest=e.MonikerKind=e.UniquenessLevel=e.WillDeleteFilesRequest=e.DidDeleteFilesNotification=e.WillRenameFilesRequest=e.DidRenameFilesNotification=e.WillCreateFilesRequest=e.DidCreateFilesNotification=e.FileOperationPatternKind=e.LinkedEditingRangeRequest=e.ShowDocumentRequest=e.SemanticTokensRegistrationType=e.SemanticTokensRefreshRequest=e.SemanticTokensRangeRequest=e.SemanticTokensDeltaRequest=e.SemanticTokensRequest=e.TokenFormat=e.CallHierarchyPrepareRequest=e.CallHierarchyOutgoingCallsRequest=e.CallHierarchyIncomingCallsRequest=e.WorkDoneProgressCancelNotification=e.WorkDoneProgressCreateRequest=e.WorkDoneProgress=e.SelectionRangeRequest=e.DeclarationRequest=e.FoldingRangeRefreshRequest=e.FoldingRangeRequest=e.ColorPresentationRequest=e.DocumentColorRequest=e.ConfigurationRequest=e.DidChangeWorkspaceFoldersNotification=e.WorkspaceFoldersRequest=e.TypeDefinitionRequest=e.ImplementationRequest=e.ApplyWorkspaceEditRequest=e.ExecuteCommandRequest=e.PrepareRenameRequest=e.RenameRequest=e.PrepareSupportDefaultBehavior=e.DocumentOnTypeFormattingRequest=e.DocumentRangesFormattingRequest=e.DocumentRangeFormattingRequest=e.DocumentFormattingRequest=e.DocumentLinkResolveRequest=e.DocumentLinkRequest=e.CodeLensRefreshRequest=e.CodeLensResolveRequest=e.CodeLensRequest=e.WorkspaceSymbolResolveRequest=void 0,e.InlineCompletionRequest=e.DidCloseNotebookDocumentNotification=e.DidSaveNotebookDocumentNotification=e.DidChangeNotebookDocumentNotification=e.NotebookCellArrayChange=e.DidOpenNotebookDocumentNotification=e.NotebookDocumentSyncRegistrationType=e.NotebookDocument=e.NotebookCell=e.ExecutionSummary=e.NotebookCellKind=e.DiagnosticRefreshRequest=e.WorkspaceDiagnosticRequest=e.DocumentDiagnosticRequest=e.DocumentDiagnosticReportKind=e.DiagnosticServerCancellationData=e.InlayHintRefreshRequest=e.InlayHintResolveRequest=e.InlayHintRequest=e.InlineValueRefreshRequest=e.InlineValueRequest=e.TypeHierarchySupertypesRequest=e.TypeHierarchySubtypesRequest=e.TypeHierarchyPrepareRequest=void 0;const t=no,n=to,r=ao,o=uo;Object.defineProperty(e,"ImplementationRequest",{enumerable:!0,get:function(){return o.ImplementationRequest}});const i=go;Object.defineProperty(e,"TypeDefinitionRequest",{enumerable:!0,get:function(){return i.TypeDefinitionRequest}});const s=bo;Object.defineProperty(e,"WorkspaceFoldersRequest",{enumerable:!0,get:function(){return s.WorkspaceFoldersRequest}}),Object.defineProperty(e,"DidChangeWorkspaceFoldersNotification",{enumerable:!0,get:function(){return s.DidChangeWorkspaceFoldersNotification}});const a=Oo;Object.defineProperty(e,"ConfigurationRequest",{enumerable:!0,get:function(){return a.ConfigurationRequest}});const c=No;Object.defineProperty(e,"DocumentColorRequest",{enumerable:!0,get:function(){return c.DocumentColorRequest}}),Object.defineProperty(e,"ColorPresentationRequest",{enumerable:!0,get:function(){return c.ColorPresentationRequest}});const l=Mo;Object.defineProperty(e,"FoldingRangeRequest",{enumerable:!0,get:function(){return l.FoldingRangeRequest}}),Object.defineProperty(e,"FoldingRangeRefreshRequest",{enumerable:!0,get:function(){return l.FoldingRangeRefreshRequest}});const u=Wo;Object.defineProperty(e,"DeclarationRequest",{enumerable:!0,get:function(){return u.DeclarationRequest}});const d=$o;Object.defineProperty(e,"SelectionRangeRequest",{enumerable:!0,get:function(){return d.SelectionRangeRequest}});const f=Jo;Object.defineProperty(e,"WorkDoneProgress",{enumerable:!0,get:function(){return f.WorkDoneProgress}}),Object.defineProperty(e,"WorkDoneProgressCreateRequest",{enumerable:!0,get:function(){return f.WorkDoneProgressCreateRequest}}),Object.defineProperty(e,"WorkDoneProgressCancelNotification",{enumerable:!0,get:function(){return f.WorkDoneProgressCancelNotification}});const h=oi;Object.defineProperty(e,"CallHierarchyIncomingCallsRequest",{enumerable:!0,get:function(){return h.CallHierarchyIncomingCallsRequest}}),Object.defineProperty(e,"CallHierarchyOutgoingCallsRequest",{enumerable:!0,get:function(){return h.CallHierarchyOutgoingCallsRequest}}),Object.defineProperty(e,"CallHierarchyPrepareRequest",{enumerable:!0,get:function(){return h.CallHierarchyPrepareRequest}});const p=fi;Object.defineProperty(e,"TokenFormat",{enumerable:!0,get:function(){return p.TokenFormat}}),Object.defineProperty(e,"SemanticTokensRequest",{enumerable:!0,get:function(){return p.SemanticTokensRequest}}),Object.defineProperty(e,"SemanticTokensDeltaRequest",{enumerable:!0,get:function(){return p.SemanticTokensDeltaRequest}}),Object.defineProperty(e,"SemanticTokensRangeRequest",{enumerable:!0,get:function(){return p.SemanticTokensRangeRequest}}),Object.defineProperty(e,"SemanticTokensRefreshRequest",{enumerable:!0,get:function(){return p.SemanticTokensRefreshRequest}}),Object.defineProperty(e,"SemanticTokensRegistrationType",{enumerable:!0,get:function(){return p.SemanticTokensRegistrationType}});const g=Oi;Object.defineProperty(e,"ShowDocumentRequest",{enumerable:!0,get:function(){return g.ShowDocumentRequest}});const m=Ni;Object.defineProperty(e,"LinkedEditingRangeRequest",{enumerable:!0,get:function(){return m.LinkedEditingRangeRequest}});const v=Ii;Object.defineProperty(e,"FileOperationPatternKind",{enumerable:!0,get:function(){return v.FileOperationPatternKind}}),Object.defineProperty(e,"DidCreateFilesNotification",{enumerable:!0,get:function(){return v.DidCreateFilesNotification}}),Object.defineProperty(e,"WillCreateFilesRequest",{enumerable:!0,get:function(){return v.WillCreateFilesRequest}}),Object.defineProperty(e,"DidRenameFilesNotification",{enumerable:!0,get:function(){return v.DidRenameFilesNotification}}),Object.defineProperty(e,"WillRenameFilesRequest",{enumerable:!0,get:function(){return v.WillRenameFilesRequest}}),Object.defineProperty(e,"DidDeleteFilesNotification",{enumerable:!0,get:function(){return v.DidDeleteFilesNotification}}),Object.defineProperty(e,"WillDeleteFilesRequest",{enumerable:!0,get:function(){return v.WillDeleteFilesRequest}});const y=Ji;Object.defineProperty(e,"UniquenessLevel",{enumerable:!0,get:function(){return y.UniquenessLevel}}),Object.defineProperty(e,"MonikerKind",{enumerable:!0,get:function(){return y.MonikerKind}}),Object.defineProperty(e,"MonikerRequest",{enumerable:!0,get:function(){return y.MonikerRequest}});const b=rs;Object.defineProperty(e,"TypeHierarchyPrepareRequest",{enumerable:!0,get:function(){return b.TypeHierarchyPrepareRequest}}),Object.defineProperty(e,"TypeHierarchySubtypesRequest",{enumerable:!0,get:function(){return b.TypeHierarchySubtypesRequest}}),Object.defineProperty(e,"TypeHierarchySupertypesRequest",{enumerable:!0,get:function(){return b.TypeHierarchySupertypesRequest}});const w=ds;Object.defineProperty(e,"InlineValueRequest",{enumerable:!0,get:function(){return w.InlineValueRequest}}),Object.defineProperty(e,"InlineValueRefreshRequest",{enumerable:!0,get:function(){return w.InlineValueRefreshRequest}});const x=vs;Object.defineProperty(e,"InlayHintRequest",{enumerable:!0,get:function(){return x.InlayHintRequest}}),Object.defineProperty(e,"InlayHintResolveRequest",{enumerable:!0,get:function(){return x.InlayHintResolveRequest}}),Object.defineProperty(e,"InlayHintRefreshRequest",{enumerable:!0,get:function(){return x.InlayHintRefreshRequest}});const T=Os;Object.defineProperty(e,"DiagnosticServerCancellationData",{enumerable:!0,get:function(){return T.DiagnosticServerCancellationData}}),Object.defineProperty(e,"DocumentDiagnosticReportKind",{enumerable:!0,get:function(){return T.DocumentDiagnosticReportKind}}),Object.defineProperty(e,"DocumentDiagnosticRequest",{enumerable:!0,get:function(){return T.DocumentDiagnosticRequest}}),Object.defineProperty(e,"WorkspaceDiagnosticRequest",{enumerable:!0,get:function(){return T.WorkspaceDiagnosticRequest}}),Object.defineProperty(e,"DiagnosticRefreshRequest",{enumerable:!0,get:function(){return T.DiagnosticRefreshRequest}});const D=Ls;Object.defineProperty(e,"NotebookCellKind",{enumerable:!0,get:function(){return D.NotebookCellKind}}),Object.defineProperty(e,"ExecutionSummary",{enumerable:!0,get:function(){return D.ExecutionSummary}}),Object.defineProperty(e,"NotebookCell",{enumerable:!0,get:function(){return D.NotebookCell}}),Object.defineProperty(e,"NotebookDocument",{enumerable:!0,get:function(){return D.NotebookDocument}}),Object.defineProperty(e,"NotebookDocumentSyncRegistrationType",{enumerable:!0,get:function(){return D.NotebookDocumentSyncRegistrationType}}),Object.defineProperty(e,"DidOpenNotebookDocumentNotification",{enumerable:!0,get:function(){return D.DidOpenNotebookDocumentNotification}}),Object.defineProperty(e,"NotebookCellArrayChange",{enumerable:!0,get:function(){return D.NotebookCellArrayChange}}),Object.defineProperty(e,"DidChangeNotebookDocumentNotification",{enumerable:!0,get:function(){return D.DidChangeNotebookDocumentNotification}}),Object.defineProperty(e,"DidSaveNotebookDocumentNotification",{enumerable:!0,get:function(){return D.DidSaveNotebookDocumentNotification}}),Object.defineProperty(e,"DidCloseNotebookDocumentNotification",{enumerable:!0,get:function(){return D.DidCloseNotebookDocumentNotification}});const S=aa;var O,R,j,C,N,k,P,E,I,_,M,q,A,L,F,H,W,U,B,z,$,V,K,G,J,Y,X,Z,Q,ee,te,ne,re,oe,ie,se,ae,ce,le,ue,de,fe,he,pe,ge,me,ve,ye,be,we,xe,Te,De,Se,Oe,Re,je,Ce,Ne,ke,Pe,Ee,Ie,_e,Me,qe,Ae,Le,Fe,He,We,Ue,Be,ze,$e,Ve,Ke,Ge,Je,Ye,Xe,Ze,Qe,et,tt,nt,rt,ot,it,st,at,ct,lt,ut,dt,ft,ht,pt,gt,mt,vt,yt,bt,wt,xt,Tt,Dt,St,Ot,Rt,jt,Ct,Nt,kt,Pt,Et,It,_t,Mt,qt,At;Object.defineProperty(e,"InlineCompletionRequest",{enumerable:!0,get:function(){return S.InlineCompletionRequest}}),(O||(e.TextDocumentFilter=O={})).is=function(e){const t=e;return r.string(t)||r.string(t.language)||r.string(t.scheme)||r.string(t.pattern)},(R||(e.NotebookDocumentFilter=R={})).is=function(e){const t=e;return r.objectLiteral(t)&&(r.string(t.notebookType)||r.string(t.scheme)||r.string(t.pattern))},(j||(e.NotebookCellTextDocumentFilter=j={})).is=function(e){const t=e;return r.objectLiteral(t)&&(r.string(t.notebook)||R.is(t.notebook))&&(void 0===t.language||r.string(t.language))},(C||(e.DocumentSelector=C={})).is=function(e){if(!Array.isArray(e))return!1;for(let t of e)if(!r.string(t)&&!O.is(t)&&!j.is(t))return!1;return!0},(k=N||(e.RegistrationRequest=N={})).method="client/registerCapability",k.messageDirection=t.MessageDirection.serverToClient,k.type=new t.ProtocolRequestType(k.method),(E=P||(e.UnregistrationRequest=P={})).method="client/unregisterCapability",E.messageDirection=t.MessageDirection.serverToClient,E.type=new t.ProtocolRequestType(E.method),(_=I||(e.ResourceOperationKind=I={})).Create="create",_.Rename="rename",_.Delete="delete",(q=M||(e.FailureHandlingKind=M={})).Abort="abort",q.Transactional="transactional",q.TextOnlyTransactional="textOnlyTransactional",q.Undo="undo",(L=A||(e.PositionEncodingKind=A={})).UTF8="utf-8",L.UTF16="utf-16",L.UTF32="utf-32",(F||(e.StaticRegistrationOptions=F={})).hasId=function(e){const t=e;return t&&r.string(t.id)&&t.id.length>0},(H||(e.TextDocumentRegistrationOptions=H={})).is=function(e){const t=e;return t&&(null===t.documentSelector||C.is(t.documentSelector))},(U=W||(e.WorkDoneProgressOptions=W={})).is=function(e){const t=e;return r.objectLiteral(t)&&(void 0===t.workDoneProgress||r.boolean(t.workDoneProgress))},U.hasWorkDoneProgress=function(e){const t=e;return t&&r.boolean(t.workDoneProgress)},(z=B||(e.InitializeRequest=B={})).method="initialize",z.messageDirection=t.MessageDirection.clientToServer,z.type=new t.ProtocolRequestType(z.method),($||(e.InitializeErrorCodes=$={})).unknownProtocolVersion=1,(K=V||(e.InitializedNotification=V={})).method="initialized",K.messageDirection=t.MessageDirection.clientToServer,K.type=new t.ProtocolNotificationType(K.method),(J=G||(e.ShutdownRequest=G={})).method="shutdown",J.messageDirection=t.MessageDirection.clientToServer,J.type=new t.ProtocolRequestType0(J.method),(X=Y||(e.ExitNotification=Y={})).method="exit",X.messageDirection=t.MessageDirection.clientToServer,X.type=new t.ProtocolNotificationType0(X.method),(Q=Z||(e.DidChangeConfigurationNotification=Z={})).method="workspace/didChangeConfiguration",Q.messageDirection=t.MessageDirection.clientToServer,Q.type=new t.ProtocolNotificationType(Q.method),(te=ee||(e.MessageType=ee={})).Error=1,te.Warning=2,te.Info=3,te.Log=4,te.Debug=5,(re=ne||(e.ShowMessageNotification=ne={})).method="window/showMessage",re.messageDirection=t.MessageDirection.serverToClient,re.type=new t.ProtocolNotificationType(re.method),(ie=oe||(e.ShowMessageRequest=oe={})).method="window/showMessageRequest",ie.messageDirection=t.MessageDirection.serverToClient,ie.type=new t.ProtocolRequestType(ie.method),(ae=se||(e.LogMessageNotification=se={})).method="window/logMessage",ae.messageDirection=t.MessageDirection.serverToClient,ae.type=new t.ProtocolNotificationType(ae.method),(le=ce||(e.TelemetryEventNotification=ce={})).method="telemetry/event",le.messageDirection=t.MessageDirection.serverToClient,le.type=new t.ProtocolNotificationType(le.method),(de=ue||(e.TextDocumentSyncKind=ue={})).None=0,de.Full=1,de.Incremental=2,(he=fe||(e.DidOpenTextDocumentNotification=fe={})).method="textDocument/didOpen",he.messageDirection=t.MessageDirection.clientToServer,he.type=new t.ProtocolNotificationType(he.method),(ge=pe||(e.TextDocumentContentChangeEvent=pe={})).isIncremental=function(e){let t=e;return null!=t&&"string"==typeof t.text&&void 0!==t.range&&(void 0===t.rangeLength||"number"==typeof t.rangeLength)},ge.isFull=function(e){let t=e;return null!=t&&"string"==typeof t.text&&void 0===t.range&&void 0===t.rangeLength},(ve=me||(e.DidChangeTextDocumentNotification=me={})).method="textDocument/didChange",ve.messageDirection=t.MessageDirection.clientToServer,ve.type=new t.ProtocolNotificationType(ve.method),(be=ye||(e.DidCloseTextDocumentNotification=ye={})).method="textDocument/didClose",be.messageDirection=t.MessageDirection.clientToServer,be.type=new t.ProtocolNotificationType(be.method),(xe=we||(e.DidSaveTextDocumentNotification=we={})).method="textDocument/didSave",xe.messageDirection=t.MessageDirection.clientToServer,xe.type=new t.ProtocolNotificationType(xe.method),(De=Te||(e.TextDocumentSaveReason=Te={})).Manual=1,De.AfterDelay=2,De.FocusOut=3,(Oe=Se||(e.WillSaveTextDocumentNotification=Se={})).method="textDocument/willSave",Oe.messageDirection=t.MessageDirection.clientToServer,Oe.type=new t.ProtocolNotificationType(Oe.method),(je=Re||(e.WillSaveTextDocumentWaitUntilRequest=Re={})).method="textDocument/willSaveWaitUntil",je.messageDirection=t.MessageDirection.clientToServer,je.type=new t.ProtocolRequestType(je.method),(Ne=Ce||(e.DidChangeWatchedFilesNotification=Ce={})).method="workspace/didChangeWatchedFiles",Ne.messageDirection=t.MessageDirection.clientToServer,Ne.type=new t.ProtocolNotificationType(Ne.method),(Pe=ke||(e.FileChangeType=ke={})).Created=1,Pe.Changed=2,Pe.Deleted=3,(Ee||(e.RelativePattern=Ee={})).is=function(e){const t=e;return r.objectLiteral(t)&&(n.URI.is(t.baseUri)||n.WorkspaceFolder.is(t.baseUri))&&r.string(t.pattern)},(_e=Ie||(e.WatchKind=Ie={})).Create=1,_e.Change=2,_e.Delete=4,(qe=Me||(e.PublishDiagnosticsNotification=Me={})).method="textDocument/publishDiagnostics",qe.messageDirection=t.MessageDirection.serverToClient,qe.type=new t.ProtocolNotificationType(qe.method),(Le=Ae||(e.CompletionTriggerKind=Ae={})).Invoked=1,Le.TriggerCharacter=2,Le.TriggerForIncompleteCompletions=3,(He=Fe||(e.CompletionRequest=Fe={})).method="textDocument/completion",He.messageDirection=t.MessageDirection.clientToServer,He.type=new t.ProtocolRequestType(He.method),(Ue=We||(e.CompletionResolveRequest=We={})).method="completionItem/resolve",Ue.messageDirection=t.MessageDirection.clientToServer,Ue.type=new t.ProtocolRequestType(Ue.method),(ze=Be||(e.HoverRequest=Be={})).method="textDocument/hover",ze.messageDirection=t.MessageDirection.clientToServer,ze.type=new t.ProtocolRequestType(ze.method),(Ve=$e||(e.SignatureHelpTriggerKind=$e={})).Invoked=1,Ve.TriggerCharacter=2,Ve.ContentChange=3,(Ge=Ke||(e.SignatureHelpRequest=Ke={})).method="textDocument/signatureHelp",Ge.messageDirection=t.MessageDirection.clientToServer,Ge.type=new t.ProtocolRequestType(Ge.method),(Ye=Je||(e.DefinitionRequest=Je={})).method="textDocument/definition",Ye.messageDirection=t.MessageDirection.clientToServer,Ye.type=new t.ProtocolRequestType(Ye.method),(Ze=Xe||(e.ReferencesRequest=Xe={})).method="textDocument/references",Ze.messageDirection=t.MessageDirection.clientToServer,Ze.type=new t.ProtocolRequestType(Ze.method),(et=Qe||(e.DocumentHighlightRequest=Qe={})).method="textDocument/documentHighlight",et.messageDirection=t.MessageDirection.clientToServer,et.type=new t.ProtocolRequestType(et.method),(nt=tt||(e.DocumentSymbolRequest=tt={})).method="textDocument/documentSymbol",nt.messageDirection=t.MessageDirection.clientToServer,nt.type=new t.ProtocolRequestType(nt.method),(ot=rt||(e.CodeActionRequest=rt={})).method="textDocument/codeAction",ot.messageDirection=t.MessageDirection.clientToServer,ot.type=new t.ProtocolRequestType(ot.method),(st=it||(e.CodeActionResolveRequest=it={})).method="codeAction/resolve",st.messageDirection=t.MessageDirection.clientToServer,st.type=new t.ProtocolRequestType(st.method),(ct=at||(e.WorkspaceSymbolRequest=at={})).method="workspace/symbol",ct.messageDirection=t.MessageDirection.clientToServer,ct.type=new t.ProtocolRequestType(ct.method),(ut=lt||(e.WorkspaceSymbolResolveRequest=lt={})).method="workspaceSymbol/resolve",ut.messageDirection=t.MessageDirection.clientToServer,ut.type=new t.ProtocolRequestType(ut.method),(ft=dt||(e.CodeLensRequest=dt={})).method="textDocument/codeLens",ft.messageDirection=t.MessageDirection.clientToServer,ft.type=new t.ProtocolRequestType(ft.method),(pt=ht||(e.CodeLensResolveRequest=ht={})).method="codeLens/resolve",pt.messageDirection=t.MessageDirection.clientToServer,pt.type=new t.ProtocolRequestType(pt.method),(mt=gt||(e.CodeLensRefreshRequest=gt={})).method="workspace/codeLens/refresh",mt.messageDirection=t.MessageDirection.serverToClient,mt.type=new t.ProtocolRequestType0(mt.method),(yt=vt||(e.DocumentLinkRequest=vt={})).method="textDocument/documentLink",yt.messageDirection=t.MessageDirection.clientToServer,yt.type=new t.ProtocolRequestType(yt.method),(wt=bt||(e.DocumentLinkResolveRequest=bt={})).method="documentLink/resolve",wt.messageDirection=t.MessageDirection.clientToServer,wt.type=new t.ProtocolRequestType(wt.method),(Tt=xt||(e.DocumentFormattingRequest=xt={})).method="textDocument/formatting",Tt.messageDirection=t.MessageDirection.clientToServer,Tt.type=new t.ProtocolRequestType(Tt.method),(St=Dt||(e.DocumentRangeFormattingRequest=Dt={})).method="textDocument/rangeFormatting",St.messageDirection=t.MessageDirection.clientToServer,St.type=new t.ProtocolRequestType(St.method),(Rt=Ot||(e.DocumentRangesFormattingRequest=Ot={})).method="textDocument/rangesFormatting",Rt.messageDirection=t.MessageDirection.clientToServer,Rt.type=new t.ProtocolRequestType(Rt.method),(Ct=jt||(e.DocumentOnTypeFormattingRequest=jt={})).method="textDocument/onTypeFormatting",Ct.messageDirection=t.MessageDirection.clientToServer,Ct.type=new t.ProtocolRequestType(Ct.method),(Nt||(e.PrepareSupportDefaultBehavior=Nt={})).Identifier=1,(Pt=kt||(e.RenameRequest=kt={})).method="textDocument/rename",Pt.messageDirection=t.MessageDirection.clientToServer,Pt.type=new t.ProtocolRequestType(Pt.method),(It=Et||(e.PrepareRenameRequest=Et={})).method="textDocument/prepareRename",It.messageDirection=t.MessageDirection.clientToServer,It.type=new t.ProtocolRequestType(It.method),(Mt=_t||(e.ExecuteCommandRequest=_t={})).method="workspace/executeCommand",Mt.messageDirection=t.MessageDirection.clientToServer,Mt.type=new t.ProtocolRequestType(Mt.method),(At=qt||(e.ApplyWorkspaceEditRequest=qt={})).method="workspace/applyEdit",At.messageDirection=t.MessageDirection.serverToClient,At.type=new t.ProtocolRequestType("workspace/applyEdit")}(so);var da={};Object.defineProperty(da,"__esModule",{value:!0}),da.createProtocolConnection=void 0;const fa=wr;da.createProtocolConnection=function(e,t,n,r){return fa.ConnectionStrategy.is(r)&&(r={connectionStrategy:r}),(0,fa.createMessageConnection)(e,t,n,r)},function(e){var t=T&&T.__createBinding||(Object.create?function(e,t,n,r){void 0===r&&(r=n);var o=Object.getOwnPropertyDescriptor(t,n);o&&!("get"in o?!t.__esModule:o.writable||o.configurable)||(o={enumerable:!0,get:function(){return t[n]}}),Object.defineProperty(e,r,o)}:function(e,t,n,r){void 0===r&&(r=n),e[r]=t[n]}),n=T&&T.__exportStar||function(e,n){for(var r in e)"default"===r||Object.prototype.hasOwnProperty.call(n,r)||t(n,e,r)};Object.defineProperty(e,"__esModule",{value:!0}),e.LSPErrorCodes=e.createProtocolConnection=void 0,n(wr,e),n(to,e),n(no,e),n(so,e);var r,o,i=da;Object.defineProperty(e,"createProtocolConnection",{enumerable:!0,get:function(){return i.createProtocolConnection}}),(o=r||(e.LSPErrorCodes=r={})).lspReservedErrorRangeStart=-32899,o.RequestFailed=-32803,o.ServerCancelled=-32802,o.ContentModified=-32801,o.RequestCancelled=-32800,o.lspReservedErrorRangeEnd=-32800}(eo),function(e){var t=T&&T.__createBinding||(Object.create?function(e,t,n,r){void 0===r&&(r=n);var o=Object.getOwnPropertyDescriptor(t,n);o&&!("get"in o?!t.__esModule:o.writable||o.configurable)||(o={enumerable:!0,get:function(){return t[n]}}),Object.defineProperty(e,r,o)}:function(e,t,n,r){void 0===r&&(r=n),e[r]=t[n]}),n=T&&T.__exportStar||function(e,n){for(var r in e)"default"===r||Object.prototype.hasOwnProperty.call(n,r)||t(n,e,r)};Object.defineProperty(e,"__esModule",{value:!0}),e.createProtocolConnection=void 0;const r=Qr;n(Qr,e),n(eo,e),e.createProtocolConnection=function(e,t,n,o){return(0,r.createMessageConnection)(e,t,n,o)}}(br);const ha=1e4,pa=Object.fromEntries(Object.entries(br.CompletionItemKind).map((([e,t])=>[t,e]))),ga=e=>e.reduce(((e,t)=>t),""),ma=S.define({combine:ga}),va=S.define({combine:ga}),ya=S.define({combine:ga});class ba{constructor(e){this.rootUri=e.rootUri,this.workspaceFolders=e.workspaceFolders,this.autoClose=e.autoClose,this.plugins=[],this.transport=e.transport,this.requestManager=new Nt.RequestManager([this.transport]),this.client=new Nt.Client(this.requestManager),this.client.onNotification((e=>{this.processNotification(e)}));const t=this.transport;t&&t.connection&&t.connection.addEventListener("message",(e=>{const n=JSON.parse(e.data);n.method&&n.id&&t.connection.send(JSON.stringify({jsonrpc:"2.0",id:n.id,result:null}))})),this.initializePromise=this.initialize()}async initialize(){const{capabilities:e}=await this.request("initialize",{capabilities:{textDocument:{hover:{dynamicRegistration:!0,contentFormat:["plaintext","markdown"]},moniker:{},synchronization:{dynamicRegistration:!0,willSave:!1,didSave:!1,willSaveWaitUntil:!1},completion:{dynamicRegistration:!0,completionItem:{snippetSupport:!1,commitCharactersSupport:!0,documentationFormat:["plaintext","markdown"],deprecatedSupport:!1,preselectSupport:!1},contextSupport:!1},signatureHelp:{dynamicRegistration:!0,signatureInformation:{documentationFormat:["plaintext","markdown"]}},declaration:{dynamicRegistration:!0,linkSupport:!0},definition:{dynamicRegistration:!0,linkSupport:!0},typeDefinition:{dynamicRegistration:!0,linkSupport:!0},implementation:{dynamicRegistration:!0,linkSupport:!0}},workspace:{didChangeConfiguration:{dynamicRegistration:!0}}},initializationOptions:null,processId:null,rootUri:this.rootUri,workspaceFolders:this.workspaceFolders},3e4);this.capabilities=e,this.notify("initialized",{}),this.ready=!0}close(){this.client.close()}textDocumentDidOpen(e){return this.notify("textDocument/didOpen",e)}textDocumentDidChange(e){return this.notify("textDocument/didChange",e)}async textDocumentHover(e){return await this.request("textDocument/hover",e,ha)}async textDocumentCompletion(e){return await this.request("textDocument/completion",e,ha)}attachPlugin(e){this.plugins.push(e)}detachPlugin(e){const t=this.plugins.indexOf(e);-1!==t&&(this.plugins.splice(t,1),this.autoClose&&this.close())}request(e,t,n){return this.client.request({method:e,params:t},n)}notify(e,t){return this.client.notify({method:e,params:t})}processNotification(e){for(const t of this.plugins)t.processNotification(e)}}class wa{constructor(e,t){this.view=e,this.allowHTMLContent=t,this.client=this.view.state.facet(ma),this.documentUri=this.view.state.facet(va),this.languageId=this.view.state.facet(ya),this.documentVersion=0,this.changesTimeout=0,this.client.attachPlugin(this),this.initialize({documentText:this.view.state.doc.toString()})}update({docChanged:e}){e&&(this.changesTimeout&&clearTimeout(this.changesTimeout),this.changesTimeout=self.setTimeout((()=>{this.sendChange({documentText:this.view.state.doc.toString()})}),500))}destroy(){this.client.detachPlugin(this)}async initialize({documentText:e}){this.client.initializePromise&&await this.client.initializePromise,this.client.textDocumentDidOpen({textDocument:{uri:this.documentUri,languageId:this.languageId,text:e,version:this.documentVersion}})}async sendChange({documentText:e}){if(this.client.ready)try{await this.client.textDocumentDidChange({textDocument:{uri:this.documentUri,version:this.documentVersion++},contentChanges:[{text:e}]})}catch(t){console.error(t)}}requestDiagnostics(e){this.sendChange({documentText:e.state.doc.toString()})}async requestHoverTooltip(e,{line:t,character:n}){if(!this.client.ready||!this.client.capabilities.hoverProvider)return null;this.sendChange({documentText:e.state.doc.toString()});const r=await this.client.textDocumentHover({textDocument:{uri:this.documentUri},position:{line:t,character:n}});if(!r)return null;const{contents:o,range:i}=r;let s,a=xa(e.state.doc,{line:t,character:n});if(i&&(a=xa(e.state.doc,i.start),s=xa(e.state.doc,i.end)),null===a)return null;const c=document.createElement("div");return c.classList.add("documentation"),this.allowHTMLContent?c.innerHTML=Da(o):c.textContent=Da(o),{pos:a,end:s,create:e=>({dom:c}),above:!0}}async requestCompletion(e,{line:t,character:n},{triggerKind:r,triggerCharacter:o}){if(!this.client.ready||!this.client.capabilities.completionProvider)return null;this.sendChange({documentText:e.state.doc.toString()});const i=await this.client.textDocumentCompletion({textDocument:{uri:this.documentUri},position:{line:t,character:n},context:{triggerKind:r,triggerCharacter:o}});if(!i)return null;let s=("items"in i?i.items:i).map((({detail:e,label:t,kind:n,textEdit:r,documentation:o,sortText:i,filterText:s})=>{var a;const c={label:t,detail:e,apply:null!==(a=null==r?void 0:r.newText)&&void 0!==a?a:t,type:n&&pa[n].toLowerCase(),sortText:null!=i?i:t,filterText:null!=s?s:t};return o&&(c.info=Da(o)),c}));const[a,c]=function(e){const t=new Set,n=new Set;for(const{apply:o}of e){const[e,...r]=o;t.add(e);for(const t of r)n.add(t)}const r=Sa(t)+Sa(n)+"*$";return[new RegExp("^"+r),new RegExp(r)]}(s),l=e.matchBefore(c);let{pos:u}=e;if(l){u=l.from;const e=l.text.toLowerCase();/^\w+$/.test(e)&&(s=s.filter((({filterText:t})=>t.toLowerCase().startsWith(e))).sort((({apply:e},{apply:t})=>{switch(!0){case e.startsWith(l.text)&&!t.startsWith(l.text):return-1;case!e.startsWith(l.text)&&t.startsWith(l.text):return 1}return 0})))}return{from:u,options:s}}processNotification(e){try{if("textDocument/publishDiagnostics"===e.method)this.processDiagnostics(e.params)}catch(t){console.error(t)}}processDiagnostics(e){if(e.uri!==this.documentUri)return;const t=e.diagnostics.map((({range:e,message:t,severity:n})=>({from:xa(this.view.state.doc,e.start),to:xa(this.view.state.doc,e.end),severity:{[br.DiagnosticSeverity.Error]:"error",[br.DiagnosticSeverity.Warning]:"warning",[br.DiagnosticSeverity.Information]:"info",[br.DiagnosticSeverity.Hint]:"info"}[n],message:t}))).filter((({from:e,to:t})=>null!==e&&null!==t&&void 0!==e&&void 0!==t)).sort(((e,t)=>{switch(!0){case e.from<t.from:return-1;case e.from>t.from:return 1}return 0}));this.view.dispatch(C(this.view.state,t))}}function xa(e,t){if(t.line>=e.lines)return;const n=e.line(t.line+1).from+t.character;return n>e.length?void 0:n}function Ta(e,t){const n=e.lineAt(t);return{line:n.number-1,character:t-n.from}}function Da(e){return Array.isArray(e)?e.map((e=>Da(e)+"\n\n")).join(""):"string"==typeof e?e:e.value}function Sa(e){let t="",n=Array.from(e).join("");return/\w/.test(n)&&(t+="\\w",n=n.replace(/\w/g,"")),`[${t}${n.replace(/[^\w\s]/g,"\\$&")}]`}function Oa(e){return[...N().map((t=>{const n=k(t);return n===e?null:n})).filter(Boolean).sort(((e,t)=>e.startsWith("import")&&!t.startsWith("import")?-1:!e.startsWith("import")&&t.startsWith("import")?1:0)),e].join("\n")}class Ra extends ba{constructor(...e){super(...e),this.documentVersion=0,this.getCompletionInternal=async(e,t)=>await this._request("getCompletions",{doc:{...e.doc,version:t}}),this.throttledGetCompletionInternal=function(e,r,o){var i=!0,s=!0;if("function"!=typeof e)throw new TypeError("Expected a function");return t(o)&&(i="leading"in o?!!o.leading:i,s="trailing"in o?!!o.trailing:s),n(e,r,{leading:i,maxWait:r,trailing:s})}(this.getCompletionInternal,200)}_request(e,t){return this.request(e,t)}isDisabled(){return!jt()}async textDocumentDidOpen(e){return this.isDisabled()?e:super.textDocumentDidOpen(e)}async textDocumentCompletion(e){return this.isDisabled()?[]:super.textDocumentCompletion(e)}async textDocumentDidChange(e){return this.isDisabled()?e:super.textDocumentDidChange({...e,contentChanges:[{text:Oa(e.contentChanges[0].text)}],textDocument:br.VersionedTextDocumentIdentifier.create(e.textDocument.uri,++this.documentVersion)})}textDocumentHover(e){return this.isDisabled()?Promise.resolve({contents:[]}):super.textDocumentHover(e)}signOut(){return this._request("signOut",{})}signInInitiate(){return this._request("signInInitiate",{})}signInConfirm(e){return this._request("signInConfirm",e)}async signedIn(){const{status:e,user:t}=await this._request("checkStatus",{});return P.debug("Copilot#signedIn",e,t),"SignedIn"===e||"AlreadySignedIn"===e||"OK"===e}acceptCompletion(e){return this._request("notifyAccepted",e)}rejectCompletions(e){return this._request("notifyRejected",e)}async getCompletion(e){if(this.isDisabled())return{completions:[]};const t=this.documentVersion;if(e.doc.version=t,0===t)return{completions:[]};var n;n=t,y.set(Rt,n);const r=await this.throttledGetCompletionInternal(e,t);var o;return o=t,y.get(Rt)===o&&y.set(Rt,null),t!==this.documentVersion?{completions:[]}:r||{completions:[]}}}async function ja(e){return new Promise(((t,n)=>{const r=new WebSocket(e);r.onopen=()=>{r.close(),t(e)},r.onerror=t=>{r.close(),n(new Error(`Failed to connect to ${e}: ${t.type}`))}}))}const Ca="/marimo.py",Na="copilot",ka=`file://${Ca}`,Pa=E((()=>new Ea));class Ea extends en{constructor(){super(),this.delegate=void 0,this.delegate=void 0}async connect(){return await w(x,(e=>"github"===e.completion.copilot)),await async function(e,t=5){for(let n=0;n<t;n++)try{return await ja(e)}catch{await new Promise((e=>setTimeout(e,1e3*(n+1))))}throw new Error(`Failed to connect to ${e}`)}("/lsp/copilot",3),this.delegate||(this.delegate=new Nt.WebSocketTransport("/lsp/copilot")),this.delegate.connect()}close(){var e;null==(e=this.delegate)||e.close()}async sendData(e,t){var n;return null==(n=this.delegate)?void 0:n.sendData(e,t)}}const Ia=E((()=>new Ra({rootUri:ka,documentUri:ka,languageId:Na,workspaceFolders:null,transport:Pa()})));function _a(){return function(e){let t=null;return[ma.of(e.client||new ba({...e,autoClose:!0})),va.of(e.documentUri),ya.of(e.languageId),O.define((n=>t=new wa(n,e.allowHTMLContent))),R(((e,n)=>{var r;return null!==(r=null==t?void 0:t.requestHoverTooltip(e,Ta(e.state.doc,n)))&&void 0!==r?r:null})),j({override:[async e=>{var n,r,o;if(null==t)return null;const{state:i,pos:s,explicit:a}=e,c=i.doc.lineAt(s);let l,u=br.CompletionTriggerKind.Invoked;return!a&&(null===(o=null===(r=null===(n=t.client.capabilities)||void 0===n?void 0:n.completionProvider)||void 0===r?void 0:r.triggerCharacters)||void 0===o?void 0:o.includes(c.text[s-c.from-1]))&&(u=br.CompletionTriggerKind.TriggerCharacter,l=c.text[s-c.from-1]),u!==br.CompletionTriggerKind.Invoked||e.matchBefore(/\w+$/)?await t.requestCompletion(e,Ta(i.doc,s),{triggerKind:u,triggerCharacter:l}):null}]})]}({rootUri:ka,documentUri:ka,workspaceFolders:[],transport:Pa(),client:Ia(),languageId:Na})}const Ma=h.memo((()=>{const[e,t]=I(Ot),[n,r]=h.useState(),[o,i]=h.useState(),[s,c]=h.useState(!1);h.useEffect((()=>{const e=Ia();e.initializePromise.catch((()=>{t(!1),r("notConnected")})),e.signedIn().then((e=>{t(e),r(e?"signedIn":"signedOut")})).catch((()=>{t(!1),r("notConnected")}))}),[]);const l=async e=>{e.preventDefault(),c(!0);try{const e=Ia(),{verificationUri:n,status:o,userCode:s}=await e.signInInitiate();"OK"===o||"AlreadySignedIn"===o?t(!0):(r("signingIn"),i({url:n,code:s}))}finally{c(!1)}},d=async e=>{if(e.preventDefault(),!o)return;const n=Ia();try{c(!0);const{status:e}=await n.signInConfirm({userCode:o.code});"OK"===e||"AlreadySignedIn"===e?(t(!0),r("signedIn")):r("signInFailed")}catch{for(let e=0;e<3;e++){await new Promise((e=>setTimeout(e,1e3)));if(await n.signedIn())return t(!0),void r("signedIn")}r("signInFailed")}finally{c(!1)}},f=async e=>{e.preventDefault();const n=Ia();t(!1),r("signedOut"),await n.signOut()};return(()=>{switch(n??(e?"signedIn":"connecting")){case"connecting":return a.jsx(_,{className:"font-normal flex",children:"Connecting..."});case"signedOut":return a.jsx(q,{onClick:l,size:"xs",variant:"link",children:"Sign in to GitHub Copilot"});case"signingIn":return a.jsxs("ol",{className:"ml-4 text-sm list-decimal [&>li]:mt-2",children:[a.jsx("li",{children:a.jsxs("div",{className:"flex items-center",children:["Copy this code:",a.jsx("strong",{className:"ml-2",children:null==o?void 0:o.code}),a.jsx(L,{className:"ml-2 cursor-pointer opacity-60 hover:opacity-100 h-3 w-3",onClick:async()=>{o&&(await F(o.code),H({description:"Copied to clipboard"}))}})]})}),a.jsxs("li",{children:["Enter the code at this link:",a.jsx("a",{href:null==o?void 0:o.url,target:"_blank",rel:"noreferrer",className:W({variant:"link",size:"xs"}),children:null==o?void 0:o.url})]}),a.jsxs("li",{children:["Click here when done:",a.jsxs(q,{size:"xs",onClick:d,className:"ml-1",children:[s&&a.jsx(A,{className:"h-3 w-3 mr-1 animate-spin"}),"Done"]})]})]});case"signInFailed":return a.jsxs("div",{className:"flex flex-col gap-1",children:[a.jsx("div",{className:"text-destructive text-sm",children:"Sign in failed. Please try again."}),s?a.jsx(A,{className:"h-3 w-3 mr-1 animate-spin"}):a.jsx(q,{onClick:l,size:"xs",variant:"link",children:"Connect to GitHub Copilot"})]});case"signedIn":return a.jsxs("div",{className:"flex items-center gap-5",children:[a.jsxs(_,{className:"font-normal flex items-center",children:[a.jsx("div",{className:"inline-flex items-center justify-center bg-[var(--grass-7)] rounded-full p-1 mr-2",children:a.jsx(M,{className:"h-3 w-3 text-white"})}),"Connected"]}),a.jsx(q,{onClick:f,size:"xs",variant:"text",children:"Disconnect"})]});case"notConnected":return a.jsxs("div",{className:"flex flex-col gap-1",children:[a.jsxs(_,{className:"font-normal flex",children:[a.jsx(u,{className:"h-4 w-4 mr-1"}),"Unable to connect"]}),a.jsxs("div",{className:"text-sm",children:["For troubleshooting, see the"," ",a.jsx("a",{className:"hyperlink",href:"https://docs.marimo.io/getting_started/index.html#github-copilot",target:"_blank",rel:"noreferrer",children:"docs"}),"."]})]})}})()}));var qa;function Aa(e){return La(),[U.of([{key:"j",run:t=>!(!B(t,!0)||!z(t))&&(e.focusDown(),!0)}]),U.of([{key:"k",run:t=>!(!$(t)||!z(t))&&(e.focusUp(),!0)}]),O.define((e=>(Fa.INSTANCES.addInstance(e),{destroy(){Fa.INSTANCES.removeInstance(e)}})))]}Ma.displayName="CopilotConfig";const La=E((()=>{V.defineAction("goToDefinition",(e=>{const t=e.cm6;return K(t)})),V.mapCommand("gd","action","goToDefinition",{},{context:"normal"})}));class Fa{constructor(){this.instances=new Set,this.isBroadcasting=!1}addInstance(e){this.instances.add(e);const t=G(e);t?t.on("vim-mode-change",(t=>{if(this.isBroadcasting)return;J("mode"in t,'Expected event to have a "mode" property');const n=t.mode;this.isBroadcasting=!0,this.broadcastModeChange(e,n),this.isBroadcasting=!1})):P.warn("Expected CodeMirror instance to have CodeMirror instance state")}removeInstance(e){this.instances.delete(e)}broadcastModeChange(e,t){J("exitInsertMode"in V,"Vim does not have an exitInsertMode method"),J("exitVisualMode"in V,"Vim does not have an exitVisualMode method");for(const n of this.instances)if(n!==e){const e=G(n);if(!e){P.warn("Expected CodeMirror instance to have CodeMirror instance state");continue}const r=e.setSelections.bind(e);e.setSelections=()=>[];const o=e.state.vim;if(!o){P.warn("Expected CodeMirror instance to have Vim state");continue}switch(t){case"normal":o.insertMode&&V.exitInsertMode(e,!0),o.visualMode&&V.exitVisualMode(e,!0);break;case"insert":o.insertMode||V.handleKey(e,"i","")}e.setSelections=r}}}qa=Fa,Fa.INSTANCES=new qa;const Ha=["default","vim"];function Wa(e,t){switch(e.preset){case"default":return[U.of(Q),U.of([{key:"Escape",preventDefault:!0,run:e=>(e.contentDOM.blur(),!0)}])];case"vim":return[X.highest(Ua("d",(e=>""===e.state.doc.toString()),(e=>""===e.state.doc.toString()&&(t.deleteCell(),!0)))),Z({status:!1}),Aa(t),U.of(Q)];default:return Y(e.preset),[]}}function Ua(e,t,n){let r="",o=0;return U.of([{any:(i,s)=>{const a=s.key,c=s.timeStamp;if(a!==e||!t(i))return r="",o=0,!1;if(r===e&&c-o<500){if(n(i))return r="",o=0,!0}return r=a,o=c,!1}}])}const Ba={markdown:!0,wasm_layouts:!1,scratchpad:!0,multi_column:!1,chat_sidebar:!1};function za(e){var t;return(null==(t=b().experimental)?void 0:t[e])??Ba[e]}function $a(){return za("multi_column")?["compact","medium","full","columns"]:["compact","medium","full"]}!function(e,t){if("undefined"==typeof window)return;const n=`__marimo__${t}`;window[n]&&P.warn(`Overwriting existing debug object ${n}`),window[n]=e}((function(e,t){const n=b();n.experimental=n.experimental??{},n.experimental[e]=t,ee({config:n})}),"setFeatureFlag");const Va=v({});function Ka(){return te(Va)}function Ga(e,t){const n=i(13),{registerAction:r,unregisterAction:o}=function(){const e=ne(Va);return{registerAction:re(((t,n)=>{e((e=>({...e,[t]:n})))})),unregisterAction:re((t=>{e((e=>{const{[t]:n,...r}=e;return r}))}))}}(),s=te(oe),a=t===ie.NOOP;let c;n[0]!==t?(c=()=>t(),n[0]=t,n[1]=c):c=n[1];const l=re(c);let u;n[2]!==s||n[3]!==e||n[4]!==t?(u=n=>{const r=s.getHotkey(e).key;if(se(r)(n)){!1!==t()&&(n.preventDefault(),n.stopPropagation())}},n[2]=s,n[3]=e,n[4]=t,n[5]=u):u=n[5];const d=re(u);let f,g;p(document,"keydown",d),n[6]!==a||n[7]!==r||n[8]!==e||n[9]!==l||n[10]!==o?(f=()=>{if(!a)return r(e,l),()=>o(e)},g=[l,e,a,r,o],n[6]=a,n[7]=r,n[8]=e,n[9]=l,n[10]=o,n[11]=f,n[12]=g):(f=n[11],g=n[12]),h.useEffect(f,g)}function Ja(e,t){const n=i(3),r=te(oe);let o;n[0]!==t||n[1]!==r?(o=e=>{for(const[n,o]of ae.entries(t)){const t=r.getHotkey(n).key;if(se(t)(e)){P.debug("Satisfied",t,e);!1!==o()&&(e.preventDefault(),e.stopPropagation())}}},n[0]=t,n[1]=r,n[2]=o):o=n[2],p(e,"keydown",o)}function Ya(e,t){const n=i(2);let r;n[0]!==t?(r=e=>{for(const[n,r]of ae.entries(t))if(se(n)(e)){P.debug("Satisfied",n,e);!1!==r()&&(e.preventDefault(),e.stopPropagation())}},n[0]=t,n[1]=r):r=n[1],p(e,"keydown",r)}const Xa=v(!1),Za=()=>{const e=i(78),[t,n]=I(Xa),[r,o]=h.useState(null);let s;e[0]===Symbol.for("react.memo_cache_sentinel")?(s=[],e[0]=s):s=e[0];const[c,l]=h.useState(s),[d,f]=ce(),p=te(oe);let g,m;e[1]!==n?(g=()=>n((e=>!e)),e[1]=n,e[2]=g):g=e[2],Ga("global.showHelp",g),e[3]!==d||e[4]!==f?(m=async e=>{const t={...d};f(e),await ee({config:e}).catch((e=>{throw f(t),e}))},e[3]=d,e[4]=f,e[5]=m):m=e[5];const v=m;let y;e[6]!==r||e[7]!==d||e[8]!==v?(y=async e=>{if(!r)return;const t=e.join("-"),n={...d,keymap:{...d.keymap,overrides:{...d.keymap.overrides,[r]:t}}};o(null),l([]),await v(n)},e[6]=r,e[7]=d,e[8]=v,e[9]=y):y=e[9];const b=y;let w;e[10]!==r||e[11]!==d||e[12]!==v?(w=async()=>{if(!r)return;const e={...d,keymap:{...d.keymap,overrides:{...d.keymap.overrides}}};delete e.keymap.overrides[r],o(null),l([]),await v(e)},e[10]=r,e[11]=d,e[12]=v,e[13]=w):w=e[13];const x=w;let T;e[14]!==d||e[15]!==v?(T=async()=>{if(!window.confirm("Are you sure you want to reset all shortcuts to their default values?"))return;const e={...d,keymap:{...d.keymap,overrides:{}}};o(null),l([]),await v(e)},e[14]=d,e[15]=v,e[16]=T):T=e[16];const D=T;if(!t)return null;let S;e[17]!==p||e[18]!==r||e[19]!==c||e[20]!==b||e[21]!==x?(S=e=>{const t=p.getHotkey(e);if(r===e){const n=le(e);return a.jsxs("div",{children:[a.jsx(ue,{defaultValue:c.join("+"),placeholder:t.name,onKeyDown:e=>{e.preventDefault();const t=[];if("Meta"===e.key||"Control"===e.key||"Alt"===e.key||"Shift"===e.key)return;if(e.metaKey&&t.push(de()?"Cmd":"Meta"),e.ctrlKey&&t.push("Ctrl"),e.altKey&&t.push("Alt"),e.shiftKey&&t.push("Shift"),"-"===e.key)return;if("Escape"===e.key&&0===t.length)return o(null),void l([]);let n=e.key.toLowerCase();" "===e.key&&(n="Space"),t.push(n),b(t)},autoFocus:!0,endAdornment:a.jsx(q,{variant:"text",size:"xs",className:"mb-0",onClick:()=>{o(null),l([])},children:a.jsx(u,{className:"w-4 h-4"})})}),a.jsxs("div",{className:"flex items-center justify-between w-full",children:[a.jsx("span",{className:"text-muted-foreground text-xs",children:"Press a key combination"}),n.key!==t.key&&a.jsxs("span",{className:"text-xs cursor-pointer text-primary",onClick:x,children:["Reset to default:"," ",a.jsx("span",{className:"font-mono",children:n.key})]})]})]},e)}return a.jsxs("div",{className:"grid grid-cols-[auto,2fr,3fr] gap-2 items-center",children:[a.jsx(fe,{className:"cursor-pointer opacity-60 hover:opacity-100 text-muted-foreground w-3 h-3",onClick:()=>{l([]),o(e)}}),a.jsx(he,{className:"justify-end",shortcut:t.key}),a.jsx("span",{children:t.name.toLowerCase()})]},e)},e[17]=p,e[18]=r,e[19]=c,e[20]=b,e[21]=x,e[22]=S):S=e[22];const O=S;let R,j,C,N,k,P,E,_,M,A,L,F,H,W,U,B,z,$,V,K,G,J,Y,X;if(e[23]!==p||e[24]!==O||e[25]!==t||e[26]!==n){const r=p.getHotkeyGroups(),o=e=>{const t=r[e];return a.jsxs("div",{className:"mb-[40px] gap-2 flex flex-col",children:[a.jsx("h3",{className:"text-lg font-medium",children:e}),t.map((e=>O(e)))]})};U=pe,B=t,e[44]!==n?(z=e=>n(e),e[44]=n,e[45]=z):z=e[45],F=ge,H="sm:items-center sm:top-0",e[46]===Symbol.for("react.memo_cache_sentinel")?(W=a.jsx(me,{}),e[46]=W):W=e[46],_=ve,M=!1,A="max-h-screen sm:max-h-[90vh] overflow-y-auto sm:max-w-[850px]",e[47]===Symbol.for("react.memo_cache_sentinel")?(L=a.jsx(ye,{children:a.jsx(be,{children:"Shortcuts"})}),e[47]=L):L=e[47],P="flex flex-row gap-3",E=a.jsxs("div",{className:"w-1/2",children:[o("Editing"),o("Markdown")]}),R="w-1/2",j=o("Navigation"),C=o("Running Cells"),N=o("Creation and Ordering"),k=o("Other"),e[23]=p,e[24]=O,e[25]=t,e[26]=n,e[27]=R,e[28]=j,e[29]=C,e[30]=N,e[31]=k,e[32]=P,e[33]=E,e[34]=_,e[35]=M,e[36]=A,e[37]=L,e[38]=F,e[39]=H,e[40]=W,e[41]=U,e[42]=B,e[43]=z}else R=e[27],j=e[28],C=e[29],N=e[30],k=e[31],P=e[32],E=e[33],_=e[34],M=e[35],A=e[36],L=e[37],F=e[38],H=e[39],W=e[40],U=e[41],B=e[42],z=e[43];return e[48]===Symbol.for("react.memo_cache_sentinel")?($=a.jsx("span",{className:"text-destructive",children:"Reset all to default"}),e[48]=$):$=e[48],e[49]!==D?(V=a.jsx(q,{className:"mt-4 hover:bg-destructive/10 hover:border-destructive",variant:"outline",size:"xs",onClick:D,tabIndex:-1,children:$}),e[49]=D,e[50]=V):V=e[50],e[51]!==R||e[52]!==j||e[53]!==C||e[54]!==N||e[55]!==k||e[56]!==V?(K=a.jsxs("div",{className:R,children:[j,C,N,k,V]}),e[51]=R,e[52]=j,e[53]=C,e[54]=N,e[55]=k,e[56]=V,e[57]=K):K=e[57],e[58]!==P||e[59]!==E||e[60]!==K?(G=a.jsxs("div",{className:P,children:[E,K]}),e[58]=P,e[59]=E,e[60]=K,e[61]=G):G=e[61],e[62]!==_||e[63]!==M||e[64]!==A||e[65]!==L||e[66]!==G?(J=a.jsxs(_,{usePortal:M,className:A,children:[L,G]}),e[62]=_,e[63]=M,e[64]=A,e[65]=L,e[66]=G,e[67]=J):J=e[67],e[68]!==F||e[69]!==H||e[70]!==W||e[71]!==J?(Y=a.jsxs(F,{className:H,children:[W,J]}),e[68]=F,e[69]=H,e[70]=W,e[71]=J,e[72]=Y):Y=e[72],e[73]!==U||e[74]!==B||e[75]!==z||e[76]!==Y?(X=a.jsx(U,{open:B,onOpenChange:z,children:Y}),e[73]=U,e[74]=B,e[75]=z,e[76]=Y,e[77]=X):X=e[77],X},Qa=["gpt-4o","gpt-4-turbo","gpt-4","gpt-3.5-turbo","claude-3-opus-20240229","claude-3-sonnet-20240229","claude-3-haiku-20240307","claude-3-5-sonnet-20240620","gemini-pro","gemini-1.5-pro-001","gemini-1.5-flash-001","gemini-1.5-pro-001","gemini-1.5-flash-002","gemini-1.5-pro-002","gemini-1.5-flash-002"],ec="flex flex-row items-center space-x-1 space-y-0",tc=[{id:"editor",label:"Editor",Icon:fe,className:"bg-[var(--blue-4)]"},{id:"display",label:"Display",Icon:lt,className:"bg-[var(--grass-4)]"},{id:"packageManagement",label:"Package Management",Icon:ut,className:"bg-[var(--red-4)]"},{id:"runtime",label:"Runtime",Icon:st,className:"bg-[var(--amber-4)]"},{id:"ai",label:"AI",Icon:it,className:"bg-[linear-gradient(45deg,var(--purple-5),var(--cyan-5))]"},{id:"labs",label:"Labs",Icon:at,className:"bg-[var(--slate-4)]"}],nc=v(tc[0].id),rc=()=>{const e=i(38),[t,n]=ce(),r=h.useRef(null),o=ne(Xa),[s,c]=I(nc);let l,u;e[0]===Symbol.for("react.memo_cache_sentinel")?(l=we(xe),e[0]=l):l=e[0],e[1]!==t?(u={resolver:l,defaultValues:t},e[1]=t,e[2]=u):u=e[2];const d=Te(u);let f;e[3]!==n?(f=async e=>{await ee({config:e}).then((()=>{n(e)}))},e[3]=n,e[4]=f):f=e[4];const p=f;let g;e[5]===Symbol.for("react.memo_cache_sentinel")?(g=De(),e[5]=g):g=e[5];const m=g;let v;e[6]!==d?(v=()=>{const e=d.getValues("completion.copilot");return!1===e?null:"codeium"===e?a.jsxs(a.Fragment,{children:[a.jsxs("p",{className:"text-sm text-muted-secondary",children:["To get a Codeium API key, follow"," ",a.jsx("a",{className:"text-link hover:underline",href:"https://docs.marimo.io/guides/editor_features/ai_completion.html#codeium-copilot",target:"_blank",rel:"noreferrer",children:"these instructions"}),"."]}),a.jsx(je,{control:d.control,name:"completion.codeium_api_key",render:e=>{const{field:t}=e;return a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{children:"API Key"}),a.jsx(ke,{children:a.jsx(ue,{"data-testid":"codeium-api-key-input",className:"m-0 inline-flex",placeholder:"key",...t,value:t.value||""})}),a.jsx(Pe,{})]})}})]}):"github"===e?a.jsx(Ma,{}):void 0},e[6]=d,e[7]=v):v=e[7];const y=v;let b;e[8]!==s||e[9]!==d||e[10]!==p||e[11]!==c||e[12]!==o||e[13]!==y?(b=()=>{switch(s){case"editor":return a.jsxs(a.Fragment,{children:[a.jsxs(oc,{title:"Autosave",children:[a.jsx(je,{control:d.control,name:"save.autosave",render:e=>{const{field:t}=e;return a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{className:"font-normal",children:"Autosave enabled"}),a.jsx(ke,{children:a.jsx(Ee,{"data-testid":"autosave-checkbox",checked:"after_delay"===t.value,disabled:t.disabled,onCheckedChange:e=>{t.onChange(e?"after_delay":"off")}})})]})}}),a.jsx(je,{control:d.control,name:"save.autosave_delay",render:e=>{const{field:t}=e;return a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{children:"Autosave delay (seconds)"}),a.jsx(ke,{children:a.jsx(Fe,{"data-testid":"autosave-delay-input",className:"m-0 w-24",isDisabled:"after_delay"!==d.getValues("save.autosave"),...t,value:t.value/1e3,minValue:1,onChange:e=>{t.onChange(1e3*e),Number.isNaN(e)||p(d.getValues())}})}),a.jsx(Pe,{})]})}})]}),a.jsxs(oc,{title:"Formatting",children:[a.jsx(je,{control:d.control,name:"save.format_on_save",render:e=>{const{field:t}=e;return a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{className:"font-normal",children:"Format on save"}),a.jsx(ke,{children:a.jsx(Ee,{"data-testid":"format-on-save-checkbox",checked:t.value,disabled:t.disabled,onCheckedChange:e=>{t.onChange(e)}})})]})}}),a.jsx(je,{control:d.control,name:"formatting.line_length",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col space-y-1",children:[a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{children:"Line length"}),a.jsx(ke,{children:a.jsx(Fe,{"data-testid":"line-length-input",className:"m-0 w-24",...t,value:t.value,minValue:1,maxValue:1e3,onChange:e=>{t.onChange(e),Number.isNaN(e)||p(d.getValues())}})}),a.jsx(Pe,{})]}),a.jsx(Ie,{children:"Maximum line length when formatting code."})]})}})]}),a.jsx(oc,{title:"Autocomplete",children:a.jsx(je,{control:d.control,name:"completion.activate_on_typing",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col space-y-1",children:[a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{className:"font-normal",children:"Autocomplete"}),a.jsx(ke,{children:a.jsx(Ee,{"data-testid":"autocomplete-checkbox",checked:t.value,disabled:t.disabled,onCheckedChange:e=>{t.onChange(Boolean(e))}})})]}),a.jsx(Ie,{children:"When unchecked, code completion is still available through a hotkey."}),a.jsx("div",{children:a.jsx(q,{variant:"link",className:"mb-0 px-0",type:"button",onClick:e=>{e.preventDefault(),e.stopPropagation(),c("ai")},children:"Edit AI autocomplete"})})]})}})}),a.jsx(oc,{title:"Keymap",children:a.jsx(je,{control:d.control,name:"keymap.preset",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col space-y-1",children:[a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{children:"Keymap"}),a.jsx(ke,{children:a.jsx(_e,{"data-testid":"keymap-select",onChange:e=>t.onChange(e.target.value),value:t.value,disabled:t.disabled,className:"inline-flex mr-2",children:Ha.map((e=>a.jsx("option",{value:e,children:e},e)))})}),a.jsx(Pe,{})]}),a.jsx("div",{children:a.jsx(q,{variant:"link",className:"mb-0 px-0",type:"button",onClick:e=>{e.preventDefault(),e.stopPropagation(),o(!0)},children:"Edit Keyboard Shortcuts"})})]})}})})]});case"display":return a.jsxs(a.Fragment,{children:[a.jsxs(oc,{title:"Display",children:[a.jsx(je,{control:d.control,name:"display.default_width",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col space-y-1",children:[a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{children:"Default width"}),a.jsx(ke,{children:a.jsx(_e,{"data-testid":"user-config-width-select",onChange:e=>t.onChange(e.target.value),value:t.value,disabled:t.disabled,className:"inline-flex mr-2",children:$a().map((e=>a.jsx("option",{value:e,children:e},e)))})}),a.jsx(Pe,{})]}),a.jsx(Ie,{children:'The default app width for new notebooks; overridden by "width" in the application config.'})]})}}),a.jsx(je,{control:d.control,name:"display.theme",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col space-y-1",children:[a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{children:"Theme"}),a.jsx(ke,{children:a.jsx(_e,{"data-testid":"theme-select",onChange:e=>t.onChange(e.target.value),value:t.value,disabled:t.disabled,className:"inline-flex mr-2",children:Le.map((e=>a.jsx("option",{value:e,children:e},e)))})}),a.jsx(Pe,{})]}),a.jsx(Ie,{children:"This theme will be applied to the user's configuration; it does not affect theme when sharing the notebook."})]})}}),a.jsx(je,{control:d.control,name:"display.code_editor_font_size",render:e=>{const{field:t}=e;return a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{children:"Code editor font size (px)"}),a.jsx(ke,{children:a.jsx("span",{className:"inline-flex mr-2",children:a.jsx(Fe,{"data-testid":"code-editor-font-size-input",className:"m-0 w-24",...t,value:t.value,minValue:8,maxValue:32,onChange:e=>{t.onChange(e),p(d.getValues())}})})}),a.jsx(Pe,{})]})}})]}),a.jsxs(oc,{title:"Outputs",children:[a.jsx(je,{control:d.control,name:"display.cell_output",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col space-y-1",children:[a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{children:"Cell output area"}),a.jsx(ke,{children:a.jsx(_e,{"data-testid":"cell-output-select",onChange:e=>t.onChange(e.target.value),value:t.value,disabled:t.disabled,className:"inline-flex mr-2",children:["above","below"].map((e=>a.jsx("option",{value:e,children:e},e)))})}),a.jsx(Pe,{})]}),a.jsx(Ie,{children:"Where to display cell's output."})]})}}),a.jsx(je,{control:d.control,name:"display.dataframes",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col space-y-1",children:[a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{children:"Dataframe viewer"}),a.jsx(ke,{children:a.jsx(_e,{"data-testid":"display-dataframes-select",onChange:e=>t.onChange(e.target.value),value:t.value,disabled:t.disabled,className:"inline-flex mr-2",children:["rich","plain"].map((e=>a.jsx("option",{value:e,children:e},e)))})}),a.jsx(Pe,{})]}),a.jsx(Ie,{children:"Whether to use marimo's rich dataframe viewer or a plain HTML table. This requires restarting your notebook to take effect."})]})}})]})]});case"packageManagement":return a.jsx(oc,{title:"Package Management",children:a.jsx(je,{control:d.control,disabled:m,name:"package_management.manager",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col space-y-1",children:[a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{children:"Manager"}),a.jsx(ke,{children:a.jsx(_e,{"data-testid":"package-manager-select",onChange:e=>t.onChange(e.target.value),value:t.value,disabled:t.disabled,className:"inline-flex mr-2",children:Ae.map((e=>a.jsx("option",{value:e,children:e},e)))})}),a.jsx(Pe,{})]}),a.jsxs(Ie,{children:["When marimo comes across a module that is not installed, you will be prompted to install it using your preferred package manager. Learn more in the"," ",a.jsx("a",{className:"text-link hover:underline",href:"https://docs.marimo.io/guides/editor_features/package_management.html",target:"_blank",rel:"noreferrer",children:"docs"}),".",a.jsx("br",{}),a.jsx("br",{}),"Running marimo in a"," ",a.jsx("a",{className:"text-link hover:underline",href:"https://docs.marimo.io/guides/editor_features/package_management.html#running-marimo-in-a-sandbox-environment-uv-only",target:"_blank",rel:"noreferrer",children:"sandboxed environment"})," ","is only supported by ",a.jsx(Me,{className:"inline",children:"uv"})]})]})}})});case"runtime":return a.jsxs(oc,{title:"Runtime configuration",children:[a.jsx(je,{control:d.control,name:"runtime.auto_instantiate",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col gap-y-1",children:[a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{className:"font-normal",children:"Autorun on startup"}),a.jsx(ke,{children:a.jsx(Ee,{"data-testid":"auto-instantiate-checkbox",disabled:t.disabled,checked:t.value,onCheckedChange:t.onChange})})]}),a.jsx(Ie,{children:"Whether to automatically run all cells on startup."})]})}}),a.jsx(je,{control:d.control,name:"runtime.on_cell_change",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col gap-y-1",children:[a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{className:"font-normal",children:"On cell change"}),a.jsx(ke,{children:a.jsx(_e,{"data-testid":"on-cell-change-select",onChange:e=>t.onChange(e.target.value),value:t.value,className:"inline-flex mr-2",children:["lazy","autorun"].map((e=>a.jsx("option",{value:e,children:e},e)))})})]}),a.jsx(Ie,{children:'Whether marimo should automatically run cells or just mark them as stale. If "autorun", marimo will automatically run affected cells when a cell is run or a UI element is interacted with; if "lazy", marimo will mark affected cells as stale but won\'t re-run them.'})]})}}),a.jsx(je,{control:d.control,name:"runtime.auto_reload",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col gap-y-1",children:[a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{className:"font-normal",children:"On module change"}),a.jsx(ke,{children:a.jsx(_e,{"data-testid":"auto-reload-select",onChange:e=>t.onChange(e.target.value),value:t.value,disabled:m,className:"inline-flex mr-2",children:["off","lazy","autorun"].map((e=>a.jsx("option",{value:e,children:e},e)))})})]}),a.jsx(Ie,{children:'Whether marimo should automatically reload modules before executing cells. If "lazy", marimo will mark cells affected by module modifications as stale; if "autorun", affected cells will be automatically re-run.'})]})}}),a.jsxs(Ie,{children:["Learn more in the"," ",a.jsx("a",{className:"text-link hover:underline",href:"https://docs.marimo.io/guides/reactivity.html#runtime-configuration",target:"_blank",rel:"noreferrer",children:"docs"}),"."]})]});case"ai":return a.jsxs(a.Fragment,{children:[a.jsxs(oc,{title:"AI Code Completion",children:[a.jsx("p",{className:"text-sm text-muted-secondary",children:"You may use GitHub Copilot or Codeium for AI code completion."}),a.jsx(je,{control:d.control,name:"completion.copilot",render:e=>{const{field:t}=e;return a.jsx("div",{className:"flex flex-col space-y-1",children:a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{children:"Provider"}),a.jsx(ke,{children:a.jsx(_e,{"data-testid":"copilot-select",onChange:e=>{"none"===e.target.value?t.onChange(!1):t.onChange(e.target.value)},value:!0===t.value?"github":!1===t.value?"none":t.value,disabled:t.disabled,className:"inline-flex mr-2",children:["none","github","codeium"].map((e=>a.jsx("option",{value:e,children:e},e)))})}),a.jsx(Pe,{})]})})}}),y()]}),a.jsxs(oc,{title:"AI Keys",children:[a.jsx(je,{control:d.control,name:"ai.open_ai.api_key",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col space-y-1",children:[a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{children:"OpenAI API Key"}),a.jsx(ke,{children:a.jsx(ue,{"data-testid":"ai-openai-api-key-input",className:"m-0 inline-flex",placeholder:"sk-proj...",...t,onChange:e=>{const n=e.target.value;n.includes("*")||t.onChange(n)}})}),a.jsx(Pe,{})]}),a.jsxs(Ie,{children:["Your OpenAI API key from"," ",a.jsx("a",{className:"text-link hover:underline",href:"https://platform.openai.com/account/api-keys",target:"_blank",rel:"noreferrer",children:"platform.openai.com"}),"."]})]})}}),a.jsx(je,{control:d.control,name:"ai.anthropic.api_key",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col space-y-1",children:[a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{children:"Anthropic API Key"}),a.jsx(ke,{children:a.jsx(ue,{"data-testid":"ai-anthropic-api-key-input",className:"m-0 inline-flex",placeholder:"sk-ant...",...t,onChange:e=>{const n=e.target.value;n.includes("*")||t.onChange(n)}})}),a.jsx(Pe,{})]}),a.jsxs(Ie,{children:["Your Anthropic API key from"," ",a.jsx("a",{className:"text-link hover:underline",href:"https://console.anthropic.com/settings/keys",target:"_blank",rel:"noreferrer",children:"console.anthropic.com"}),"."]})]})}}),a.jsx(je,{control:d.control,name:"ai.google.api_key",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col space-y-1",children:[a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{children:"Google AI API Key"}),a.jsx(ke,{children:a.jsx(ue,{"data-testid":"ai-google-api-key-input",className:"m-0 inline-flex",placeholder:"AI...",...t,onChange:e=>{const n=e.target.value;n.includes("*")||t.onChange(n)}})}),a.jsx(Pe,{})]}),a.jsxs(Ie,{children:["Your Google AI API key from"," ",a.jsx("a",{className:"text-link hover:underline",href:"https://aistudio.google.com/app/apikey",target:"_blank",rel:"noreferrer",children:"aistudio.google.com"}),"."]})]})}})]}),a.jsxs(oc,{title:"AI Assist",children:[a.jsxs("p",{className:"text-sm text-muted-secondary",children:["Add an API key to ",a.jsx(Me,{className:"inline",children:"marimo.toml"})," to activate marimo's AI assistant; see"," ",a.jsx("a",{className:"text-link hover:underline",href:"https://docs.marimo.io/guides/editor_features/ai_completion.html",target:"_blank",rel:"noreferrer",children:"docs"})," ","for more info."]}),a.jsx(je,{control:d.control,disabled:m,name:"ai.open_ai.base_url",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col space-y-1",children:[a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{children:"Base URL"}),a.jsx(ke,{children:a.jsx(ue,{"data-testid":"ai-base-url-input",className:"m-0 inline-flex",placeholder:"https://api.openai.com/v1",...t})}),a.jsx(Pe,{})]}),a.jsx(Ie,{children:"This URL can be any OpenAI-compatible API endpoint."})]})}}),a.jsx(je,{control:d.control,disabled:m,name:"ai.open_ai.model",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col space-y-1",children:[a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{children:"Model"}),a.jsx(ke,{children:a.jsx(ue,{list:"ai-model-datalist","data-testid":"ai-model-input",className:"m-0 inline-flex",placeholder:"gpt-4-turbo",...t})}),a.jsx(Pe,{})]}),a.jsx("datalist",{id:"ai-model-datalist",children:Qa.map((e=>a.jsx("option",{value:e,children:e},e)))}),a.jsx(Ie,{children:'If the model starts with "claude-", we will use your Anthropic API key. If the model starts with "gemini-", we will use your Google AI API key. Otherwise, we will use your OpenAI API key.'})]})}}),a.jsx(je,{control:d.control,name:"ai.rules",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col",children:[a.jsxs(Ce,{children:[a.jsx(Ne,{children:"Custom Rules"}),a.jsx(ke,{children:a.jsx(qe,{"data-testid":"ai-rules-input",className:"m-0 inline-flex w-full h-32 p-2 text-sm",placeholder:"e.g. Always use type hints; prefer polars over pandas",...t,value:t.value})}),a.jsx(Pe,{})]}),a.jsx(Ie,{children:"Custom rules to include in all AI completion prompts."})]})}})]})]});case"labs":return a.jsxs(oc,{title:"Experimental Features",children:[a.jsx("p",{className:"text-sm text-muted-secondary mb-4",children:"⚠️ These features are experimental and may require restarting your notebook to take effect."}),a.jsx(je,{control:d.control,name:"experimental.multi_column",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col gap-y-1",children:[a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{className:"font-normal",children:"Multi-column layout"}),a.jsx(ke,{children:a.jsx(Ee,{"data-testid":"multi-column-checkbox",checked:!0===t.value,onCheckedChange:t.onChange})})]}),a.jsx(Ie,{children:"Enable experimental support for multi-column notebook layouts. This adds a new width setting to the notebook settings menu."})]})}}),a.jsx(je,{control:d.control,name:"experimental.chat_sidebar",render:e=>{const{field:t}=e;return a.jsxs("div",{className:"flex flex-col gap-y-1",children:[a.jsxs(Ce,{className:ec,children:[a.jsx(Ne,{className:"font-normal",children:"Chat Sidebar"}),a.jsx(ke,{children:a.jsx(Ee,{"data-testid":"chat-sidebar-checkbox",checked:!0===t.value,onCheckedChange:t.onChange})})]}),a.jsx(Ie,{children:"Enable experimental chat sidebar to ask questions with an AI assistant."})]})}})]})}},e[8]=s,e[9]=d,e[10]=p,e[11]=c,e[12]=o,e[13]=y,e[14]=b):b=e[14];const w=b;let x,T,D;e[15]===Symbol.for("react.memo_cache_sentinel")?(x=a.jsx(Me,{className:"inline",children:"marimo.toml"}),T=a.jsx("br",{}),e[15]=x,e[16]=T):(x=e[15],T=e[16]),e[17]===Symbol.for("react.memo_cache_sentinel")?(D=a.jsxs("p",{className:"text-muted-secondary",children:["User configuration is stored in ",x,T,"Run ",a.jsx(Me,{className:"inline",children:"marimo config show"})," in your terminal to show your current configuration and file location."]}),e[17]=D):D=e[17];const S=D;let O,R,j,C,N,k,P,E;return e[18]!==d||e[19]!==p?(O=d.handleSubmit(p),e[18]=d,e[19]=p,e[20]=O):O=e[20],e[21]!==c?(R=e=>c(e),e[21]=c,e[22]=R):R=e[22],e[23]===Symbol.for("react.memo_cache_sentinel")?(j=a.jsxs(Se,{className:"self-start max-h-none flex flex-col gap-2 flex-shrink-0 bg-background flex-1 min-h-full",children:[tc.map(ic),a.jsx("div",{className:"flex-1"}),!De()&&S]}),e[23]=j):j=e[23],e[24]!==s||e[25]!==R?(C=a.jsx(He,{value:s,onValueChange:R,orientation:"vertical",className:"w-1/3 pr-4 border-r h-full overflow-auto p-6",children:j}),e[24]=s,e[25]=R,e[26]=C):C=e[26],e[27]!==w?(N=w(),e[27]=w,e[28]=N):N=e[28],e[29]!==N?(k=a.jsx("div",{className:"w-2/3 pl-6 gap-2 flex flex-col overflow-auto p-6",children:N}),e[29]=N,e[30]=k):k=e[30],e[31]!==O||e[32]!==C||e[33]!==k?(P=a.jsxs("form",{ref:r,onChange:O,className:"flex text-pretty overflow-hidden",children:[C,k]}),e[31]=O,e[32]=C,e[33]=k,e[34]=P):P=e[34],e[35]!==d||e[36]!==P?(E=a.jsx(Oe,{...d,children:P}),e[35]=d,e[36]=P,e[37]=E):E=e[37],E},oc=e=>{const t=i(5),{title:n,children:r}=e;let o,s;return t[0]!==n?(o=a.jsx(Tt,{className:"text-base",children:n}),t[0]=n,t[1]=o):o=t[1],t[2]!==o||t[3]!==r?(s=a.jsxs("div",{className:"flex flex-col gap-4 pb-4",children:[o,r]}),t[2]=o,t[3]=r,t[4]=s):s=t[4],s};function ic(e){return a.jsx(Re,{value:e.id,className:"w-full text-left p-2 data-[state=active]:bg-primary data-[state=active]:text-primary-foreground justify-start",children:a.jsxs("div",{className:"flex gap-4 items-center text-lg",children:[a.jsx("span",{className:g(e.className,"w-8 h-8 rounded flex items-center justify-center text-muted-foreground"),children:a.jsx(e.Icon,{className:"w-4 h-4"})}),e.label]})},e.id)}const sc=()=>{const e=i(27),[t,n]=We();let r,o;e[0]===Symbol.for("react.memo_cache_sentinel")?(r=we(Ue),e[0]=r):r=e[0],e[1]!==t?(o={resolver:r,defaultValues:t},e[1]=t,e[2]=o):o=e[2];const s=Te(o);let c;e[3]!==n?(c=async e=>{await $e({config:e}).then((()=>{n(e)})).catch((()=>{n(e)}))},e[3]=n,e[4]=c):c=e[4];const l=c;let u,d,f,p,g,m,v,y,b,w;return e[5]!==t.width?(u=[t.width],e[5]=t.width,e[6]=u):u=e[6],h.useEffect(ac,u),e[7]!==s||e[8]!==l?(d=s.handleSubmit(l),e[7]=s,e[8]=l,e[9]=d):d=e[9],e[10]===Symbol.for("react.memo_cache_sentinel")?(f=a.jsxs("div",{children:[a.jsx(xt,{children:"Application Config"}),a.jsx(Dt,{children:"Settings applied to this notebook"})]}),e[10]=f):f=e[10],e[11]!==s.control?(p=a.jsx(je,{control:s.control,name:"width",render:lc}),g=a.jsx(je,{control:s.control,name:"app_title",render:uc}),m=a.jsx(je,{control:s.control,name:"css_file",render:dc}),v=a.jsx(je,{control:s.control,name:"html_head_file",render:fc}),y=a.jsx(je,{control:s.control,name:"auto_download",render:hc}),e[11]=s.control,e[12]=p,e[13]=g,e[14]=m,e[15]=v,e[16]=y):(p=e[12],g=e[13],m=e[14],v=e[15],y=e[16]),e[17]!==d||e[18]!==p||e[19]!==g||e[20]!==m||e[21]!==v||e[22]!==y?(b=a.jsxs("form",{onChange:d,className:"flex flex-col gap-4",children:[f,p,g,m,v,y]}),e[17]=d,e[18]=p,e[19]=g,e[20]=m,e[21]=v,e[22]=y,e[23]=b):b=e[23],e[24]!==s||e[25]!==b?(w=a.jsx(Oe,{...s,children:b}),e[24]=s,e[25]=b,e[26]=w):w=e[26],w};function ac(){window.dispatchEvent(new Event("resize"))}function cc(e){return a.jsx("option",{value:e,children:e},e)}function lc(e){const{field:t}=e;return a.jsxs(Ce,{className:"flex flex-row items-center space-x-1 space-y-0",children:[a.jsx(Ne,{children:"Width"}),a.jsx(ke,{children:a.jsx(_e,{"data-testid":"app-width-select",onChange:e=>t.onChange(e.target.value),value:t.value,disabled:t.disabled,className:"inline-flex mr-2",children:$a().map(cc)})}),a.jsx(Pe,{})]})}function uc(e){const{field:t}=e;return a.jsxs("div",{className:"flex flex-col gap-y-1",children:[a.jsxs(Ce,{className:"flex flex-row items-center space-x-1 space-y-0",children:[a.jsx(Ne,{children:"App title"}),a.jsx(ke,{children:a.jsx(ue,{...t,value:t.value??"",onChange:e=>{t.onChange(e.target.value),Be.safeParse(e.target.value).success&&(document.title=e.target.value)}})}),a.jsx(Pe,{})]}),a.jsx(Ie,{children:"The application title is put in the title tag in the HTML code and typically displayed in the title bar of the browser window."})]})}function dc(e){const{field:t}=e;return a.jsxs("div",{className:"flex flex-col gap-y-1",children:[a.jsxs(Ce,{className:"flex flex-row items-center space-x-1 space-y-0",children:[a.jsx(Ne,{className:"flex-shrink-0",children:"Custom CSS"}),a.jsx(ke,{children:a.jsx(ue,{...t,value:t.value??"",placeholder:"custom.css",onChange:e=>{t.onChange(e.target.value),Be.safeParse(e.target.value).success&&(document.title=e.target.value)}})}),a.jsx(Pe,{})]}),a.jsx(Ie,{children:"A filepath to a custom css file to be injected into the notebook."})]})}function fc(e){const{field:t}=e;return a.jsxs("div",{className:"flex flex-col gap-y-1",children:[a.jsxs(Ce,{className:"flex flex-row items-center space-x-1 space-y-0",children:[a.jsx(Ne,{className:"flex-shrink-0",children:"HTML Head"}),a.jsx(ke,{children:a.jsx(ue,{...t,value:t.value??"",placeholder:"head.html",onChange:e=>{t.onChange(e.target.value)}})}),a.jsx(Pe,{})]}),a.jsxs(Ie,{children:["A filepath to an HTML file to be injected into the"," ",a.jsx(Me,{className:"inline",children:"<head/>"})," section of the notebook. Use this to add analytics, custom fonts, meta tags, or external scripts."]})]})}function hc(e){const{field:t}=e;return a.jsxs("div",{className:"flex flex-col gap-y-1",children:[a.jsxs(Ce,{className:"flex flex-col gap-2",children:[a.jsx(ke,{children:a.jsx("div",{className:"flex gap-2",children:a.jsxs("div",{className:"flex items-center space-x-2",children:[a.jsx(Ee,{id:"html-checkbox",checked:t.value.includes("html"),onCheckedChange:()=>{t.onChange(ze(t.value,"html"))}}),a.jsx(Ne,{htmlFor:"html-checkbox",children:"Auto-download HTML"})]})})}),a.jsx(Pe,{})]}),a.jsxs(Ie,{children:["When enabled, marimo will periodically save this notebook as HTML to a folder ",a.jsx(Me,{className:"inline",children:"__marimo__"})," in the notebook's directory."]})]})}const pc=v(!1),gc=e=>{const t=i(20),{showAppConfig:n}=e,r=void 0===n||n,[o,s]=I(pc);let c;t[0]===Symbol.for("react.memo_cache_sentinel")?(c=a.jsx(mt,{"aria-label":"Config","data-testid":"app-config-button",shape:"circle",size:"small",className:"h-[27px] w-[27px]",color:"hint-green",children:a.jsx(d,{content:"Settings",children:a.jsx(Ve,{strokeWidth:1.8})})}),t[0]=c):c=t[0];const l=c;let u;t[1]===Symbol.for("react.memo_cache_sentinel")?(u=a.jsx(ve,{className:"w-[80vw] h-[70vh] overflow-hidden sm:max-w-5xl top-[15vh] p-0",children:a.jsx(rc,{})}),t[1]=u):u=t[1];const f=u;if(!r){let e,n;return t[2]===Symbol.for("react.memo_cache_sentinel")?(e=a.jsx(Ke,{children:l}),t[2]=e):e=t[2],t[3]!==o||t[4]!==s?(n=a.jsxs(pe,{open:o,onOpenChange:s,children:[e,f]}),t[3]=o,t[4]=s,t[5]=n):n=t[5],n}let h,p,g,m,v,y,b,w;return t[6]===Symbol.for("react.memo_cache_sentinel")?(h=a.jsx(Ge,{asChild:!0,children:l}),t[6]=h):h=t[6],t[7]===Symbol.for("react.memo_cache_sentinel")?(p=a.jsx(sc,{}),g=a.jsx("div",{className:"h-px bg-border my-2"}),t[7]=p,t[8]=g):(p=t[7],g=t[8]),t[9]!==s?(m=()=>s(!0),t[9]=s,t[10]=m):m=t[10],t[11]===Symbol.for("react.memo_cache_sentinel")?(v=a.jsx(Ve,{strokeWidth:1.8,className:"w-4 h-4 mr-2"}),t[11]=v):v=t[11],t[12]!==m?(y=a.jsxs(Je,{children:[h,a.jsxs(Ye,{className:"w-80 overflow-auto",align:"end",side:"bottom",onFocusOutside:mc,children:[p,g,a.jsxs(q,{onClick:m,variant:"link",className:"px-0",children:[v,"User settings"]})]})]}),t[12]=m,t[13]=y):y=t[13],t[14]!==o||t[15]!==s?(b=a.jsx(pe,{open:o,onOpenChange:s,children:f}),t[14]=o,t[15]=s,t[16]=b):b=t[16],t[17]!==y||t[18]!==b?(w=a.jsxs(a.Fragment,{children:[y,b]}),t[17]=y,t[18]=b,t[19]=w):w=t[19],w};function mc(e){return e.preventDefault()}const vc=e=>a.jsx("svg",{xmlns:"http://www.w3.org/2000/svg",width:"1em",height:"1em",fill:"currentColor",viewBox:"0 0 448 512",...e,children:a.jsx("path",{d:"M439.8 200.5c-7.7-30.9-22.3-54.2-53.4-54.2h-40.1v47.4c0 36.8-31.2 67.8-66.8 67.8H172.7c-29.2 0-53.4 25-53.4 54.3v101.8c0 29 25.2 46 53.4 54.3 33.8 9.9 66.3 11.7 106.8 0 26.9-7.8 53.4-23.5 53.4-54.3v-40.7H226.2v-13.6h160.2c31.1 0 42.6-21.7 53.4-54.2 11.2-33.5 10.7-65.7 0-108.6zM286.2 404c11.1 0 20.1 9.1 20.1 20.3 0 11.3-9 20.4-20.1 20.4-11 0-20.1-9.2-20.1-20.4.1-11.3 9.1-20.3 20.1-20.3zM167.8 248.1h106.8c29.7 0 53.4-24.5 53.4-54.3V91.9c0-29-24.4-50.7-53.4-55.6-35.8-5.9-74.7-5.6-106.8.1-45.2 8-53.4 24.7-53.4 55.6v40.7h106.9v13.6h-147c-31.1 0-58.3 18.7-66.8 54.2-9.8 40.7-10.2 66.1 0 108.6 7.6 31.6 25.7 54.2 56.8 54.2H101v-48.8c0-35.3 30.5-66.4 66.8-66.4zm-6.7-142.6c-11.1 0-20.1-9.1-20.1-20.3.1-11.3 9-20.4 20.1-20.4 11 0 20.1 9.2 20.1 20.4s-9 20.3-20.1 20.3z"})}),yc=e=>a.jsx("svg",{xmlns:"http://www.w3.org/2000/svg",width:"1em",height:"1em",viewBox:"0 0 640 512",fill:"currentColor",...e,children:a.jsx("path",{d:"M593.8 59.1H46.2C20.7 59.1 0 79.8 0 105.2v301.5c0 25.5 20.7 46.2 46.2 46.2h547.7c25.5 0 46.2-20.7 46.1-46.1V105.2c0-25.4-20.7-46.1-46.2-46.1zM338.5 360.6H277v-120l-61.5 76.9-61.5-76.9v120H92.3V151.4h61.5l61.5 76.9 61.5-76.9h61.5v209.2zm135.3 3.1L381.5 256H443V151.4h61.5V256H566z"})}),bc=h.createContext(null);function wc(){const e=h.useContext(bc);if(null===e)throw new Error("No Tree Api Provided");return e}const xc=h.createContext(null);const Tc=h.createContext(null);const Dc=h.createContext(0);function Sc(){h.useContext(Dc)}function Oc(e,t,n){return Math.max(Math.min(e,n),t)}function Rc(e){return e&&e.isLeaf}function jc(e){return e&&e.isInternal&&!e.isOpen}function Cc(e){var t;return e&&e.isOpen&&!(null===(t=e.children)||void 0===t?void 0:t.length)}const Nc=(e,t)=>{let n=e;for(;n;){if(n.id===t.id)return!0;n=n.parent}return!1},kc=e=>{if(!e.parent)throw Error("Node does not have a parent");return e.parent.children.findIndex((t=>t.id===e.id))};function Pc(e,t){if(!e)return null;if(e.id===t)return e;if(e.children)for(let n of e.children){const e=Pc(n,t);if(e)return e}return null}function Ec(e,t){if(t(e),e.children)for(let n of e.children)Ec(n,t)}function Ic(e){const t=Ac(e);let n;for(let r=0;r<t.length;++r){if(t[r]===e){n=Mc(t,r);break}}null==n||n.focus()}function _c(e){const t=Ac(e);let n;for(let r=0;r<t.length;++r){if(t[r]===e){n=qc(t,r);break}}null==n||n.focus()}function Mc(e,t){return t+1<e.length?e[t+1]:e[0]}function qc(e,t){return t-1>=0?e[t-1]:e[e.length-1]}function Ac(e){return Array.from(document.querySelectorAll('button:not([disabled]), [href], input:not([disabled]), select:not([disabled]), textarea:not([disabled]), [tabindex]:not([tabindex="-1"]):not([disabled]), details:not([disabled]), summary:not(:disabled)')).filter((t=>t===e||!e.contains(t)))}function Lc(e,t){return"boolean"==typeof t?t:"string"==typeof t?e[t]:t(e)}function Fc(e){return null===e?null:Hc(e)}function Hc(e){return"string"==typeof e?e:e.id}function Wc(e,...t){if(e)return e(...t)}function Uc(e){return new Promise(((t,n)=>{let r=0;!function o(){r+=1,100===r&&n(),e()?t():setTimeout(o,10)}()}))}function Bc(e){var t,n;const r=e.focusedNode;return r?r.isOpen?0:r.parent?r.childIndex+1:0:null!==(n=null===(t=e.root.children)||void 0===t?void 0:t.length)&&void 0!==n?n:0}function zc(e){const t=e.focusedNode;return t?t.isOpen?t.id:t.parent&&!t.parent.isRoot?t.parent.id:null:null}const $c=Object.freeze(Object.defineProperty({__proto__:null,access:Lc,bound:Oc,dfs:Pc,focusNextElement:Ic,focusPrevElement:_c,getInsertIndex:Bc,getInsertParentId:zc,identify:Hc,identifyNull:Fc,indexOf:kc,isClosed:jc,isDescendant:Nc,isItem:Rc,isOpenWithEmptyChildren:Cc,mergeRefs:function(...e){return t=>{e.forEach((e=>{"function"==typeof e?e(t):null!=e&&(e.current=t)}))}},noop:function(){},safeRun:Wc,waitFor:Uc,walk:Ec},Symbol.toStringTag,{value:"Module"})),Vc={display:"flex",alignItems:"center",zIndex:1},Kc={flex:1,height:"2px",background:"#4B91E2",borderRadius:"1px"},Gc={width:"4px",height:"4px",boxShadow:"0 0 0 3px #4B91E2",borderRadius:"50%"},Jc=o.memo((function({top:e,left:t,indent:n}){const r={position:"absolute",pointerEvents:"none",top:e-2+"px",left:t+"px",right:n+"px"};return a.jsxs("div",{style:Object.assign(Object.assign({},Vc),r),children:[a.jsx("div",{style:Object.assign({},Gc)}),a.jsx("div",{style:Object.assign({},Kc)})]})}));function Yc({node:e,attrs:t,innerRef:n,children:r}){return a.jsx("div",Object.assign({},t,{ref:n,onFocus:e=>e.stopPropagation(),onClick:e.handleClick,children:r}))}function Xc(e){return a.jsxs("div",{ref:e.dragHandle,style:e.style,children:[a.jsx("span",{onClick:t=>{t.stopPropagation(),e.node.toggle()},children:e.node.isLeaf?"🌳":e.node.isOpen?"🗁":"🗀"})," ",e.node.isEditing?a.jsx(Qc,Object.assign({},e)):a.jsx(Zc,Object.assign({},e))]})}function Zc(e){return a.jsx(a.Fragment,{children:a.jsx("span",{children:e.node.data.name})})}function Qc({node:e}){const t=h.useRef();return h.useEffect((()=>{var e,n;null===(e=t.current)||void 0===e||e.focus(),null===(n=t.current)||void 0===n||n.select()}),[]),a.jsx("input",{ref:t,defaultValue:e.data.name,onBlur:()=>e.reset(),onKeyDown:n=>{var r;"Escape"===n.key&&e.reset(),"Enter"===n.key&&e.submit((null===(r=t.current)||void 0===r?void 0:r.value)||"")}})}function el(e){return{type:"EDIT",id:e}}function tl(e){return{type:"FOCUS",id:e}}class nl{constructor(e){this.handleClick=e=>{e.metaKey&&!this.tree.props.disableMultiSelection?this.isSelected?this.deselect():this.selectMulti():e.shiftKey&&!this.tree.props.disableMultiSelection?this.selectContiguous():(this.select(),this.activate())},this.tree=e.tree,this.id=e.id,this.data=e.data,this.level=e.level,this.children=e.children,this.parent=e.parent,this.isDraggable=e.isDraggable,this.rowIndex=e.rowIndex}get isRoot(){return this.id===rl}get isLeaf(){return!Array.isArray(this.children)}get isInternal(){return!this.isLeaf}get isOpen(){return!this.isLeaf&&this.tree.isOpen(this.id)}get isClosed(){return!this.isLeaf&&!this.tree.isOpen(this.id)}get isEditable(){return this.tree.isEditable(this.data)}get isEditing(){return this.tree.editingId===this.id}get isSelected(){return this.tree.isSelected(this.id)}get isOnlySelection(){return this.isSelected&&this.tree.hasOneSelection}get isSelectedStart(){var e;return this.isSelected&&!(null===(e=this.prev)||void 0===e?void 0:e.isSelected)}get isSelectedEnd(){var e;return this.isSelected&&!(null===(e=this.next)||void 0===e?void 0:e.isSelected)}get isFocused(){return this.tree.isFocused(this.id)}get isDragging(){return this.tree.isDragging(this.id)}get willReceiveDrop(){return this.tree.willReceiveDrop(this.id)}get state(){return{isClosed:this.isClosed,isDragging:this.isDragging,isEditing:this.isEditing,isFocused:this.isFocused,isInternal:this.isInternal,isLeaf:this.isLeaf,isOpen:this.isOpen,isSelected:this.isSelected,isSelectedEnd:this.isSelectedEnd,isSelectedStart:this.isSelectedStart,willReceiveDrop:this.willReceiveDrop}}get childIndex(){return this.parent&&this.parent.children?this.parent.children.findIndex((e=>e.id===this.id)):-1}get next(){return null===this.rowIndex?null:this.tree.at(this.rowIndex+1)}get prev(){return null===this.rowIndex?null:this.tree.at(this.rowIndex-1)}get nextSibling(){var e,t;const n=this.childIndex;return null!==(t=null===(e=this.parent)||void 0===e?void 0:e.children[n+1])&&void 0!==t?t:null}isAncestorOf(e){if(!e)return!1;let t=e;for(;t;){if(t.id===this.id)return!0;t=t.parent}return!1}select(){this.tree.select(this)}deselect(){this.tree.deselect(this)}selectMulti(){this.tree.selectMulti(this)}selectContiguous(){this.tree.selectContiguous(this)}activate(){this.tree.activate(this)}focus(){this.tree.focus(this)}toggle(){this.tree.toggle(this)}open(){this.tree.open(this)}openParents(){this.tree.openParents(this)}close(){this.tree.close(this)}submit(e){this.tree.submit(this,e)}reset(){this.tree.reset()}clone(){return new nl(Object.assign({},this))}edit(){return this.tree.edit(this)}}const rl="__REACT_ARBORIST_INTERNAL_ROOT__";function ol(e){var t;function n(t,r,o){const i=e.accessId(t),s=new nl({tree:e,data:t,level:r,parent:o,id:i,children:null,isDraggable:e.isDraggable(t),rowIndex:null}),a=e.accessChildren(t);return a&&(s.children=a.map((e=>n(e,r+1,s)))),s}const r=new nl({tree:e,id:rl,data:{id:rl},level:-1,parent:null,children:null,isDraggable:!0,rowIndex:null}),o=null!==(t=e.props.data)&&void 0!==t?t:[];return r.children=o.map((e=>n(e,0,r))),r}const il={open:(e,t)=>({type:"VISIBILITY_OPEN",id:e,filtered:t}),close:(e,t)=>({type:"VISIBILITY_CLOSE",id:e,filtered:t}),toggle:(e,t)=>({type:"VISIBILITY_TOGGLE",id:e,filtered:t}),clear:e=>({type:"VISIBILITY_CLEAR",filtered:e})};function sl(e={},t){if("VISIBILITY_OPEN"===t.type)return Object.assign(Object.assign({},e),{[t.id]:!0});if("VISIBILITY_CLOSE"===t.type)return Object.assign(Object.assign({},e),{[t.id]:!1});if("VISIBILITY_TOGGLE"===t.type){const n=e[t.id];return Object.assign(Object.assign({},e),{[t.id]:!n})}return"VISIBILITY_CLEAR"===t.type?{}:e}const al=e=>{var t;return{nodes:{open:{filtered:{},unfiltered:null!==(t=null==e?void 0:e.initialOpenState)&&void 0!==t?t:{}},focus:{id:null,treeFocused:!1},edit:{id:null},drag:{id:null,selectedIds:[],destinationParentId:null,destinationIndex:null},selection:{ids:new Set,anchor:null,mostRecent:null}},dnd:{cursor:{type:"none"},dragId:null,dragIds:[],parentId:null,index:-1}}},cl=e=>({type:"SELECTION_ONLY",id:Hc(e)}),ll=e=>({type:"SELECTION_ADD",ids:(Array.isArray(e)?e:[e]).map(Hc)}),ul=e=>({type:"SELECTION_REMOVE",ids:(Array.isArray(e)?e:[e]).map(Hc)}),dl=e=>Object.assign({type:"SELECTION_SET"},e),fl=e=>({type:"SELECTION_MOST_RECENT",id:null===e?null:Hc(e)}),hl=e=>({type:"SELECTION_ANCHOR",id:null===e?null:Hc(e)});const pl={cursor:e=>({type:"DND_CURSOR",cursor:e}),dragStart:(e,t)=>({type:"DND_DRAG_START",id:e,dragIds:t}),dragEnd:()=>({type:"DND_DRAG_END"}),hovering:(e,t)=>({type:"DND_HOVERING",parentId:e,index:t})};const gl={position:"fixed",pointerEvents:"none",zIndex:100,left:0,top:0,width:"100%",height:"100%"},ml=e=>{if(!e)return{display:"none"};const{x:t,y:n}=e;return{transform:`translate(${t}px, ${n}px)`}},vl=e=>{if(!e)return{display:"none"};const{x:t,y:n}=e;return{transform:`translate(${t+10}px, ${n+10}px)`}};function yl({offset:e,mouse:t,id:n,dragIds:r,isDragging:o}){return a.jsxs(bl,{isDragging:o,children:[a.jsx(wl,{offset:e,children:a.jsx(Tl,{id:n,dragIds:r})}),a.jsx(xl,{mouse:t,count:r.length})]})}const bl=h.memo((function(e){return e.isDragging?a.jsx("div",{style:gl,children:e.children}):null}));function wl(e){return a.jsx("div",{className:"row preview",style:ml(e.offset),children:e.children})}function xl(e){const{count:t,mouse:n}=e;return t>1?a.jsx("div",{className:"selected-count",style:vl(n),children:t}):null}const Tl=h.memo((function(e){const t=wc(),n=t.get(e.id);return n?a.jsx(t.renderNode,{preview:!0,node:n,style:{paddingLeft:n.level*t.indent,opacity:.2,background:"transparent"},tree:t}):null}));var Dl=Number.isNaN||function(e){return"number"==typeof e&&e!=e};function Sl(e,t){if(e.length!==t.length)return!1;for(var n=0;n<e.length;n++)if(r=e[n],o=t[n],!(r===o||Dl(r)&&Dl(o)))return!1;var r,o;return!0}function Ol(e,t){var n;void 0===t&&(t=Sl);var r,o=[],i=!1;return function(){for(var s=[],a=0;a<arguments.length;a++)s[a]=arguments[a];return i&&n===this&&t(s,o)||(r=e.apply(this,s),i=!0,n=this,o=s),r}}var Rl="object"==typeof performance&&"function"==typeof performance.now?function(){return performance.now()}:function(){return Date.now()};function jl(e){cancelAnimationFrame(e.id)}var Cl=-1;function Nl(e){if(void 0===e&&(e=!1),-1===Cl||e){var t=document.createElement("div"),n=t.style;n.width="50px",n.height="50px",n.overflow="scroll",document.body.appendChild(t),Cl=t.offsetWidth-t.clientWidth,document.body.removeChild(t)}return Cl}var kl=null;function Pl(e){if(void 0===e&&(e=!1),null===kl||e){var t=document.createElement("div"),n=t.style;n.width="50px",n.height="50px",n.overflow="scroll",n.direction="rtl";var r=document.createElement("div"),o=r.style;return o.width="100px",o.height="100px",t.appendChild(r),document.body.appendChild(t),t.scrollLeft>0?kl="positive-descending":(t.scrollLeft=1,kl=0===t.scrollLeft?"negative":"positive-ascending"),document.body.removeChild(t),kl}return kl}var El=function(e,t){return e};function Il(e){var t,n=e.getItemOffset,r=e.getEstimatedTotalSize,o=e.getItemSize,i=e.getOffsetForIndexAndAlignment,s=e.getStartIndexForOffset,a=e.getStopIndexForStartIndex,c=e.initInstanceProps,l=e.shouldResetStyleCacheOnItemSizeChange,u=e.validateProps;return(t=function(e){var t,d;function f(t){var r;return(r=e.call(this,t)||this)._instanceProps=c(r.props,ht(r)),r._outerRef=void 0,r._resetIsScrollingTimeoutId=null,r.state={instance:ht(r),isScrolling:!1,scrollDirection:"forward",scrollOffset:"number"==typeof r.props.initialScrollOffset?r.props.initialScrollOffset:0,scrollUpdateWasRequested:!1},r._callOnItemsRendered=void 0,r._callOnItemsRendered=Ol((function(e,t,n,o){return r.props.onItemsRendered({overscanStartIndex:e,overscanStopIndex:t,visibleStartIndex:n,visibleStopIndex:o})})),r._callOnScroll=void 0,r._callOnScroll=Ol((function(e,t,n){return r.props.onScroll({scrollDirection:e,scrollOffset:t,scrollUpdateWasRequested:n})})),r._getItemStyle=void 0,r._getItemStyle=function(e){var t,i=r.props,s=i.direction,a=i.itemSize,c=i.layout,u=r._getItemStyleCache(l&&a,l&&c,l&&s);if(u.hasOwnProperty(e))t=u[e];else{var d=n(r.props,e,r._instanceProps),f=o(r.props,e,r._instanceProps),h="horizontal"===s||"horizontal"===c,p="rtl"===s,g=h?d:0;u[e]=t={position:"absolute",left:p?void 0:g,right:p?g:void 0,top:h?0:d,height:h?"100%":f,width:h?f:"100%"}}return t},r._getItemStyleCache=void 0,r._getItemStyleCache=Ol((function(e,t,n){return{}})),r._onScrollHorizontal=function(e){var t=e.currentTarget,n=t.clientWidth,o=t.scrollLeft,i=t.scrollWidth;r.setState((function(e){if(e.scrollOffset===o)return null;var t=r.props.direction,s=o;if("rtl"===t)switch(Pl()){case"negative":s=-o;break;case"positive-descending":s=i-n-o}return s=Math.max(0,Math.min(s,i-n)),{isScrolling:!0,scrollDirection:e.scrollOffset<s?"forward":"backward",scrollOffset:s,scrollUpdateWasRequested:!1}}),r._resetIsScrollingDebounced)},r._onScrollVertical=function(e){var t=e.currentTarget,n=t.clientHeight,o=t.scrollHeight,i=t.scrollTop;r.setState((function(e){if(e.scrollOffset===i)return null;var t=Math.max(0,Math.min(i,o-n));return{isScrolling:!0,scrollDirection:e.scrollOffset<t?"forward":"backward",scrollOffset:t,scrollUpdateWasRequested:!1}}),r._resetIsScrollingDebounced)},r._outerRefSetter=function(e){var t=r.props.outerRef;r._outerRef=e,"function"==typeof t?t(e):null!=t&&"object"==typeof t&&t.hasOwnProperty("current")&&(t.current=e)},r._resetIsScrollingDebounced=function(){var e,t,n,o;null!==r._resetIsScrollingTimeoutId&&jl(r._resetIsScrollingTimeoutId),r._resetIsScrollingTimeoutId=(e=r._resetIsScrolling,t=150,n=Rl(),o={id:requestAnimationFrame((function r(){Rl()-n>=t?e.call(null):o.id=requestAnimationFrame(r)}))})},r._resetIsScrolling=function(){r._resetIsScrollingTimeoutId=null,r.setState({isScrolling:!1},(function(){r._getItemStyleCache(-1,null)}))},r}d=e,(t=f).prototype=Object.create(d.prototype),t.prototype.constructor=t,ft(t,d),f.getDerivedStateFromProps=function(e,t){return _l(e,t),u(e),null};var p=f.prototype;return p.scrollTo=function(e){e=Math.max(0,e),this.setState((function(t){return t.scrollOffset===e?null:{scrollDirection:t.scrollOffset<e?"forward":"backward",scrollOffset:e,scrollUpdateWasRequested:!0}}),this._resetIsScrollingDebounced)},p.scrollToItem=function(e,t){void 0===t&&(t="auto");var n=this.props,r=n.itemCount,o=n.layout,s=this.state.scrollOffset;e=Math.max(0,Math.min(e,r-1));var a=0;if(this._outerRef){var c=this._outerRef;a="vertical"===o?c.scrollWidth>c.clientWidth?Nl():0:c.scrollHeight>c.clientHeight?Nl():0}this.scrollTo(i(this.props,e,t,s,this._instanceProps,a))},p.componentDidMount=function(){var e=this.props,t=e.direction,n=e.initialScrollOffset,r=e.layout;if("number"==typeof n&&null!=this._outerRef){var o=this._outerRef;"horizontal"===t||"horizontal"===r?o.scrollLeft=n:o.scrollTop=n}this._callPropsCallbacks()},p.componentDidUpdate=function(){var e=this.props,t=e.direction,n=e.layout,r=this.state,o=r.scrollOffset;if(r.scrollUpdateWasRequested&&null!=this._outerRef){var i=this._outerRef;if("horizontal"===t||"horizontal"===n)if("rtl"===t)switch(Pl()){case"negative":i.scrollLeft=-o;break;case"positive-ascending":i.scrollLeft=o;break;default:var s=i.clientWidth,a=i.scrollWidth;i.scrollLeft=a-s-o}else i.scrollLeft=o;else i.scrollTop=o}this._callPropsCallbacks()},p.componentWillUnmount=function(){null!==this._resetIsScrollingTimeoutId&&jl(this._resetIsScrollingTimeoutId)},p.render=function(){var e=this.props,t=e.children,n=e.className,o=e.direction,i=e.height,s=e.innerRef,a=e.innerElementType,c=e.innerTagName,l=e.itemCount,u=e.itemData,d=e.itemKey,f=void 0===d?El:d,p=e.layout,g=e.outerElementType,m=e.outerTagName,v=e.style,y=e.useIsScrolling,b=e.width,w=this.state.isScrolling,x="horizontal"===o||"horizontal"===p,T=x?this._onScrollHorizontal:this._onScrollVertical,D=this._getRangeToRender(),S=D[0],O=D[1],R=[];if(l>0)for(var j=S;j<=O;j++)R.push(h.createElement(t,{data:u,key:f(j,u),index:j,isScrolling:y?w:void 0,style:this._getItemStyle(j)}));var C=r(this.props,this._instanceProps);return h.createElement(g||m||"div",{className:n,onScroll:T,ref:this._outerRefSetter,style:Xe({position:"relative",height:i,width:b,overflow:"auto",WebkitOverflowScrolling:"touch",willChange:"transform",direction:o},v)},h.createElement(a||c||"div",{children:R,ref:s,style:{height:x?"100%":C,pointerEvents:w?"none":void 0,width:x?C:"100%"}}))},p._callPropsCallbacks=function(){if("function"==typeof this.props.onItemsRendered&&this.props.itemCount>0){var e=this._getRangeToRender(),t=e[0],n=e[1],r=e[2],o=e[3];this._callOnItemsRendered(t,n,r,o)}if("function"==typeof this.props.onScroll){var i=this.state,s=i.scrollDirection,a=i.scrollOffset,c=i.scrollUpdateWasRequested;this._callOnScroll(s,a,c)}},p._getRangeToRender=function(){var e=this.props,t=e.itemCount,n=e.overscanCount,r=this.state,o=r.isScrolling,i=r.scrollDirection,c=r.scrollOffset;if(0===t)return[0,0,0,0];var l=s(this.props,c,this._instanceProps),u=a(this.props,l,c,this._instanceProps),d=o&&"backward"!==i?1:Math.max(1,n),f=o&&"forward"!==i?1:Math.max(1,n);return[Math.max(0,l-d),Math.max(0,Math.min(t-1,u+f)),l,u]},f}(h.PureComponent)).defaultProps={direction:"ltr",itemData:void 0,layout:"vertical",overscanCount:2,useIsScrolling:!1},t}var _l=function(e,t){e.children,e.direction,e.height,e.layout,e.innerTagName,e.outerTagName,e.width,t.instance},Ml=Il({getItemOffset:function(e,t){return t*e.itemSize},getItemSize:function(e,t){return e.itemSize},getEstimatedTotalSize:function(e){var t=e.itemCount;return e.itemSize*t},getOffsetForIndexAndAlignment:function(e,t,n,r,o,i){var s=e.direction,a=e.height,c=e.itemCount,l=e.itemSize,u=e.layout,d=e.width,f="horizontal"===s||"horizontal"===u?d:a,h=Math.max(0,c*l-f),p=Math.min(h,t*l),g=Math.max(0,t*l-f+l+i);switch("smart"===n&&(n=r>=g-f&&r<=p+f?"auto":"center"),n){case"start":return p;case"end":return g;case"center":var m=Math.round(g+(p-g)/2);return m<Math.ceil(f/2)?0:m>h+Math.floor(f/2)?h:m;default:return r>=g&&r<=p?r:r<g?g:p}},getStartIndexForOffset:function(e,t){var n=e.itemCount,r=e.itemSize;return Math.max(0,Math.min(n-1,Math.floor(t/r)))},getStopIndexForStartIndex:function(e,t,n){var r=e.direction,o=e.height,i=e.itemCount,s=e.itemSize,a=e.layout,c=e.width,l=t*s,u="horizontal"===r||"horizontal"===a?c:o,d=Math.ceil((u+n-l)/s);return Math.max(0,Math.min(i-1,t+d-1))},initInstanceProps:function(e){},shouldResetStyleCacheOnItemSizeChange:!0,validateProps:function(e){e.itemSize}});function ql(){var e,t;const n=wc(),r=function(){const e=h.useContext(Tc);if(null===e)throw new Error("Provide a DnDContext");return e}().cursor;if(!r||"line"!==r.type)return null;const o=n.indent,i=n.rowHeight*r.index+(null!==(t=null!==(e=n.props.padding)&&void 0!==e?e:n.props.paddingTop)&&void 0!==t?t:0),s=o*r.level,c=n.renderCursor;return a.jsx(c,{top:i,left:s,indent:o})}const Al=h.forwardRef((function(e,t){const{children:n}=e,r=function(e,t){var n={};for(var r in e)Object.prototype.hasOwnProperty.call(e,r)&&t.indexOf(r)<0&&(n[r]=e[r]);if(null!=e&&"function"==typeof Object.getOwnPropertySymbols){var o=0;for(r=Object.getOwnPropertySymbols(e);o<r.length;o++)t.indexOf(r[o])<0&&Object.prototype.propertyIsEnumerable.call(e,r[o])&&(n[r[o]]=e[r[o]])}return n}(e,["children"]),o=wc();return a.jsxs("div",Object.assign({ref:t},r,{onClick:e=>{e.currentTarget===e.target&&o.deselectAll()},children:[a.jsx(Ll,{}),n]}))})),Ll=()=>{const e=wc();return a.jsx("div",{style:{height:e.visibleNodes.length*e.rowHeight,width:"100%",position:"absolute",left:"0",right:"0"},children:a.jsx(ql,{})})};const Fl=h.forwardRef((function(e,t){var n,r,o,i,{style:s}=e,c=function(e,t){var n={};for(var r in e)Object.prototype.hasOwnProperty.call(e,r)&&t.indexOf(r)<0&&(n[r]=e[r]);if(null!=e&&"function"==typeof Object.getOwnPropertySymbols){var o=0;for(r=Object.getOwnPropertySymbols(e);o<r.length;o++)t.indexOf(r[o])<0&&Object.prototype.propertyIsEnumerable.call(e,r[o])&&(n[r[o]]=e[r[o]])}return n}(e,["style"]);const l=wc(),u=null!==(r=null!==(n=l.props.padding)&&void 0!==n?n:l.props.paddingTop)&&void 0!==r?r:0,d=null!==(i=null!==(o=l.props.padding)&&void 0!==o?o:l.props.paddingBottom)&&void 0!==i?i:0;return a.jsx("div",Object.assign({ref:t,style:Object.assign(Object.assign({},s),{height:`${parseFloat(s.height)+u+d}px`})},c))}));var Hl,Wl,Ul=h.createContext({dragDropManager:void 0});function Bl(e,t){for(var n=arguments.length,r=new Array(n>2?n-2:0),o=2;o<n;o++)r[o-2]=arguments[o];if(!e){var i;if(void 0===t)i=new Error("Minified exception occurred; use the non-minified dev environment for the full error message and additional helpful warnings.");else{var s=0;(i=new Error(t.replace(/%s/g,(function(){return r[s++]})))).name="Invariant Violation"}throw i.framesToPop=1,i}}(Wl=Hl||(Hl={})).SOURCE="SOURCE",Wl.TARGET="TARGET";var zl="dnd-core/INIT_COORDS",$l="dnd-core/BEGIN_DRAG",Vl="dnd-core/PUBLISH_DRAG_SOURCE",Kl="dnd-core/HOVER",Gl="dnd-core/DROP",Jl="dnd-core/END_DRAG";function Yl(e,t){return{type:zl,payload:{sourceClientOffset:t||null,clientOffset:e||null}}}function Xl(e){return(Xl="function"==typeof Symbol&&"symbol"==typeof Symbol.iterator?function(e){return typeof e}:function(e){return e&&"function"==typeof Symbol&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e})(e)}function Zl(e){return"object"===Xl(e)}var Ql={type:zl,payload:{clientOffset:null,sourceClientOffset:null}};function eu(e){return function(){var t=arguments.length>0&&void 0!==arguments[0]?arguments[0]:[],n=arguments.length>1&&void 0!==arguments[1]?arguments[1]:{publishSource:!0},r=n.publishSource,o=void 0===r||r,i=n.clientOffset,s=n.getSourceClientOffset,a=e.getMonitor(),c=e.getRegistry();e.dispatch(Yl(i)),function(e,t,n){Bl(!t.isDragging(),"Cannot call beginDrag while dragging."),e.forEach((function(e){Bl(n.getSource(e),"Expected sourceIds to be registered.")}))}(t,a,c);var l=function(e,t){for(var n=null,r=e.length-1;r>=0;r--)if(t.canDragSource(e[r])){n=e[r];break}return n}(t,a);if(null!==l){var u=null;if(i){if(!s)throw new Error("getSourceClientOffset must be defined");!function(e){Bl("function"==typeof e,"When clientOffset is provided, getSourceClientOffset must be a function.")}(s),u=s(l)}e.dispatch(Yl(i,u));var d=c.getSource(l).beginDrag(a,l);if(null!=d){!function(e){Bl(Zl(e),"Item must be an object.")}(d),c.pinSource(l);var f=c.getSourceType(l);return{type:$l,payload:{itemType:f,item:d,sourceId:l,clientOffset:i||null,sourceClientOffset:u||null,isSourcePublic:!!o}}}}else e.dispatch(Ql)}}function tu(e){return function(){if(e.getMonitor().isDragging())return{type:Vl}}}function nu(e,t){return null===t?null===e:Array.isArray(e)?e.some((function(e){return e===t})):e===t}function ru(e){return function(t){var n=(arguments.length>1&&void 0!==arguments[1]?arguments[1]:{}).clientOffset;!function(e){Bl(Array.isArray(e),"Expected targetIds to be an array.")}(t);var r=t.slice(0),o=e.getMonitor(),i=e.getRegistry();return function(e,t,n){Bl(t.isDragging(),"Cannot call hover while not dragging."),Bl(!t.didDrop(),"Cannot call hover after drop.");for(var r=0;r<e.length;r++){var o=e[r];Bl(e.lastIndexOf(o)===r,"Expected targetIds to be unique in the passed array."),Bl(n.getTarget(o),"Expected targetIds to be registered.")}}(r,o,i),function(e,t,n){for(var r=e.length-1;r>=0;r--){var o=e[r];nu(t.getTargetType(o),n)||e.splice(r,1)}}(r,i,o.getItemType()),function(e,t,n){e.forEach((function(e){n.getTarget(e).hover(t,e)}))}(r,o,i),{type:Kl,payload:{targetIds:r,clientOffset:n||null}}}}function ou(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter((function(t){return Object.getOwnPropertyDescriptor(e,t).enumerable}))),n.push.apply(n,r)}return n}function iu(e){for(var t=1;t<arguments.length;t++){var n=null!=arguments[t]?arguments[t]:{};t%2?ou(Object(n),!0).forEach((function(t){su(e,t,n[t])})):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):ou(Object(n)).forEach((function(t){Object.defineProperty(e,t,Object.getOwnPropertyDescriptor(n,t))}))}return e}function su(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function au(e){return function(){var t=arguments.length>0&&void 0!==arguments[0]?arguments[0]:{},n=e.getMonitor(),r=e.getRegistry();!function(e){Bl(e.isDragging(),"Cannot call drop while not dragging."),Bl(!e.didDrop(),"Cannot call drop twice during one drag operation.")}(n);var o=function(e){var t=e.getTargetIds().filter(e.canDropOnTarget,e);return t.reverse(),t}(n);o.forEach((function(o,i){var s=function(e,t,n,r){var o=n.getTarget(e),i=o?o.drop(r,e):void 0;(function(e){Bl(void 0===e||Zl(e),"Drop result must either be an object or undefined.")})(i),void 0===i&&(i=0===t?{}:r.getDropResult());return i}(o,i,r,n),a={type:Gl,payload:{dropResult:iu(iu({},t),s)}};e.dispatch(a)}))}}function cu(e){return function(){var t=e.getMonitor(),n=e.getRegistry();!function(e){Bl(e.isDragging(),"Cannot call endDrag while not dragging.")}(t);var r=t.getSourceId();null!=r&&(n.getSource(r,!0).endDrag(t,r),n.unpinSource());return{type:Jl}}}function lu(e,t,n){return t&&function(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,r.key,r)}}(e.prototype,t),e}function uu(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}var du=function(){function e(t,n){var r=this;!function(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}(this,e),uu(this,"store",void 0),uu(this,"monitor",void 0),uu(this,"backend",void 0),uu(this,"isSetUp",!1),uu(this,"handleRefCountChange",(function(){var e=r.store.getState().refCount>0;r.backend&&(e&&!r.isSetUp?(r.backend.setup(),r.isSetUp=!0):!e&&r.isSetUp&&(r.backend.teardown(),r.isSetUp=!1))})),this.store=t,this.monitor=n,t.subscribe(this.handleRefCountChange)}return lu(e,[{key:"receiveBackend",value:function(e){this.backend=e}},{key:"getMonitor",value:function(){return this.monitor}},{key:"getBackend",value:function(){return this.backend}},{key:"getRegistry",value:function(){return this.monitor.registry}},{key:"getActions",value:function(){var e=this,t=this.store.dispatch;var n=function(e){return{beginDrag:eu(e),publishDragSource:tu(e),hover:ru(e),drop:au(e),endDrag:cu(e)}}(this);return Object.keys(n).reduce((function(r,o){var i,s=n[o];return r[o]=(i=s,function(){for(var n=arguments.length,r=new Array(n),o=0;o<n;o++)r[o]=arguments[o];var s=i.apply(e,r);void 0!==s&&t(s)}),r}),{})}},{key:"dispatch",value:function(e){this.store.dispatch(e)}}]),e}();function fu(e){return"Minified Redux error #"+e+"; visit https://redux.js.org/Errors?code="+e+" for the full message or use the non-minified dev environment for full errors. "}var hu="function"==typeof Symbol&&Symbol.observable||"@@observable",pu=function(){return Math.random().toString(36).substring(7).split("").join(".")},gu={INIT:"@@redux/INIT"+pu(),REPLACE:"@@redux/REPLACE"+pu(),PROBE_UNKNOWN_ACTION:function(){return"@@redux/PROBE_UNKNOWN_ACTION"+pu()}};function mu(e,t,n){var r;if("function"==typeof t&&"function"==typeof n||"function"==typeof n&&"function"==typeof arguments[3])throw new Error(fu(0));if("function"==typeof t&&void 0===n&&(n=t,t=void 0),void 0!==n){if("function"!=typeof n)throw new Error(fu(1));return n(mu)(e,t)}if("function"!=typeof e)throw new Error(fu(2));var o=e,i=t,s=[],a=s,c=!1;function l(){a===s&&(a=s.slice())}function u(){if(c)throw new Error(fu(3));return i}function d(e){if("function"!=typeof e)throw new Error(fu(4));if(c)throw new Error(fu(5));var t=!0;return l(),a.push(e),function(){if(t){if(c)throw new Error(fu(6));t=!1,l();var n=a.indexOf(e);a.splice(n,1),s=null}}}function f(e){if(!function(e){if("object"!=typeof e||null===e)return!1;for(var t=e;null!==Object.getPrototypeOf(t);)t=Object.getPrototypeOf(t);return Object.getPrototypeOf(e)===t}(e))throw new Error(fu(7));if(void 0===e.type)throw new Error(fu(8));if(c)throw new Error(fu(9));try{c=!0,i=o(i,e)}finally{c=!1}for(var t=s=a,n=0;n<t.length;n++){(0,t[n])()}return e}return f({type:gu.INIT}),(r={dispatch:f,subscribe:d,getState:u,replaceReducer:function(e){if("function"!=typeof e)throw new Error(fu(10));o=e,f({type:gu.REPLACE})}})[hu]=function(){var e,t=d;return(e={subscribe:function(e){if("object"!=typeof e||null===e)throw new Error(fu(11));function n(){e.next&&e.next(u())}return n(),{unsubscribe:t(n)}}})[hu]=function(){return this},e},r}var vu=function(e,t){return e===t};function yu(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter((function(t){return Object.getOwnPropertyDescriptor(e,t).enumerable}))),n.push.apply(n,r)}return n}function bu(e){for(var t=1;t<arguments.length;t++){var n=null!=arguments[t]?arguments[t]:{};t%2?yu(Object(n),!0).forEach((function(t){wu(e,t,n[t])})):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):yu(Object(n)).forEach((function(t){Object.defineProperty(e,t,Object.getOwnPropertyDescriptor(n,t))}))}return e}function wu(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}var xu={initialSourceClientOffset:null,initialClientOffset:null,clientOffset:null};function Tu(){var e,t,n=arguments.length>0&&void 0!==arguments[0]?arguments[0]:xu,r=arguments.length>1?arguments[1]:void 0,o=r.payload;switch(r.type){case zl:case $l:return{initialSourceClientOffset:o.sourceClientOffset,initialClientOffset:o.clientOffset,clientOffset:o.clientOffset};case Kl:return e=n.clientOffset,t=o.clientOffset,!e&&!t||e&&t&&e.x===t.x&&e.y===t.y?n:bu(bu({},n),{},{clientOffset:o.clientOffset});case Jl:case Gl:return xu;default:return n}}var Du="dnd-core/ADD_SOURCE",Su="dnd-core/ADD_TARGET",Ou="dnd-core/REMOVE_SOURCE",Ru="dnd-core/REMOVE_TARGET";function ju(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter((function(t){return Object.getOwnPropertyDescriptor(e,t).enumerable}))),n.push.apply(n,r)}return n}function Cu(e){for(var t=1;t<arguments.length;t++){var n=null!=arguments[t]?arguments[t]:{};t%2?ju(Object(n),!0).forEach((function(t){Nu(e,t,n[t])})):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):ju(Object(n)).forEach((function(t){Object.defineProperty(e,t,Object.getOwnPropertyDescriptor(n,t))}))}return e}function Nu(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}var ku={itemType:null,item:null,sourceId:null,targetIds:[],dropResult:null,didDrop:!1,isSourcePublic:null};function Pu(){var e,t,n=arguments.length>0&&void 0!==arguments[0]?arguments[0]:ku,r=arguments.length>1?arguments[1]:void 0,o=r.payload;switch(r.type){case $l:return Cu(Cu({},n),{},{itemType:o.itemType,item:o.item,sourceId:o.sourceId,isSourcePublic:o.isSourcePublic,dropResult:null,didDrop:!1});case Vl:return Cu(Cu({},n),{},{isSourcePublic:!0});case Kl:return Cu(Cu({},n),{},{targetIds:o.targetIds});case Ru:return-1===n.targetIds.indexOf(o.targetId)?n:Cu(Cu({},n),{},{targetIds:(e=n.targetIds,t=o.targetId,e.filter((function(e){return e!==t})))});case Gl:return Cu(Cu({},n),{},{dropResult:o.dropResult,didDrop:!0,targetIds:[]});case Jl:return Cu(Cu({},n),{},{itemType:null,item:null,sourceId:null,dropResult:null,didDrop:!1,isSourcePublic:null,targetIds:[]});default:return n}}function Eu(){var e=arguments.length>0&&void 0!==arguments[0]?arguments[0]:0;switch((arguments.length>1?arguments[1]:void 0).type){case Du:case Su:return e+1;case Ou:case Ru:return e-1;default:return e}}var Iu=[],_u=[];function Mu(){var e=arguments.length>1?arguments[1]:void 0;switch(e.type){case Kl:break;case Du:case Su:case Ru:case Ou:return Iu;default:return _u}var t=e.payload,n=t.targetIds,r=void 0===n?[]:n,o=t.prevTargetIds,i=void 0===o?[]:o,s=function(e,t){var n=new Map,r=function(e){n.set(e,n.has(e)?n.get(e)+1:1)};e.forEach(r),t.forEach(r);var o=[];return n.forEach((function(e,t){1===e&&o.push(t)})),o}(r,i),a=s.length>0||!function(e,t){var n=arguments.length>2&&void 0!==arguments[2]?arguments[2]:vu;if(e.length!==t.length)return!1;for(var r=0;r<e.length;++r)if(!n(e[r],t[r]))return!1;return!0}(r,i);if(!a)return Iu;var c=i[i.length-1],l=r[r.length-1];return c!==l&&(c&&s.push(c),l&&s.push(l)),s}function qu(){return(arguments.length>0&&void 0!==arguments[0]?arguments[0]:0)+1}function Au(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter((function(t){return Object.getOwnPropertyDescriptor(e,t).enumerable}))),n.push.apply(n,r)}return n}function Lu(e){for(var t=1;t<arguments.length;t++){var n=null!=arguments[t]?arguments[t]:{};t%2?Au(Object(n),!0).forEach((function(t){Fu(e,t,n[t])})):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):Au(Object(n)).forEach((function(t){Object.defineProperty(e,t,Object.getOwnPropertyDescriptor(n,t))}))}return e}function Fu(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function Hu(){var e,t,n,r=arguments.length>0&&void 0!==arguments[0]?arguments[0]:{},o=arguments.length>1?arguments[1]:void 0;return{dirtyHandlerIds:Mu(r.dirtyHandlerIds,{type:o.type,payload:Lu(Lu({},o.payload),{},{prevTargetIds:(e=r,t="dragOperation.targetIds",n=[],t.split(".").reduce((function(e,t){return e&&e[t]?e[t]:n||null}),e))})}),dragOffset:Tu(r.dragOffset,o),refCount:Eu(r.refCount,o),dragOperation:Pu(r.dragOperation,o),stateId:qu(r.stateId)}}function Wu(e,t){return{x:e.x-t.x,y:e.y-t.y}}function Uu(e,t,n){return t&&function(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,r.key,r)}}(e.prototype,t),e}function Bu(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}Iu.__IS_NONE__=!0,_u.__IS_ALL__=!0;var zu=function(){function e(t,n){!function(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}(this,e),Bu(this,"store",void 0),Bu(this,"registry",void 0),this.store=t,this.registry=n}return Uu(e,[{key:"subscribeToStateChange",value:function(e){var t=this,n=(arguments.length>1&&void 0!==arguments[1]?arguments[1]:{handlerIds:void 0}).handlerIds;Bl("function"==typeof e,"listener must be a function."),Bl(void 0===n||Array.isArray(n),"handlerIds, when specified, must be an array of strings.");var r=this.store.getState().stateId;return this.store.subscribe((function(){var o=t.store.getState(),i=o.stateId;try{var s=i===r||i===r+1&&!function(e,t){return e!==Iu&&(e===_u||void 0===t||(n=e,t.filter((function(e){return n.indexOf(e)>-1}))).length>0);var n}(o.dirtyHandlerIds,n);s||e()}finally{r=i}}))}},{key:"subscribeToOffsetChange",value:function(e){var t=this;Bl("function"==typeof e,"listener must be a function.");var n=this.store.getState().dragOffset;return this.store.subscribe((function(){var r=t.store.getState().dragOffset;r!==n&&(n=r,e())}))}},{key:"canDragSource",value:function(e){if(!e)return!1;var t=this.registry.getSource(e);return Bl(t,"Expected to find a valid source. sourceId=".concat(e)),!this.isDragging()&&t.canDrag(this,e)}},{key:"canDropOnTarget",value:function(e){if(!e)return!1;var t=this.registry.getTarget(e);return Bl(t,"Expected to find a valid target. targetId=".concat(e)),!(!this.isDragging()||this.didDrop())&&(nu(this.registry.getTargetType(e),this.getItemType())&&t.canDrop(this,e))}},{key:"isDragging",value:function(){return Boolean(this.getItemType())}},{key:"isDraggingSource",value:function(e){if(!e)return!1;var t=this.registry.getSource(e,!0);return Bl(t,"Expected to find a valid source. sourceId=".concat(e)),!(!this.isDragging()||!this.isSourcePublic())&&(this.registry.getSourceType(e)===this.getItemType()&&t.isDragging(this,e))}},{key:"isOverTarget",value:function(e){if(!e)return!1;var t=(arguments.length>1&&void 0!==arguments[1]?arguments[1]:{shallow:!1}).shallow;if(!this.isDragging())return!1;var n=this.registry.getTargetType(e),r=this.getItemType();if(r&&!nu(n,r))return!1;var o=this.getTargetIds();if(!o.length)return!1;var i=o.indexOf(e);return t?i===o.length-1:i>-1}},{key:"getItemType",value:function(){return this.store.getState().dragOperation.itemType}},{key:"getItem",value:function(){return this.store.getState().dragOperation.item}},{key:"getSourceId",value:function(){return this.store.getState().dragOperation.sourceId}},{key:"getTargetIds",value:function(){return this.store.getState().dragOperation.targetIds}},{key:"getDropResult",value:function(){return this.store.getState().dragOperation.dropResult}},{key:"didDrop",value:function(){return this.store.getState().dragOperation.didDrop}},{key:"isSourcePublic",value:function(){return Boolean(this.store.getState().dragOperation.isSourcePublic)}},{key:"getInitialClientOffset",value:function(){return this.store.getState().dragOffset.initialClientOffset}},{key:"getInitialSourceClientOffset",value:function(){return this.store.getState().dragOffset.initialSourceClientOffset}},{key:"getClientOffset",value:function(){return this.store.getState().dragOffset.clientOffset}},{key:"getSourceClientOffset",value:function(){return e=this.store.getState().dragOffset,r=e.clientOffset,o=e.initialClientOffset,i=e.initialSourceClientOffset,r&&o&&i?Wu((n=i,{x:(t=r).x+n.x,y:t.y+n.y}),o):null;var e,t,n,r,o,i}},{key:"getDifferenceFromInitialOffset",value:function(){return e=this.store.getState().dragOffset,t=e.clientOffset,n=e.initialClientOffset,t&&n?Wu(t,n):null;var e,t,n}}]),e}(),$u=0;function Vu(e){return(Vu="function"==typeof Symbol&&"symbol"==typeof Symbol.iterator?function(e){return typeof e}:function(e){return e&&"function"==typeof Symbol&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e})(e)}function Ku(e,t){t&&Array.isArray(e)?e.forEach((function(e){return Ku(e,!1)})):Bl("string"==typeof e||"symbol"===Vu(e),t?"Type can only be a string, a symbol, or an array of either.":"Type can only be a string or a symbol.")}const Gu="undefined"!=typeof global?global:self,Ju=Gu.MutationObserver||Gu.WebKitMutationObserver;function Yu(e){return function(){const t=setTimeout(r,0),n=setInterval(r,50);function r(){clearTimeout(t),clearInterval(n),e()}}}const Xu="function"==typeof Ju?function(e){let t=1;const n=new Ju(e),r=document.createTextNode("");return n.observe(r,{characterData:!0}),function(){t=-t,r.data=t}}:Yu;class Zu{call(){try{this.task&&this.task()}catch(e){this.onError(e)}finally{this.task=null,this.release(this)}}constructor(e,t){this.onError=e,this.release=t,this.task=null}}const Qu=new class{enqueueTask(e){const{queue:t,requestFlush:n}=this;t.length||(n(),this.flushing=!0),t[t.length]=e}constructor(){this.queue=[],this.pendingErrors=[],this.flushing=!1,this.index=0,this.capacity=1024,this.flush=()=>{const{queue:e}=this;for(;this.index<e.length;){const t=this.index;if(this.index++,e[t].call(),this.index>this.capacity){for(let t=0,n=e.length-this.index;t<n;t++)e[t]=e[t+this.index];e.length-=this.index,this.index=0}}e.length=0,this.index=0,this.flushing=!1},this.registerPendingError=e=>{this.pendingErrors.push(e),this.requestErrorThrow()},this.requestFlush=Xu(this.flush),this.requestErrorThrow=Yu((()=>{if(this.pendingErrors.length)throw this.pendingErrors.shift()}))}},ed=new class{create(e){const t=this.freeTasks,n=t.length?t.pop():new Zu(this.onError,(e=>t[t.length]=e));return n.task=e,n}constructor(e){this.onError=e,this.freeTasks=[]}}(Qu.registerPendingError);function td(e,t,n){return t&&function(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,r.key,r)}}(e.prototype,t),e}function nd(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}function rd(e,t){return function(e){if(Array.isArray(e))return e}(e)||function(e,t){var n=null==e?null:"undefined"!=typeof Symbol&&e[Symbol.iterator]||e["@@iterator"];if(null==n)return;var r,o,i=[],s=!0,a=!1;try{for(n=n.call(e);!(s=(r=n.next()).done)&&(i.push(r.value),!t||i.length!==t);s=!0);}catch(Ih){a=!0,o=Ih}finally{try{s||null==n.return||n.return()}finally{if(a)throw o}}return i}(e,t)||function(e,t){if(!e)return;if("string"==typeof e)return od(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);"Object"===n&&e.constructor&&(n=e.constructor.name);if("Map"===n||"Set"===n)return Array.from(e);if("Arguments"===n||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return od(e,t)}(e,t)||function(){throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.")}()}function od(e,t){(null==t||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function id(e){var t=($u++).toString();switch(e){case Hl.SOURCE:return"S".concat(t);case Hl.TARGET:return"T".concat(t);default:throw new Error("Unknown Handler Role: ".concat(e))}}function sd(e){switch(e[0]){case"S":return Hl.SOURCE;case"T":return Hl.TARGET;default:Bl(!1,"Cannot parse handler ID: ".concat(e))}}function ad(e,t){var n=e.entries(),r=!1;do{var o=n.next(),i=o.done;if(rd(o.value,2)[1]===t)return!0;r=!!i}while(!r);return!1}var cd=function(){function e(t){!function(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}(this,e),nd(this,"types",new Map),nd(this,"dragSources",new Map),nd(this,"dropTargets",new Map),nd(this,"pinnedSourceId",null),nd(this,"pinnedSource",null),nd(this,"store",void 0),this.store=t}return td(e,[{key:"addSource",value:function(e,t){Ku(e),function(e){Bl("function"==typeof e.canDrag,"Expected canDrag to be a function."),Bl("function"==typeof e.beginDrag,"Expected beginDrag to be a function."),Bl("function"==typeof e.endDrag,"Expected endDrag to be a function.")}(t);var n=this.addHandler(Hl.SOURCE,e,t);return this.store.dispatch(function(e){return{type:Du,payload:{sourceId:e}}}(n)),n}},{key:"addTarget",value:function(e,t){Ku(e,!0),function(e){Bl("function"==typeof e.canDrop,"Expected canDrop to be a function."),Bl("function"==typeof e.hover,"Expected hover to be a function."),Bl("function"==typeof e.drop,"Expected beginDrag to be a function.")}(t);var n=this.addHandler(Hl.TARGET,e,t);return this.store.dispatch(function(e){return{type:Su,payload:{targetId:e}}}(n)),n}},{key:"containsHandler",value:function(e){return ad(this.dragSources,e)||ad(this.dropTargets,e)}},{key:"getSource",value:function(e){var t=arguments.length>1&&void 0!==arguments[1]&&arguments[1];return Bl(this.isSourceId(e),"Expected a valid source ID."),t&&e===this.pinnedSourceId?this.pinnedSource:this.dragSources.get(e)}},{key:"getTarget",value:function(e){return Bl(this.isTargetId(e),"Expected a valid target ID."),this.dropTargets.get(e)}},{key:"getSourceType",value:function(e){return Bl(this.isSourceId(e),"Expected a valid source ID."),this.types.get(e)}},{key:"getTargetType",value:function(e){return Bl(this.isTargetId(e),"Expected a valid target ID."),this.types.get(e)}},{key:"isSourceId",value:function(e){return sd(e)===Hl.SOURCE}},{key:"isTargetId",value:function(e){return sd(e)===Hl.TARGET}},{key:"removeSource",value:function(e){var t,n=this;Bl(this.getSource(e),"Expected an existing source."),this.store.dispatch(function(e){return{type:Ou,payload:{sourceId:e}}}(e)),t=function(){n.dragSources.delete(e),n.types.delete(e)},Qu.enqueueTask(ed.create(t))}},{key:"removeTarget",value:function(e){Bl(this.getTarget(e),"Expected an existing target."),this.store.dispatch(function(e){return{type:Ru,payload:{targetId:e}}}(e)),this.dropTargets.delete(e),this.types.delete(e)}},{key:"pinSource",value:function(e){var t=this.getSource(e);Bl(t,"Expected an existing source."),this.pinnedSourceId=e,this.pinnedSource=t}},{key:"unpinSource",value:function(){Bl(this.pinnedSource,"No source is pinned at the time."),this.pinnedSourceId=null,this.pinnedSource=null}},{key:"addHandler",value:function(e,t,n){var r=id(e);return this.types.set(r,t),e===Hl.SOURCE?this.dragSources.set(r,n):e===Hl.TARGET&&this.dropTargets.set(r,n),r}}]),e}();function ld(e){var t=arguments.length>1&&void 0!==arguments[1]?arguments[1]:void 0,n=arguments.length>2&&void 0!==arguments[2]?arguments[2]:{},r=function(e){var t="undefined"!=typeof window&&window.__REDUX_DEVTOOLS_EXTENSION__;return mu(Hu,e&&t&&t({name:"dnd-core",instanceId:"dnd-core"}))}(arguments.length>3&&void 0!==arguments[3]&&arguments[3]),o=new zu(r,new cd(r)),i=new du(r,o),s=e(i,t,n);return i.receiveBackend(s),i}var ud=["children"];function dd(e,t){return function(e){if(Array.isArray(e))return e}(e)||function(e,t){var n=null==e?null:"undefined"!=typeof Symbol&&e[Symbol.iterator]||e["@@iterator"];if(null==n)return;var r,o,i=[],s=!0,a=!1;try{for(n=n.call(e);!(s=(r=n.next()).done)&&(i.push(r.value),!t||i.length!==t);s=!0);}catch(Ih){a=!0,o=Ih}finally{try{s||null==n.return||n.return()}finally{if(a)throw o}}return i}(e,t)||function(e,t){if(!e)return;if("string"==typeof e)return fd(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);"Object"===n&&e.constructor&&(n=e.constructor.name);if("Map"===n||"Set"===n)return Array.from(e);if("Arguments"===n||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return fd(e,t)}(e,t)||function(){throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.")}()}function fd(e,t){(null==t||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function hd(e,t){if(null==e)return{};var n,r,o=function(e,t){if(null==e)return{};var n,r,o={},i=Object.keys(e);for(r=0;r<i.length;r++)n=i[r],t.indexOf(n)>=0||(o[n]=e[n]);return o}(e,t);if(Object.getOwnPropertySymbols){var i=Object.getOwnPropertySymbols(e);for(r=0;r<i.length;r++)n=i[r],t.indexOf(n)>=0||Object.prototype.propertyIsEnumerable.call(e,n)&&(o[n]=e[n])}return o}var pd=0,gd=Symbol.for("__REACT_DND_CONTEXT_INSTANCE__"),md=h.memo((function(e){var t=e.children,n=function(e){if("manager"in e){return[{dragDropManager:e.manager},!1]}var t=function(e){var t=arguments.length>1&&void 0!==arguments[1]?arguments[1]:vd(),n=arguments.length>2?arguments[2]:void 0,r=arguments.length>3?arguments[3]:void 0,o=t;o[gd]||(o[gd]={dragDropManager:ld(e,t,n,r)});return o[gd]}(e.backend,e.context,e.options,e.debugMode),n=!e.context;return[t,n]}(hd(e,ud)),r=dd(n,2),o=r[0],i=r[1];return h.useEffect((function(){if(i){var e=vd();return++pd,function(){0==--pd&&(e[gd]=null)}}}),[]),a.jsx(Ul.Provider,Object.assign({value:o},{children:t}),void 0)}));function vd(){return"undefined"!=typeof global?global:window}function yd(e,t,n){return t&&function(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,r.key,r)}}(e.prototype,t),e}function bd(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}var wd=!1,xd=!1,Td=function(){function e(t){!function(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}(this,e),bd(this,"internalMonitor",void 0),bd(this,"sourceId",null),this.internalMonitor=t.getMonitor()}return yd(e,[{key:"receiveHandlerId",value:function(e){this.sourceId=e}},{key:"getHandlerId",value:function(){return this.sourceId}},{key:"canDrag",value:function(){Bl(!wd,"You may not call monitor.canDrag() inside your canDrag() implementation. Read more: http://react-dnd.github.io/react-dnd/docs/api/drag-source-monitor");try{return wd=!0,this.internalMonitor.canDragSource(this.sourceId)}finally{wd=!1}}},{key:"isDragging",value:function(){if(!this.sourceId)return!1;Bl(!xd,"You may not call monitor.isDragging() inside your isDragging() implementation. Read more: http://react-dnd.github.io/react-dnd/docs/api/drag-source-monitor");try{return xd=!0,this.internalMonitor.isDraggingSource(this.sourceId)}finally{xd=!1}}},{key:"subscribeToStateChange",value:function(e,t){return this.internalMonitor.subscribeToStateChange(e,t)}},{key:"isDraggingSource",value:function(e){return this.internalMonitor.isDraggingSource(e)}},{key:"isOverTarget",value:function(e,t){return this.internalMonitor.isOverTarget(e,t)}},{key:"getTargetIds",value:function(){return this.internalMonitor.getTargetIds()}},{key:"isSourcePublic",value:function(){return this.internalMonitor.isSourcePublic()}},{key:"getSourceId",value:function(){return this.internalMonitor.getSourceId()}},{key:"subscribeToOffsetChange",value:function(e){return this.internalMonitor.subscribeToOffsetChange(e)}},{key:"canDragSource",value:function(e){return this.internalMonitor.canDragSource(e)}},{key:"canDropOnTarget",value:function(e){return this.internalMonitor.canDropOnTarget(e)}},{key:"getItemType",value:function(){return this.internalMonitor.getItemType()}},{key:"getItem",value:function(){return this.internalMonitor.getItem()}},{key:"getDropResult",value:function(){return this.internalMonitor.getDropResult()}},{key:"didDrop",value:function(){return this.internalMonitor.didDrop()}},{key:"getInitialClientOffset",value:function(){return this.internalMonitor.getInitialClientOffset()}},{key:"getInitialSourceClientOffset",value:function(){return this.internalMonitor.getInitialSourceClientOffset()}},{key:"getSourceClientOffset",value:function(){return this.internalMonitor.getSourceClientOffset()}},{key:"getClientOffset",value:function(){return this.internalMonitor.getClientOffset()}},{key:"getDifferenceFromInitialOffset",value:function(){return this.internalMonitor.getDifferenceFromInitialOffset()}}]),e}();function Dd(e,t,n){return t&&function(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,r.key,r)}}(e.prototype,t),e}function Sd(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}var Od=!1,Rd=function(){function e(t){!function(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}(this,e),Sd(this,"internalMonitor",void 0),Sd(this,"targetId",null),this.internalMonitor=t.getMonitor()}return Dd(e,[{key:"receiveHandlerId",value:function(e){this.targetId=e}},{key:"getHandlerId",value:function(){return this.targetId}},{key:"subscribeToStateChange",value:function(e,t){return this.internalMonitor.subscribeToStateChange(e,t)}},{key:"canDrop",value:function(){if(!this.targetId)return!1;Bl(!Od,"You may not call monitor.canDrop() inside your canDrop() implementation. Read more: http://react-dnd.github.io/react-dnd/docs/api/drop-target-monitor");try{return Od=!0,this.internalMonitor.canDropOnTarget(this.targetId)}finally{Od=!1}}},{key:"isOver",value:function(e){return!!this.targetId&&this.internalMonitor.isOverTarget(this.targetId,e)}},{key:"getItemType",value:function(){return this.internalMonitor.getItemType()}},{key:"getItem",value:function(){return this.internalMonitor.getItem()}},{key:"getDropResult",value:function(){return this.internalMonitor.getDropResult()}},{key:"didDrop",value:function(){return this.internalMonitor.didDrop()}},{key:"getInitialClientOffset",value:function(){return this.internalMonitor.getInitialClientOffset()}},{key:"getInitialSourceClientOffset",value:function(){return this.internalMonitor.getInitialSourceClientOffset()}},{key:"getSourceClientOffset",value:function(){return this.internalMonitor.getSourceClientOffset()}},{key:"getClientOffset",value:function(){return this.internalMonitor.getClientOffset()}},{key:"getDifferenceFromInitialOffset",value:function(){return this.internalMonitor.getDifferenceFromInitialOffset()}}]),e}();function jd(e){return function(){var t=arguments.length>0&&void 0!==arguments[0]?arguments[0]:null,n=arguments.length>1&&void 0!==arguments[1]?arguments[1]:null;if(!h.isValidElement(t)){var r=t;return e(r,n),r}var o=t;return function(e){if("string"!=typeof e.type){var t=e.type.displayName||e.type.name||"the component";throw new Error("Only native element nodes can now be passed to React DnD connectors."+"You can either wrap ".concat(t," into a <div>, or turn it into a ")+"drag source or a drop target itself.")}}(o),function(e,t){var n=e.ref;return Bl("string"!=typeof n,"Cannot connect React DnD to an element with an existing string ref. Please convert it to use a callback ref instead, or wrap it into a <span> or <div>. Read more: https://reactjs.org/docs/refs-and-the-dom.html#callback-refs"),n?h.cloneElement(e,{ref:function(e){Nd(n,e),Nd(t,e)}}):h.cloneElement(e,{ref:t})}(o,n?function(t){return e(t,n)}:e)}}function Cd(e){var t={};return Object.keys(e).forEach((function(n){var r=e[n];if(n.endsWith("Ref"))t[n]=e[n];else{var o=jd(r);t[n]=function(){return o}}})),t}function Nd(e,t){"function"==typeof e?e(t):e.current=t}function kd(e){return(kd="function"==typeof Symbol&&"symbol"==typeof Symbol.iterator?function(e){return typeof e}:function(e){return e&&"function"==typeof Symbol&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e})(e)}function Pd(e){return null!==e&&"object"===kd(e)&&Object.prototype.hasOwnProperty.call(e,"current")}function Ed(e,t,n,r){var o=void 0;if(void 0!==o)return!!o;if(e===t)return!0;if("object"!=typeof e||!e||"object"!=typeof t||!t)return!1;var i=Object.keys(e),s=Object.keys(t);if(i.length!==s.length)return!1;for(var a=Object.prototype.hasOwnProperty.bind(t),c=0;c<i.length;c++){var l=i[c];if(!a(l))return!1;var u=e[l],d=t[l];if(!1===(o=void 0)||void 0===o&&u!==d)return!1}return!0}function Id(e,t,n){return t&&function(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,r.key,r)}}(e.prototype,t),e}function _d(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}var Md=function(){function e(t){var n=this;!function(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}(this,e),_d(this,"hooks",Cd({dragSource:function(e,t){n.clearDragSource(),n.dragSourceOptions=t||null,Pd(e)?n.dragSourceRef=e:n.dragSourceNode=e,n.reconnectDragSource()},dragPreview:function(e,t){n.clearDragPreview(),n.dragPreviewOptions=t||null,Pd(e)?n.dragPreviewRef=e:n.dragPreviewNode=e,n.reconnectDragPreview()}})),_d(this,"handlerId",null),_d(this,"dragSourceRef",null),_d(this,"dragSourceNode",void 0),_d(this,"dragSourceOptionsInternal",null),_d(this,"dragSourceUnsubscribe",void 0),_d(this,"dragPreviewRef",null),_d(this,"dragPreviewNode",void 0),_d(this,"dragPreviewOptionsInternal",null),_d(this,"dragPreviewUnsubscribe",void 0),_d(this,"lastConnectedHandlerId",null),_d(this,"lastConnectedDragSource",null),_d(this,"lastConnectedDragSourceOptions",null),_d(this,"lastConnectedDragPreview",null),_d(this,"lastConnectedDragPreviewOptions",null),_d(this,"backend",void 0),this.backend=t}return Id(e,[{key:"receiveHandlerId",value:function(e){this.handlerId!==e&&(this.handlerId=e,this.reconnect())}},{key:"connectTarget",get:function(){return this.dragSource}},{key:"dragSourceOptions",get:function(){return this.dragSourceOptionsInternal},set:function(e){this.dragSourceOptionsInternal=e}},{key:"dragPreviewOptions",get:function(){return this.dragPreviewOptionsInternal},set:function(e){this.dragPreviewOptionsInternal=e}},{key:"reconnect",value:function(){this.reconnectDragSource(),this.reconnectDragPreview()}},{key:"reconnectDragSource",value:function(){var e=this.dragSource,t=this.didHandlerIdChange()||this.didConnectedDragSourceChange()||this.didDragSourceOptionsChange();t&&this.disconnectDragSource(),this.handlerId&&(e?t&&(this.lastConnectedHandlerId=this.handlerId,this.lastConnectedDragSource=e,this.lastConnectedDragSourceOptions=this.dragSourceOptions,this.dragSourceUnsubscribe=this.backend.connectDragSource(this.handlerId,e,this.dragSourceOptions)):this.lastConnectedDragSource=e)}},{key:"reconnectDragPreview",value:function(){var e=this.dragPreview,t=this.didHandlerIdChange()||this.didConnectedDragPreviewChange()||this.didDragPreviewOptionsChange();t&&this.disconnectDragPreview(),this.handlerId&&(e?t&&(this.lastConnectedHandlerId=this.handlerId,this.lastConnectedDragPreview=e,this.lastConnectedDragPreviewOptions=this.dragPreviewOptions,this.dragPreviewUnsubscribe=this.backend.connectDragPreview(this.handlerId,e,this.dragPreviewOptions)):this.lastConnectedDragPreview=e)}},{key:"didHandlerIdChange",value:function(){return this.lastConnectedHandlerId!==this.handlerId}},{key:"didConnectedDragSourceChange",value:function(){return this.lastConnectedDragSource!==this.dragSource}},{key:"didConnectedDragPreviewChange",value:function(){return this.lastConnectedDragPreview!==this.dragPreview}},{key:"didDragSourceOptionsChange",value:function(){return!Ed(this.lastConnectedDragSourceOptions,this.dragSourceOptions)}},{key:"didDragPreviewOptionsChange",value:function(){return!Ed(this.lastConnectedDragPreviewOptions,this.dragPreviewOptions)}},{key:"disconnectDragSource",value:function(){this.dragSourceUnsubscribe&&(this.dragSourceUnsubscribe(),this.dragSourceUnsubscribe=void 0)}},{key:"disconnectDragPreview",value:function(){this.dragPreviewUnsubscribe&&(this.dragPreviewUnsubscribe(),this.dragPreviewUnsubscribe=void 0,this.dragPreviewNode=null,this.dragPreviewRef=null)}},{key:"dragSource",get:function(){return this.dragSourceNode||this.dragSourceRef&&this.dragSourceRef.current}},{key:"dragPreview",get:function(){return this.dragPreviewNode||this.dragPreviewRef&&this.dragPreviewRef.current}},{key:"clearDragSource",value:function(){this.dragSourceNode=null,this.dragSourceRef=null}},{key:"clearDragPreview",value:function(){this.dragPreviewNode=null,this.dragPreviewRef=null}}]),e}();function qd(e,t,n){return t&&function(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,r.key,r)}}(e.prototype,t),e}function Ad(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}var Ld=function(){function e(t){var n=this;!function(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}(this,e),Ad(this,"hooks",Cd({dropTarget:function(e,t){n.clearDropTarget(),n.dropTargetOptions=t,Pd(e)?n.dropTargetRef=e:n.dropTargetNode=e,n.reconnect()}})),Ad(this,"handlerId",null),Ad(this,"dropTargetRef",null),Ad(this,"dropTargetNode",void 0),Ad(this,"dropTargetOptionsInternal",null),Ad(this,"unsubscribeDropTarget",void 0),Ad(this,"lastConnectedHandlerId",null),Ad(this,"lastConnectedDropTarget",null),Ad(this,"lastConnectedDropTargetOptions",null),Ad(this,"backend",void 0),this.backend=t}return qd(e,[{key:"connectTarget",get:function(){return this.dropTarget}},{key:"reconnect",value:function(){var e=this.didHandlerIdChange()||this.didDropTargetChange()||this.didOptionsChange();e&&this.disconnectDropTarget();var t=this.dropTarget;this.handlerId&&(t?e&&(this.lastConnectedHandlerId=this.handlerId,this.lastConnectedDropTarget=t,this.lastConnectedDropTargetOptions=this.dropTargetOptions,this.unsubscribeDropTarget=this.backend.connectDropTarget(this.handlerId,t,this.dropTargetOptions)):this.lastConnectedDropTarget=t)}},{key:"receiveHandlerId",value:function(e){e!==this.handlerId&&(this.handlerId=e,this.reconnect())}},{key:"dropTargetOptions",get:function(){return this.dropTargetOptionsInternal},set:function(e){this.dropTargetOptionsInternal=e}},{key:"didHandlerIdChange",value:function(){return this.lastConnectedHandlerId!==this.handlerId}},{key:"didDropTargetChange",value:function(){return this.lastConnectedDropTarget!==this.dropTarget}},{key:"didOptionsChange",value:function(){return!Ed(this.lastConnectedDropTargetOptions,this.dropTargetOptions)}},{key:"disconnectDropTarget",value:function(){this.unsubscribeDropTarget&&(this.unsubscribeDropTarget(),this.unsubscribeDropTarget=void 0)}},{key:"dropTarget",get:function(){return this.dropTargetNode||this.dropTargetRef&&this.dropTargetRef.current}},{key:"clearDropTarget",value:function(){this.dropTargetRef=null,this.dropTargetNode=null}}]),e}();var Fd="undefined"!=typeof window?h.useLayoutEffect:h.useEffect;function Hd(e){return(Hd="function"==typeof Symbol&&"symbol"==typeof Symbol.iterator?function(e){return typeof e}:function(e){return e&&"function"==typeof Symbol&&e.constructor===Symbol&&e!==Symbol.prototype?"symbol":typeof e})(e)}function Wd(e,t,n){return t&&function(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,r.key,r)}}(e.prototype,t),e}function Ud(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}var Bd=function(){function e(t,n,r){!function(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}(this,e),Ud(this,"spec",void 0),Ud(this,"monitor",void 0),Ud(this,"connector",void 0),this.spec=t,this.monitor=n,this.connector=r}return Wd(e,[{key:"beginDrag",value:function(){var e,t=this.spec,n=this.monitor;return null!==(e="object"===Hd(t.item)?t.item:"function"==typeof t.item?t.item(n):{})&&void 0!==e?e:null}},{key:"canDrag",value:function(){var e=this.spec,t=this.monitor;return"boolean"==typeof e.canDrag?e.canDrag:"function"!=typeof e.canDrag||e.canDrag(t)}},{key:"isDragging",value:function(e,t){var n=this.spec,r=this.monitor,o=n.isDragging;return o?o(r):t===e.getSourceId()}},{key:"endDrag",value:function(){var e=this.spec,t=this.monitor,n=this.connector,r=e.end;r&&r(t.getItem(),t),n.reconnect()}}]),e}();function zd(){var e=h.useContext(Ul).dragDropManager;return Bl(null!=e,"Expected drag drop context"),e}function $d(e,t){return function(e){if(Array.isArray(e))return e}(e)||function(e,t){var n=null==e?null:"undefined"!=typeof Symbol&&e[Symbol.iterator]||e["@@iterator"];if(null==n)return;var r,o,i=[],s=!0,a=!1;try{for(n=n.call(e);!(s=(r=n.next()).done)&&(i.push(r.value),!t||i.length!==t);s=!0);}catch(Ih){a=!0,o=Ih}finally{try{s||null==n.return||n.return()}finally{if(a)throw o}}return i}(e,t)||function(e,t){if(!e)return;if("string"==typeof e)return Vd(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);"Object"===n&&e.constructor&&(n=e.constructor.name);if("Map"===n||"Set"===n)return Array.from(e);if("Arguments"===n||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return Vd(e,t)}(e,t)||function(){throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.")}()}function Vd(e,t){(null==t||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function Kd(e,t,n){var r=zd(),o=function(e,t,n){var r=h.useMemo((function(){return new Bd(e,t,n)}),[t,n]);return h.useEffect((function(){r.spec=e}),[e]),r}(e,t,n),i=function(e){return h.useMemo((function(){var t=e.type;return Bl(null!=t,"spec.type must be defined"),t}),[e])}(e);Fd((function(){if(null!=i){var e=function(e,t,n){var r=n.getRegistry(),o=r.addSource(e,t);return[o,function(){return r.removeSource(o)}]}(i,o,r),s=$d(e,2),a=s[0],c=s[1];return t.receiveHandlerId(a),n.receiveHandlerId(a),c}}),[r,t,n,o,i])}function Gd(e){return function(e){if(Array.isArray(e))return Jd(e)}(e)||function(e){if("undefined"!=typeof Symbol&&null!=e[Symbol.iterator]||null!=e["@@iterator"])return Array.from(e)}(e)||function(e,t){if(!e)return;if("string"==typeof e)return Jd(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);"Object"===n&&e.constructor&&(n=e.constructor.name);if("Map"===n||"Set"===n)return Array.from(e);if("Arguments"===n||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return Jd(e,t)}(e)||function(){throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.")}()}function Jd(e,t){(null==t||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function Yd(e,t){var n=Gd(t||[]);return null==t&&"function"!=typeof e&&n.push(e),h.useMemo((function(){return"function"==typeof e?e():e}),n)}function Xd(e,t){return function(e){if(Array.isArray(e))return e}(e)||function(e,t){var n=null==e?null:"undefined"!=typeof Symbol&&e[Symbol.iterator]||e["@@iterator"];if(null==n)return;var r,o,i=[],s=!0,a=!1;try{for(n=n.call(e);!(s=(r=n.next()).done)&&(i.push(r.value),!t||i.length!==t);s=!0);}catch(Ih){a=!0,o=Ih}finally{try{s||null==n.return||n.return()}finally{if(a)throw o}}return i}(e,t)||function(e,t){if(!e)return;if("string"==typeof e)return Zd(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);"Object"===n&&e.constructor&&(n=e.constructor.name);if("Map"===n||"Set"===n)return Array.from(e);if("Arguments"===n||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return Zd(e,t)}(e,t)||function(){throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.")}()}function Zd(e,t){(null==t||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function Qd(e,t,n){var r=Xd(h.useState((function(){return t(e)})),2),o=r[0],i=r[1],s=h.useCallback((function(){var r=t(e);ot(o,r)||(i(r),n&&n())}),[o,e,n]);return Fd(s),[o,s]}function ef(e,t){return function(e){if(Array.isArray(e))return e}(e)||function(e,t){var n=null==e?null:"undefined"!=typeof Symbol&&e[Symbol.iterator]||e["@@iterator"];if(null==n)return;var r,o,i=[],s=!0,a=!1;try{for(n=n.call(e);!(s=(r=n.next()).done)&&(i.push(r.value),!t||i.length!==t);s=!0);}catch(Ih){a=!0,o=Ih}finally{try{s||null==n.return||n.return()}finally{if(a)throw o}}return i}(e,t)||function(e,t){if(!e)return;if("string"==typeof e)return tf(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);"Object"===n&&e.constructor&&(n=e.constructor.name);if("Map"===n||"Set"===n)return Array.from(e);if("Arguments"===n||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return tf(e,t)}(e,t)||function(){throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.")}()}function tf(e,t){(null==t||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function nf(e,t,n){return function(e,t,n){var r=ef(Qd(e,t,n),2),o=r[0],i=r[1];return Fd((function(){var t=e.getHandlerId();if(null!=t)return e.subscribeToStateChange(i,{handlerIds:[t]})}),[e,i]),o}(t,e||function(){return{}},(function(){return n.reconnect()}))}function rf(e){return h.useMemo((function(){return e.hooks.dragSource()}),[e])}function of(e){return h.useMemo((function(){return e.hooks.dragPreview()}),[e])}function sf(e,t){var n=Yd(e,t);Bl(!n.begin,"useDrag::spec.begin was deprecated in v14. Replace spec.begin() with spec.item(). (see more here - https://react-dnd.github.io/react-dnd/docs/api/use-drag)");var r,o=(r=zd(),h.useMemo((function(){return new Td(r)}),[r])),i=function(e,t){var n=zd(),r=h.useMemo((function(){return new Md(n.getBackend())}),[n]);return Fd((function(){return r.dragSourceOptions=e||null,r.reconnect(),function(){return r.disconnectDragSource()}}),[r,e]),Fd((function(){return r.dragPreviewOptions=t||null,r.reconnect(),function(){return r.disconnectDragPreview()}}),[r,t]),r}(n.options,n.previewOptions);return Kd(n,o,i),[nf(n.collect,o,i),rf(i),of(i)]}function af(e,t,n){return t&&function(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,r.key,r)}}(e.prototype,t),e}function cf(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}var lf=function(){function e(t,n){!function(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}(this,e),cf(this,"spec",void 0),cf(this,"monitor",void 0),this.spec=t,this.monitor=n}return af(e,[{key:"canDrop",value:function(){var e=this.spec,t=this.monitor;return!e.canDrop||e.canDrop(t.getItem(),t)}},{key:"hover",value:function(){var e=this.spec,t=this.monitor;e.hover&&e.hover(t.getItem(),t)}},{key:"drop",value:function(){var e=this.spec,t=this.monitor;if(e.drop)return e.drop(t.getItem(),t)}}]),e}();function uf(e,t){return function(e){if(Array.isArray(e))return e}(e)||function(e,t){var n=null==e?null:"undefined"!=typeof Symbol&&e[Symbol.iterator]||e["@@iterator"];if(null==n)return;var r,o,i=[],s=!0,a=!1;try{for(n=n.call(e);!(s=(r=n.next()).done)&&(i.push(r.value),!t||i.length!==t);s=!0);}catch(Ih){a=!0,o=Ih}finally{try{s||null==n.return||n.return()}finally{if(a)throw o}}return i}(e,t)||function(e,t){if(!e)return;if("string"==typeof e)return df(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);"Object"===n&&e.constructor&&(n=e.constructor.name);if("Map"===n||"Set"===n)return Array.from(e);if("Arguments"===n||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return df(e,t)}(e,t)||function(){throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.")}()}function df(e,t){(null==t||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function ff(e,t,n){var r=zd(),o=function(e,t){var n=h.useMemo((function(){return new lf(e,t)}),[t]);return h.useEffect((function(){n.spec=e}),[e]),n}(e,t),i=function(e){var t=e.accept;return h.useMemo((function(){return Bl(null!=e.accept,"accept must be defined"),Array.isArray(t)?t:[t]}),[t])}(e);Fd((function(){var e=function(e,t,n){var r=n.getRegistry(),o=r.addTarget(e,t);return[o,function(){return r.removeTarget(o)}]}(i,o,r),s=uf(e,2),a=s[0],c=s[1];return t.receiveHandlerId(a),n.receiveHandlerId(a),c}),[r,t,o,n,i.map((function(e){return e.toString()})).join("|")])}function hf(e){return h.useMemo((function(){return e.hooks.dropTarget()}),[e])}function pf(e,t){var n,r=Yd(e,t),o=(n=zd(),h.useMemo((function(){return new Rd(n)}),[n])),i=function(e){var t=zd(),n=h.useMemo((function(){return new Ld(t.getBackend())}),[t]);return Fd((function(){return n.dropTargetOptions=e||null,n.reconnect(),function(){return n.disconnectDropTarget()}}),[e]),n}(r.options);return ff(r,o,i),[nf(r.collect,o,i),hf(i)]}function gf(e,t){return function(e){if(Array.isArray(e))return e}(e)||function(e,t){var n=null==e?null:"undefined"!=typeof Symbol&&e[Symbol.iterator]||e["@@iterator"];if(null==n)return;var r,o,i=[],s=!0,a=!1;try{for(n=n.call(e);!(s=(r=n.next()).done)&&(i.push(r.value),!t||i.length!==t);s=!0);}catch(Ih){a=!0,o=Ih}finally{try{s||null==n.return||n.return()}finally{if(a)throw o}}return i}(e,t)||function(e,t){if(!e)return;if("string"==typeof e)return mf(e,t);var n=Object.prototype.toString.call(e).slice(8,-1);"Object"===n&&e.constructor&&(n=e.constructor.name);if("Map"===n||"Set"===n)return Array.from(e);if("Arguments"===n||/^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n))return mf(e,t)}(e,t)||function(){throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.")}()}function mf(e,t){(null==t||t>e.length)&&(t=e.length);for(var n=0,r=new Array(t);n<t;n++)r[n]=e[n];return r}function vf(e){var t=null;return function(){return null==t&&(t=e()),t}}function yf(e,t,n){return t&&function(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,r.key,r)}}(e.prototype,t),e}function bf(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}var wf=function(){function e(t){!function(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}(this,e),bf(this,"entered",[]),bf(this,"isNodeInDocument",void 0),this.isNodeInDocument=t}return yf(e,[{key:"enter",value:function(e){var t=this,n=this.entered.length;return this.entered=function(e,t){var n=new Set,r=function(e){return n.add(e)};e.forEach(r),t.forEach(r);var o=[];return n.forEach((function(e){return o.push(e)})),o}(this.entered.filter((function(n){return t.isNodeInDocument(n)&&(!n.contains||n.contains(e))})),[e]),0===n&&this.entered.length>0}},{key:"leave",value:function(e){var t,n,r=this.entered.length;return this.entered=(t=this.entered.filter(this.isNodeInDocument),n=e,t.filter((function(e){return e!==n}))),r>0&&0===this.entered.length}},{key:"reset",value:function(){this.entered=[]}}]),e}(),xf=vf((function(){return/firefox/i.test(navigator.userAgent)})),Tf=vf((function(){return Boolean(window.safari)}));function Df(e,t,n){return t&&function(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,r.key,r)}}(e.prototype,t),e}function Sf(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}var Of=function(){function e(t,n){!function(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}(this,e),Sf(this,"xs",void 0),Sf(this,"ys",void 0),Sf(this,"c1s",void 0),Sf(this,"c2s",void 0),Sf(this,"c3s",void 0);for(var r=t.length,o=[],i=0;i<r;i++)o.push(i);o.sort((function(e,n){return t[e]<t[n]?-1:1}));for(var s,a,c=[],l=[],u=0;u<r-1;u++)s=t[u+1]-t[u],a=n[u+1]-n[u],c.push(s),l.push(a/s);for(var d=[l[0]],f=0;f<c.length-1;f++){var h=l[f],p=l[f+1];if(h*p<=0)d.push(0);else{s=c[f];var g=c[f+1],m=s+g;d.push(3*m/((m+g)/h+(m+s)/p))}}d.push(l[l.length-1]);for(var v,y=[],b=[],w=0;w<d.length-1;w++){v=l[w];var x=d[w],T=1/c[w],D=x+d[w+1]-v-v;y.push((v-x-D)*T),b.push(D*T*T)}this.xs=t,this.ys=n,this.c1s=d,this.c2s=y,this.c3s=b}return Df(e,[{key:"interpolate",value:function(e){var t=this.xs,n=this.ys,r=this.c1s,o=this.c2s,i=this.c3s,s=t.length-1;if(e===t[s])return n[s];for(var a,c=0,l=i.length-1;c<=l;){var u=t[a=Math.floor(.5*(c+l))];if(u<e)c=a+1;else{if(!(u>e))return n[a];l=a-1}}var d=e-t[s=Math.max(0,l)],f=d*d;return n[s]+r[s]*d+o[s]*f+i[s]*d*f}}]),e}();function Rf(e){var t=1===e.nodeType?e:e.parentElement;if(!t)return null;var n=t.getBoundingClientRect(),r=n.top;return{x:n.left,y:r}}function jf(e){return{x:e.clientX,y:e.clientY}}function Cf(e,t,n,r,o){var i,s,a,c="IMG"===(i=t).nodeName&&(xf()||!(null!==(s=document.documentElement)&&void 0!==s&&s.contains(i))),l=Rf(c?e:t),u={x:n.x-l.x,y:n.y-l.y},d=e.offsetWidth,f=e.offsetHeight,h=r.anchorX,p=r.anchorY,g=function(e,t,n,r){var o=e?t.width:n,i=e?t.height:r;return Tf()&&e&&(i/=window.devicePixelRatio,o/=window.devicePixelRatio),{dragPreviewWidth:o,dragPreviewHeight:i}}(c,t,d,f),m=g.dragPreviewWidth,v=g.dragPreviewHeight,y=o.offsetX,b=o.offsetY,w=0===b||b;return{x:0===y||y?y:new Of([0,.5,1],[u.x,u.x/d*m,u.x+m-d]).interpolate(h),y:w?b:(a=new Of([0,.5,1],[u.y,u.y/f*v,u.y+v-f]).interpolate(p),Tf()&&c&&(a+=(window.devicePixelRatio-1)*v),a)}}var Nf="__NATIVE_FILE__",kf="__NATIVE_URL__",Pf="__NATIVE_TEXT__",Ef="__NATIVE_HTML__";const If=Object.freeze(Object.defineProperty({__proto__:null,FILE:Nf,HTML:Ef,TEXT:Pf,URL:kf},Symbol.toStringTag,{value:"Module"}));function _f(e,t,n){var r=t.reduce((function(t,n){return t||e.getData(n)}),"");return null!=r?r:n}var Mf;function qf(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}var Af=(qf(Mf={},Nf,{exposeProperties:{files:function(e){return Array.prototype.slice.call(e.files)},items:function(e){return e.items},dataTransfer:function(e){return e}},matchesTypes:["Files"]}),qf(Mf,Ef,{exposeProperties:{html:function(e,t){return _f(e,t,"")},dataTransfer:function(e){return e}},matchesTypes:["Html","text/html"]}),qf(Mf,kf,{exposeProperties:{urls:function(e,t){return _f(e,t,"").split("\n")},dataTransfer:function(e){return e}},matchesTypes:["Url","text/uri-list"]}),qf(Mf,Pf,{exposeProperties:{text:function(e,t){return _f(e,t,"")},dataTransfer:function(e){return e}},matchesTypes:["Text","text/plain"]}),Mf);function Lf(e,t,n){return t&&function(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,r.key,r)}}(e.prototype,t),e}function Ff(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}var Hf=function(){function e(t){!function(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}(this,e),Ff(this,"item",void 0),Ff(this,"config",void 0),this.config=t,this.item={},this.initializeExposedProperties()}return Lf(e,[{key:"initializeExposedProperties",value:function(){var e=this;Object.keys(this.config.exposeProperties).forEach((function(t){Object.defineProperty(e.item,t,{configurable:!0,enumerable:!0,get:function(){return console.warn("Browser doesn't allow reading \"".concat(t,'" until the drop event.')),null}})}))}},{key:"loadDataTransfer",value:function(e){var t=this;if(e){var n={};Object.keys(this.config.exposeProperties).forEach((function(r){n[r]={value:t.config.exposeProperties[r](e,t.config.matchesTypes),configurable:!0,enumerable:!0}})),Object.defineProperties(this.item,n)}}},{key:"canDrag",value:function(){return!0}},{key:"beginDrag",value:function(){return this.item}},{key:"isDragging",value:function(e,t){return t===e.getSourceId()}},{key:"endDrag",value:function(){}}]),e}();function Wf(e){if(!e)return null;var t=Array.prototype.slice.call(e.types||[]);return Object.keys(Af).filter((function(e){return Af[e].matchesTypes.some((function(e){return t.indexOf(e)>-1}))}))[0]||null}function Uf(e,t,n){return t&&function(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,r.key,r)}}(e.prototype,t),e}function Bf(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}var zf=function(){function e(t,n){!function(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}(this,e),Bf(this,"ownerDocument",null),Bf(this,"globalContext",void 0),Bf(this,"optionsArgs",void 0),this.globalContext=t,this.optionsArgs=n}return Uf(e,[{key:"window",get:function(){return this.globalContext?this.globalContext:"undefined"!=typeof window?window:void 0}},{key:"document",get:function(){var e;return null!==(e=this.globalContext)&&void 0!==e&&e.document?this.globalContext.document:this.window?this.window.document:void 0}},{key:"rootElement",get:function(){var e;return(null===(e=this.optionsArgs)||void 0===e?void 0:e.rootElement)||this.window}}]),e}();function $f(e,t){var n=Object.keys(e);if(Object.getOwnPropertySymbols){var r=Object.getOwnPropertySymbols(e);t&&(r=r.filter((function(t){return Object.getOwnPropertyDescriptor(e,t).enumerable}))),n.push.apply(n,r)}return n}function Vf(e){for(var t=1;t<arguments.length;t++){var n=null!=arguments[t]?arguments[t]:{};t%2?$f(Object(n),!0).forEach((function(t){Gf(e,t,n[t])})):Object.getOwnPropertyDescriptors?Object.defineProperties(e,Object.getOwnPropertyDescriptors(n)):$f(Object(n)).forEach((function(t){Object.defineProperty(e,t,Object.getOwnPropertyDescriptor(n,t))}))}return e}function Kf(e,t,n){return t&&function(e,t){for(var n=0;n<t.length;n++){var r=t[n];r.enumerable=r.enumerable||!1,r.configurable=!0,"value"in r&&(r.writable=!0),Object.defineProperty(e,r.key,r)}}(e.prototype,t),e}function Gf(e,t,n){return t in e?Object.defineProperty(e,t,{value:n,enumerable:!0,configurable:!0,writable:!0}):e[t]=n,e}var Jf,Yf=function(){function e(t,n,r){var o=this;!function(e,t){if(!(e instanceof t))throw new TypeError("Cannot call a class as a function")}(this,e),Gf(this,"options",void 0),Gf(this,"actions",void 0),Gf(this,"monitor",void 0),Gf(this,"registry",void 0),Gf(this,"enterLeaveCounter",void 0),Gf(this,"sourcePreviewNodes",new Map),Gf(this,"sourcePreviewNodeOptions",new Map),Gf(this,"sourceNodes",new Map),Gf(this,"sourceNodeOptions",new Map),Gf(this,"dragStartSourceIds",null),Gf(this,"dropTargetIds",[]),Gf(this,"dragEnterTargetIds",[]),Gf(this,"currentNativeSource",null),Gf(this,"currentNativeHandle",null),Gf(this,"currentDragSourceNode",null),Gf(this,"altKeyPressed",!1),Gf(this,"mouseMoveTimeoutTimer",null),Gf(this,"asyncEndDragFrameId",null),Gf(this,"dragOverTargetIds",null),Gf(this,"lastClientOffset",null),Gf(this,"hoverRafId",null),Gf(this,"getSourceClientOffset",(function(e){var t=o.sourceNodes.get(e);return t&&Rf(t)||null})),Gf(this,"endDragNativeItem",(function(){o.isDraggingNativeItem()&&(o.actions.endDrag(),o.currentNativeHandle&&o.registry.removeSource(o.currentNativeHandle),o.currentNativeHandle=null,o.currentNativeSource=null)})),Gf(this,"isNodeInDocument",(function(e){return Boolean(e&&o.document&&o.document.body&&o.document.body.contains(e))})),Gf(this,"endDragIfSourceWasRemovedFromDOM",(function(){var e=o.currentDragSourceNode;null==e||o.isNodeInDocument(e)||o.clearCurrentDragSourceNode()&&o.monitor.isDragging()&&o.actions.endDrag()})),Gf(this,"handleTopDragStartCapture",(function(){o.clearCurrentDragSourceNode(),o.dragStartSourceIds=[]})),Gf(this,"handleTopDragStart",(function(e){if(!e.defaultPrevented){var t=o.dragStartSourceIds;o.dragStartSourceIds=null;var n=jf(e);o.monitor.isDragging()&&o.actions.endDrag(),o.actions.beginDrag(t||[],{publishSource:!1,getSourceClientOffset:o.getSourceClientOffset,clientOffset:n});var r=e.dataTransfer,i=Wf(r);if(o.monitor.isDragging()){if(r&&"function"==typeof r.setDragImage){var s=o.monitor.getSourceId(),a=o.sourceNodes.get(s),c=o.sourcePreviewNodes.get(s)||a;if(c){var l=o.getCurrentSourcePreviewNodeOptions(),u=Cf(a,c,n,{anchorX:l.anchorX,anchorY:l.anchorY},{offsetX:l.offsetX,offsetY:l.offsetY});r.setDragImage(c,u.x,u.y)}}try{null==r||r.setData("application/json",{})}catch(Ih){}o.setCurrentDragSourceNode(e.target),o.getCurrentSourcePreviewNodeOptions().captureDraggingState?o.actions.publishDragSource():setTimeout((function(){return o.actions.publishDragSource()}),0)}else if(i)o.beginDragNativeItem(i);else{if(r&&!r.types&&(e.target&&!e.target.hasAttribute||!e.target.hasAttribute("draggable")))return;e.preventDefault()}}})),Gf(this,"handleTopDragEndCapture",(function(){o.clearCurrentDragSourceNode()&&o.monitor.isDragging()&&o.actions.endDrag()})),Gf(this,"handleTopDragEnterCapture",(function(e){if(o.dragEnterTargetIds=[],o.enterLeaveCounter.enter(e.target)&&!o.monitor.isDragging()){var t=e.dataTransfer,n=Wf(t);n&&o.beginDragNativeItem(n,t)}})),Gf(this,"handleTopDragEnter",(function(e){var t=o.dragEnterTargetIds;(o.dragEnterTargetIds=[],o.monitor.isDragging())&&(o.altKeyPressed=e.altKey,t.length>0&&o.actions.hover(t,{clientOffset:jf(e)}),t.some((function(e){return o.monitor.canDropOnTarget(e)}))&&(e.preventDefault(),e.dataTransfer&&(e.dataTransfer.dropEffect=o.getCurrentDropEffect())))})),Gf(this,"handleTopDragOverCapture",(function(){o.dragOverTargetIds=[]})),Gf(this,"handleTopDragOver",(function(e){var t=o.dragOverTargetIds;if(o.dragOverTargetIds=[],!o.monitor.isDragging())return e.preventDefault(),void(e.dataTransfer&&(e.dataTransfer.dropEffect="none"));o.altKeyPressed=e.altKey,o.lastClientOffset=jf(e),null===o.hoverRafId&&"undefined"!=typeof requestAnimationFrame&&(o.hoverRafId=requestAnimationFrame((function(){o.monitor.isDragging()&&o.actions.hover(t||[],{clientOffset:o.lastClientOffset}),o.hoverRafId=null}))),(t||[]).some((function(e){return o.monitor.canDropOnTarget(e)}))?(e.preventDefault(),e.dataTransfer&&(e.dataTransfer.dropEffect=o.getCurrentDropEffect())):o.isDraggingNativeItem()?e.preventDefault():(e.preventDefault(),e.dataTransfer&&(e.dataTransfer.dropEffect="none"))})),Gf(this,"handleTopDragLeaveCapture",(function(e){o.isDraggingNativeItem()&&e.preventDefault(),o.enterLeaveCounter.leave(e.target)&&o.isDraggingNativeItem()&&setTimeout((function(){return o.endDragNativeItem()}),0)})),Gf(this,"handleTopDropCapture",(function(e){var t;(o.dropTargetIds=[],o.isDraggingNativeItem())?(e.preventDefault(),null===(t=o.currentNativeSource)||void 0===t||t.loadDataTransfer(e.dataTransfer)):Wf(e.dataTransfer)&&e.preventDefault();o.enterLeaveCounter.reset()})),Gf(this,"handleTopDrop",(function(e){var t=o.dropTargetIds;o.dropTargetIds=[],o.actions.hover(t,{clientOffset:jf(e)}),o.actions.drop({dropEffect:o.getCurrentDropEffect()}),o.isDraggingNativeItem()?o.endDragNativeItem():o.monitor.isDragging()&&o.actions.endDrag()})),Gf(this,"handleSelectStart",(function(e){var t=e.target;"function"==typeof t.dragDrop&&("INPUT"===t.tagName||"SELECT"===t.tagName||"TEXTAREA"===t.tagName||t.isContentEditable||(e.preventDefault(),t.dragDrop()))})),this.options=new zf(n,r),this.actions=t.getActions(),this.monitor=t.getMonitor(),this.registry=t.getRegistry(),this.enterLeaveCounter=new wf(this.isNodeInDocument)}return Kf(e,[{key:"profile",value:function(){var e,t;return{sourcePreviewNodes:this.sourcePreviewNodes.size,sourcePreviewNodeOptions:this.sourcePreviewNodeOptions.size,sourceNodeOptions:this.sourceNodeOptions.size,sourceNodes:this.sourceNodes.size,dragStartSourceIds:(null===(e=this.dragStartSourceIds)||void 0===e?void 0:e.length)||0,dropTargetIds:this.dropTargetIds.length,dragEnterTargetIds:this.dragEnterTargetIds.length,dragOverTargetIds:(null===(t=this.dragOverTargetIds)||void 0===t?void 0:t.length)||0}}},{key:"window",get:function(){return this.options.window}},{key:"document",get:function(){return this.options.document}},{key:"rootElement",get:function(){return this.options.rootElement}},{key:"setup",value:function(){var e=this.rootElement;if(void 0!==e){if(e.__isReactDndBackendSetUp)throw new Error("Cannot have two HTML5 backends at the same time.");e.__isReactDndBackendSetUp=!0,this.addEventListeners(e)}}},{key:"teardown",value:function(){var e,t=this.rootElement;void 0!==t&&(t.__isReactDndBackendSetUp=!1,this.removeEventListeners(this.rootElement),this.clearCurrentDragSourceNode(),this.asyncEndDragFrameId&&(null===(e=this.window)||void 0===e||e.cancelAnimationFrame(this.asyncEndDragFrameId)))}},{key:"connectDragPreview",value:function(e,t,n){var r=this;return this.sourcePreviewNodeOptions.set(e,n),this.sourcePreviewNodes.set(e,t),function(){r.sourcePreviewNodes.delete(e),r.sourcePreviewNodeOptions.delete(e)}}},{key:"connectDragSource",value:function(e,t,n){var r=this;this.sourceNodes.set(e,t),this.sourceNodeOptions.set(e,n);var o=function(t){return r.handleDragStart(t,e)},i=function(e){return r.handleSelectStart(e)};return t.setAttribute("draggable","true"),t.addEventListener("dragstart",o),t.addEventListener("selectstart",i),function(){r.sourceNodes.delete(e),r.sourceNodeOptions.delete(e),t.removeEventListener("dragstart",o),t.removeEventListener("selectstart",i),t.setAttribute("draggable","false")}}},{key:"connectDropTarget",value:function(e,t){var n=this,r=function(t){return n.handleDragEnter(t,e)},o=function(t){return n.handleDragOver(t,e)},i=function(t){return n.handleDrop(t,e)};return t.addEventListener("dragenter",r),t.addEventListener("dragover",o),t.addEventListener("drop",i),function(){t.removeEventListener("dragenter",r),t.removeEventListener("dragover",o),t.removeEventListener("drop",i)}}},{key:"addEventListeners",value:function(e){e.addEventListener&&(e.addEventListener("dragstart",this.handleTopDragStart),e.addEventListener("dragstart",this.handleTopDragStartCapture,!0),e.addEventListener("dragend",this.handleTopDragEndCapture,!0),e.addEventListener("dragenter",this.handleTopDragEnter),e.addEventListener("dragenter",this.handleTopDragEnterCapture,!0),e.addEventListener("dragleave",this.handleTopDragLeaveCapture,!0),e.addEventListener("dragover",this.handleTopDragOver),e.addEventListener("dragover",this.handleTopDragOverCapture,!0),e.addEventListener("drop",this.handleTopDrop),e.addEventListener("drop",this.handleTopDropCapture,!0))}},{key:"removeEventListeners",value:function(e){e.removeEventListener&&(e.removeEventListener("dragstart",this.handleTopDragStart),e.removeEventListener("dragstart",this.handleTopDragStartCapture,!0),e.removeEventListener("dragend",this.handleTopDragEndCapture,!0),e.removeEventListener("dragenter",this.handleTopDragEnter),e.removeEventListener("dragenter",this.handleTopDragEnterCapture,!0),e.removeEventListener("dragleave",this.handleTopDragLeaveCapture,!0),e.removeEventListener("dragover",this.handleTopDragOver),e.removeEventListener("dragover",this.handleTopDragOverCapture,!0),e.removeEventListener("drop",this.handleTopDrop),e.removeEventListener("drop",this.handleTopDropCapture,!0))}},{key:"getCurrentSourceNodeOptions",value:function(){var e=this.monitor.getSourceId(),t=this.sourceNodeOptions.get(e);return Vf({dropEffect:this.altKeyPressed?"copy":"move"},t||{})}},{key:"getCurrentDropEffect",value:function(){return this.isDraggingNativeItem()?"copy":this.getCurrentSourceNodeOptions().dropEffect}},{key:"getCurrentSourcePreviewNodeOptions",value:function(){var e=this.monitor.getSourceId();return Vf({anchorX:.5,anchorY:.5,captureDraggingState:!1},this.sourcePreviewNodeOptions.get(e)||{})}},{key:"isDraggingNativeItem",value:function(){var e=this.monitor.getItemType();return Object.keys(If).some((function(t){return If[t]===e}))}},{key:"beginDragNativeItem",value:function(e,t){this.clearCurrentDragSourceNode(),this.currentNativeSource=function(e,t){var n=new Hf(Af[e]);return n.loadDataTransfer(t),n}(e,t),this.currentNativeHandle=this.registry.addSource(e,this.currentNativeSource),this.actions.beginDrag([this.currentNativeHandle])}},{key:"setCurrentDragSourceNode",value:function(e){var t=this;this.clearCurrentDragSourceNode(),this.currentDragSourceNode=e;this.mouseMoveTimeoutTimer=setTimeout((function(){var e;return null===(e=t.rootElement)||void 0===e?void 0:e.addEventListener("mousemove",t.endDragIfSourceWasRemovedFromDOM,!0)}),1e3)}},{key:"clearCurrentDragSourceNode",value:function(){if(this.currentDragSourceNode){var e;if(this.currentDragSourceNode=null,this.rootElement)null===(e=this.window)||void 0===e||e.clearTimeout(this.mouseMoveTimeoutTimer||void 0),this.rootElement.removeEventListener("mousemove",this.endDragIfSourceWasRemovedFromDOM,!0);return this.mouseMoveTimeoutTimer=null,!0}return!1}},{key:"handleDragStart",value:function(e,t){e.defaultPrevented||(this.dragStartSourceIds||(this.dragStartSourceIds=[]),this.dragStartSourceIds.unshift(t))}},{key:"handleDragEnter",value:function(e,t){this.dragEnterTargetIds.unshift(t)}},{key:"handleDragOver",value:function(e,t){null===this.dragOverTargetIds&&(this.dragOverTargetIds=[]),this.dragOverTargetIds.unshift(t)}},{key:"handleDrop",value:function(e,t){this.dropTargetIds.unshift(t)}}]),e}();var Xf=function(e,t,n){return new Yf(e,t,n)};function Zf(e){const t=wc(),n=t.selectedIds,[r,o,i]=sf((()=>({canDrag:()=>e.isDraggable,type:"NODE",item:()=>{const r=t.isSelected(e.id)?Array.from(n):[e.id];return t.dispatch(pl.dragStart(e.id,r)),{id:e.id}},end:()=>{t.hideCursor();let{parentId:e,index:n,dragIds:r}=t.state.dnd;t.canDrop()&&(Wc(t.props.onMove,{dragIds:r,parentId:e===rl?null:e,index:null===n?0:n,dragNodes:t.dragNodes,parentNode:t.get(e)}),t.open(e)),t.dispatch(pl.dragEnd())}})),[n,e]);return h.useEffect((()=>{i((Jf||((Jf=new Image).src="data:image/gif;base64,R0lGODlhAQABAAAAACH5BAEKAAEALAAAAAABAAEAAAICTAEAOw=="),Jf))}),[i]),o}function Qf(e,t){return{parentId:e||null,index:t}}function eh(e,t){return{type:"line",index:e,level:t}}function th(e,t){var n;let r=e;for(;r.parent&&r.level>t;)r=r.parent;return{parentId:(null===(n=r.parent)||void 0===n?void 0:n.id)||null,index:kc(r)+1}}function nh(e){var t;const n=function(e,t){const n=e.getBoundingClientRect(),r=t.x-Math.round(n.x),o=t.y-Math.round(n.y),i=n.height,s=o<i/2,a=!s,c=i/4,l=o>c&&o<i-c;return{x:r,inTopHalf:s,inBottomHalf:a,inMiddle:l,atTop:!l&&s,atBottom:!l&&a}}(e.element,e.offset),r=e.indent,o=Math.round(Math.max(0,n.x-r)/r),{node:i,nextNode:s,prevNode:a}=e,[c,l]=function(e,t,n,r){return e?e.isInternal?r.atTop?[t,e]:r.inMiddle?[e,e]:[e,n]:r.inTopHalf?[t,e]:[e,n]:[t,null]}(i,a,s,n);if(i&&i.isInternal&&n.inMiddle)return{drop:Qf(i.id,null),cursor:(u=i.id,{type:"highlight",id:u})};var u;if(!c)return{drop:Qf(null===(t=null==l?void 0:l.parent)||void 0===t?void 0:t.id,0),cursor:eh(0,0)};if(Rc(c)){const e=Oc(o,(null==l?void 0:l.level)||0,c.level);return{drop:th(c,e),cursor:eh(c.rowIndex+1,e)}}if(jc(c)){const e=Oc(o,(null==l?void 0:l.level)||0,c.level);return{drop:th(c,e),cursor:eh(c.rowIndex+1,e)}}if(Cc(c)){const e=Oc(o,0,c.level+1);return e>c.level?{drop:Qf(c.id,0),cursor:eh(c.rowIndex+1,e)}:{drop:th(c,e),cursor:eh(c.rowIndex+1,e)}}return{drop:Qf(null==c?void 0:c.id,0),cursor:eh(c.rowIndex+1,c.level+1)}}const rh=o.memo((function({index:e,style:t}){Sc(),function(){const e=h.useContext(xc);if(null===e)throw new Error("Provide a NodesContext")}();const n=wc(),r=function(e){const t=wc(),n=t.at(e);if(!n)throw new Error(`Could not find node for index: ${e}`);return h.useMemo((()=>{const r=n.clone();return t.visibleNodes[e]=r,r}),[...Object.values(n.state),n])}(e),o=h.useRef(null),i=Zf(r),s=function(e,t){const n=wc(),[r,o]=pf((()=>({accept:"NODE",canDrop:()=>n.canDrop(),hover:(r,o)=>{const i=o.getClientOffset();if(!e.current||!i)return;const{cursor:s,drop:a}=nh({element:e.current,offset:i,indent:n.indent,node:t,prevNode:t.prev,nextNode:t.next});a&&n.dispatch(pl.hovering(a.parentId,a.index)),o.canDrop()?s&&n.showCursor(s):n.hideCursor()},drop:(e,t)=>{if(!t.canDrop())return null}})),[t,e.current,n.props]);return o}(o,r),c=h.useCallback((e=>{o.current=e,s(e)}),[s]),l=n.indent*r.level,u=h.useMemo((()=>({paddingLeft:l})),[l]),d=h.useMemo((()=>{var e,r;return Object.assign(Object.assign({},t),{top:parseFloat(t.top)+(null!==(r=null!==(e=n.props.padding)&&void 0!==e?e:n.props.paddingTop)&&void 0!==r?r:0)})}),[t,n.props.padding,n.props.paddingTop]),f={role:"treeitem","aria-level":r.level+1,"aria-selected":r.isSelected,style:d,tabIndex:-1,className:n.props.rowClassName};h.useEffect((()=>{var e;!r.isEditing&&r.isFocused&&(null===(e=o.current)||void 0===e||e.focus({preventScroll:!0}))}),[r.isEditing,r.isFocused,o.current]);const p=n.renderNode,g=n.renderRow;return a.jsx(g,{node:r,innerRef:c,attrs:f,children:a.jsx(p,{node:r,tree:n,style:u,dragHandle:i})})}));let oh="",ih=null;function sh(){Sc();const e=wc();return a.jsx("div",{role:"tree",style:{height:e.height,width:e.width,minHeight:0,minWidth:0},onContextMenu:e.props.onContextMenu,onClick:e.props.onClick,tabIndex:0,onFocus:t=>{t.currentTarget.contains(t.relatedTarget)||e.onFocus()},onBlur:t=>{t.currentTarget.contains(t.relatedTarget)||e.onBlur()},onKeyDown:t=>{var n;if(e.isEditing)return;if("Backspace"===t.key){if(!e.props.onDelete)return;const t=Array.from(e.selectedIds);if(t.length>1){let n=e.mostRecentNode;for(;n&&n.isSelected;)n=n.nextSibling;n||(n=e.lastNode),e.focus(n,{scroll:!1}),e.delete(Array.from(t))}else{const t=e.focusedNode;if(t){const n=t.nextSibling,r=t.parent;e.focus(n||r,{scroll:!1}),e.delete(t)}}return}if("Tab"===t.key&&!t.shiftKey)return t.preventDefault(),void Ic(t.currentTarget);if("Tab"===t.key&&t.shiftKey)return t.preventDefault(),void _c(t.currentTarget);if("ArrowDown"===t.key){t.preventDefault();const n=e.nextNode;if(t.metaKey)return e.select(e.focusedNode),void e.activate(e.focusedNode);if(!t.shiftKey||e.props.disableMultiSelection)return void e.focus(n);{if(!n)return;const t=e.focusedNode;return void(t?t.isSelected?e.selectContiguous(n):e.selectMulti(n):e.focus(e.firstNode))}}if("ArrowUp"===t.key){t.preventDefault();const n=e.prevNode;if(!t.shiftKey||e.props.disableMultiSelection)return void e.focus(n);{if(!n)return;const t=e.focusedNode;return void(t?t.isSelected?e.selectContiguous(n):e.selectMulti(n):e.focus(e.lastNode))}}if("ArrowRight"===t.key){const t=e.focusedNode;if(!t)return;return void(t.isInternal&&t.isOpen?e.focus(e.nextNode):t.isInternal&&e.open(t.id))}if("ArrowLeft"===t.key){const t=e.focusedNode;if(!t||t.isRoot)return;return void(t.isInternal&&t.isOpen?e.close(t.id):(null===(n=t.parent)||void 0===n?void 0:n.isRoot)||e.focus(t.parent))}if("a"===t.key&&t.metaKey&&!e.props.disableMultiSelection)return t.preventDefault(),void e.selectAll();if("a"===t.key&&!t.metaKey&&e.props.onCreate)return void e.createLeaf();if("A"===t.key&&!t.metaKey){if(!e.props.onCreate)return;return void e.createInternal()}if("Home"===t.key)return t.preventDefault(),void e.focus(e.firstNode);if("End"===t.key)return t.preventDefault(),void e.focus(e.lastNode);if("Enter"===t.key){const t=e.focusedNode;if(!t)return;if(!t.isEditable||!e.props.onRename)return;return void setTimeout((()=>{t&&e.edit(t)}))}if(" "===t.key){t.preventDefault();const n=e.focusedNode;if(!n)return;return void(n.isLeaf?(n.select(),n.activate()):n.toggle())}if("*"===t.key){const t=e.focusedNode;if(!t)return;return void e.openSiblings(t)}if("PageUp"===t.key)return t.preventDefault(),void e.pageUp();"PageDown"===t.key&&(t.preventDefault(),e.pageDown()),clearTimeout(ih),oh+=t.key,ih=setTimeout((()=>{oh=""}),600);const r=e.visibleNodes.find((e=>{const t=e.data.name;return"string"==typeof t&&t.toLowerCase().startsWith(oh)}));r&&e.focus(r.id)},children:a.jsx(Ml,{className:e.props.className,outerRef:e.listEl,itemCount:e.visibleNodes.length,height:e.height,width:e.width,itemSize:e.rowHeight,overscanCount:e.overscanCount,itemKey:t=>{var n;return(null===(n=e.visibleNodes[t])||void 0===n?void 0:n.id)||t},outerElementType:Al,innerElementType:Fl,onScroll:e.props.onScroll,onItemsRendered:e.onItemsRendered.bind(e),ref:e.list,children:rh})})}function ah(e){return e.isFiltered?function(e,t){const n={},r=[];function o(e){if(!e.isRoot&&t(e)){n[e.id]=!0;let t=e.parent;for(;t;)n[t.id]=!0,t=t.parent}if(e.children)for(let t of e.children)o(t)}function i(e){var t;e.level>=0&&n[e.id]&&r.push(e),e.isOpen&&(null===(t=e.children)||void 0===t||t.forEach(i))}return o(e),i(e),r.forEach(ch),r}(e.root,e.isMatch.bind(e)):function(e){const t=[];function n(e){var r;e.level>=0&&t.push(e),e.isOpen&&(null===(r=e.children)||void 0===r||r.forEach(n))}return n(e),t.forEach(ch),t}(e.root)}function ch(e,t){e.rowIndex=t}const lh=e=>e.reduce(((e,t,n)=>(e[t.id]=n,e)),{});var uh=function(e,t,n,r){return new(n||(n=Promise))((function(o,i){function s(e){try{c(r.next(e))}catch(t){i(t)}}function a(e){try{c(r.throw(e))}catch(t){i(t)}}function c(e){var t;e.done?o(e.value):(t=e.value,t instanceof n?t:new n((function(e){e(t)}))).then(s,a)}c((r=r.apply(e,t||[])).next())}))};const{safeRun:dh,identify:fh,identifyNull:hh}=$c;class ph{constructor(e,t,n,r){this.store=e,this.props=t,this.list=n,this.listEl=r,this.visibleStartIndex=0,this.visibleStopIndex=0,this.root=ol(this),this.visibleNodes=ah(this),this.idToIndex=lh(this.visibleNodes)}update(e){this.props=e,this.root=ol(this),this.visibleNodes=ah(this),this.idToIndex=lh(this.visibleNodes)}dispatch(e){return this.store.dispatch(e)}get state(){return this.store.getState()}get openState(){return this.state.nodes.open.unfiltered}get width(){var e;return null!==(e=this.props.width)&&void 0!==e?e:300}get height(){var e;return null!==(e=this.props.height)&&void 0!==e?e:500}get indent(){var e;return null!==(e=this.props.indent)&&void 0!==e?e:24}get rowHeight(){var e;return null!==(e=this.props.rowHeight)&&void 0!==e?e:24}get overscanCount(){var e;return null!==(e=this.props.overscanCount)&&void 0!==e?e:1}get searchTerm(){return(this.props.searchTerm||"").trim()}get matchFn(){var e;const t=null!==(e=this.props.searchMatch)&&void 0!==e?e:(e,t)=>JSON.stringify(Object.values(e.data)).toLocaleLowerCase().includes(t.toLocaleLowerCase());return e=>t(e,this.searchTerm)}accessChildren(e){var t;return null!==(t=Lc(e,this.props.childrenAccessor||"children"))&&void 0!==t?t:null}accessId(e){const t=Lc(e,this.props.idAccessor||"id");if(!t)throw new Error("Data must contain an 'id' property or props.idAccessor must return a string");return t}get firstNode(){var e;return null!==(e=this.visibleNodes[0])&&void 0!==e?e:null}get lastNode(){var e;return null!==(e=this.visibleNodes[this.visibleNodes.length-1])&&void 0!==e?e:null}get focusedNode(){var e;return null!==(e=this.get(this.state.nodes.focus.id))&&void 0!==e?e:null}get mostRecentNode(){var e;return null!==(e=this.get(this.state.nodes.selection.mostRecent))&&void 0!==e?e:null}get nextNode(){const e=this.indexOf(this.focusedNode);return null===e?null:this.at(e+1)}get prevNode(){const e=this.indexOf(this.focusedNode);return null===e?null:this.at(e-1)}get(e){return e&&e in this.idToIndex&&this.visibleNodes[this.idToIndex[e]]||null}at(e){return this.visibleNodes[e]||null}nodesBetween(e,t){var n;if(null===e||null===t)return[];const r=null!==(n=this.indexOf(e))&&void 0!==n?n:0,o=this.indexOf(t);if(null===o)return[];const i=Math.min(r,o),s=Math.max(r,o);return this.visibleNodes.slice(i,s+1)}indexOf(e){const t=Fc(e);return t?this.idToIndex[t]:null}get editingId(){return this.state.nodes.edit.id}createInternal(){return this.create({type:"internal"})}createLeaf(){return this.create({type:"leaf"})}create(e={}){var t,n;return uh(this,void 0,void 0,(function*(){const r=void 0===e.parentId?zc(this):e.parentId,o=null!==(t=e.index)&&void 0!==t?t:Bc(this),i=null!==(n=e.type)&&void 0!==n?n:"leaf",s=yield dh(this.props.onCreate,{type:i,parentId:r,index:o,parentNode:this.get(r)});s&&(this.focus(s),setTimeout((()=>{this.edit(s).then((()=>{this.select(s),this.activate(s)}))})))}))}delete(e){return uh(this,void 0,void 0,(function*(){if(!e)return;const t=(Array.isArray(e)?e:[e]).map(fh),n=t.map((e=>this.get(e))).filter((e=>!!e));yield dh(this.props.onDelete,{nodes:n,ids:t})}))}edit(e){const t=fh(e);return this.resolveEdit({cancelled:!0}),this.scrollTo(t),this.dispatch(el(t)),new Promise((e=>{ph.editPromise=e}))}submit(e,t){return uh(this,void 0,void 0,(function*(){if(!e)return;const n=fh(e);yield dh(this.props.onRename,{id:n,name:t,node:this.get(n)}),this.dispatch(el(null)),this.resolveEdit({cancelled:!1,value:t}),setTimeout((()=>this.onFocus()))}))}reset(){this.dispatch(el(null)),this.resolveEdit({cancelled:!0}),setTimeout((()=>this.onFocus()))}activate(e){const t=this.get(hh(e));t&&dh(this.props.onActivate,t)}resolveEdit(e){const t=ph.editPromise;t&&t(e),ph.editPromise=null}get selectedIds(){return this.state.nodes.selection.ids}get selectedNodes(){let e=[];for(let t of Array.from(this.selectedIds)){const n=this.get(t);n&&e.push(n)}return e}focus(e,t={}){e&&(this.props.selectionFollowsFocus?this.select(e):(this.dispatch(tl(fh(e))),!1!==t.scroll&&this.scrollTo(e),this.focusedNode&&dh(this.props.onFocus,this.focusedNode)))}pageUp(){var e,t;const n=this.visibleStartIndex,r=this.visibleStopIndex-n;let o=null!==(t=null===(e=this.focusedNode)||void 0===e?void 0:e.rowIndex)&&void 0!==t?t:0;o=o>n?n:Math.max(n-r,0),this.focus(this.at(o))}pageDown(){var e,t;const n=this.visibleStartIndex,r=this.visibleStopIndex,o=r-n;let i=null!==(t=null===(e=this.focusedNode)||void 0===e?void 0:e.rowIndex)&&void 0!==t?t:0;i=i<r?r:Math.min(i+o,this.visibleNodes.length-1),this.focus(this.at(i))}select(e,t={}){if(!e)return;const n=!1!==t.focus,r=fh(e);n&&this.dispatch(tl(r)),this.dispatch(cl(r)),this.dispatch(hl(r)),this.dispatch(fl(r)),this.scrollTo(r,t.align),this.focusedNode&&n&&dh(this.props.onFocus,this.focusedNode),dh(this.props.onSelect,this.selectedNodes)}deselect(e){if(!e)return;const t=fh(e);this.dispatch(ul(t))}selectMulti(e){const t=this.get(hh(e));t&&(this.dispatch(tl(t.id)),this.dispatch(ll(t.id)),this.dispatch(hl(t.id)),this.dispatch(fl(t.id)),this.scrollTo(t),this.focusedNode&&dh(this.props.onFocus,this.focusedNode),dh(this.props.onSelect,this.selectedNodes))}selectContiguous(e){if(!e)return;const t=fh(e),{anchor:n,mostRecent:r}=this.state.nodes.selection;this.dispatch(tl(t)),this.dispatch(ul(this.nodesBetween(n,r))),this.dispatch(ll(this.nodesBetween(n,hh(t)))),this.dispatch(fl(t)),this.scrollTo(t),this.focusedNode&&dh(this.props.onFocus,this.focusedNode),dh(this.props.onSelect,this.selectedNodes)}deselectAll(){this.setSelection({ids:[],anchor:null,mostRecent:null}),dh(this.props.onSelect,this.selectedNodes)}selectAll(){var e;this.setSelection({ids:Object.keys(this.idToIndex),anchor:this.firstNode,mostRecent:this.lastNode}),this.dispatch(tl(null===(e=this.lastNode)||void 0===e?void 0:e.id)),this.focusedNode&&dh(this.props.onFocus,this.focusedNode),dh(this.props.onSelect,this.selectedNodes)}setSelection(e){var t;const n=new Set(null===(t=e.ids)||void 0===t?void 0:t.map(fh)),r=hh(e.anchor),o=hh(e.mostRecent);this.dispatch(dl({ids:n,anchor:r,mostRecent:o})),dh(this.props.onSelect,this.selectedNodes)}get cursorParentId(){const{cursor:e}=this.state.dnd;return"highlight"===e.type?e.id:null}get cursorOverFolder(){return"highlight"===this.state.dnd.cursor.type}get dragNodes(){return this.state.dnd.dragIds.map((e=>this.get(e))).filter((e=>!!e))}get dragNode(){return this.get(this.state.nodes.drag.id)}get dragDestinationParent(){return this.get(this.state.nodes.drag.destinationParentId)}get dragDestinationIndex(){return this.state.nodes.drag.destinationIndex}canDrop(){var e;if(this.isFiltered)return!1;const t=null!==(e=this.get(this.state.dnd.parentId))&&void 0!==e?e:this.root,n=this.dragNodes,r=this.props.disableDrop;for(const o of n){if(!o)return!1;if(!t)return!1;if(o.isInternal&&Nc(t,o))return!1}return"function"==typeof r?!r({parentNode:t,dragNodes:this.dragNodes,index:this.state.dnd.index||0}):"string"==typeof r?!t.data[r]:"boolean"!=typeof r||!r}hideCursor(){this.dispatch(pl.cursor({type:"none"}))}showCursor(e){this.dispatch(pl.cursor(e))}open(e){const t=hh(e);t&&(this.isOpen(t)||(this.dispatch(il.open(t,this.isFiltered)),dh(this.props.onToggle,t)))}close(e){const t=hh(e);t&&this.isOpen(t)&&(this.dispatch(il.close(t,this.isFiltered)),dh(this.props.onToggle,t))}toggle(e){const t=hh(e);if(t)return this.isOpen(t)?this.close(t):this.open(t)}openParents(e){const t=hh(e);if(!t)return;const n=Pc(this.root,t);let r=null==n?void 0:n.parent;for(;r;)this.open(r.id),r=r.parent}openSiblings(e){const t=e.parent;if(t){if(t.children){const n=e.isOpen;for(let e of t.children)e.isInternal&&(n?this.close(e.id):this.open(e.id));this.scrollTo(this.focusedNode)}}else this.toggle(e.id)}openAll(){Ec(this.root,(e=>{e.isInternal&&e.open()}))}closeAll(){Ec(this.root,(e=>{e.isInternal&&e.close()}))}scrollTo(e,t="smart"){if(!e)return;const n=fh(e);return this.openParents(n),Uc((()=>n in this.idToIndex)).then((()=>{var e;const r=this.idToIndex[n];void 0!==r&&(null===(e=this.list.current)||void 0===e||e.scrollToItem(r,t))})).catch((()=>{}))}get isEditing(){return null!==this.state.nodes.edit.id}get isFiltered(){var e;return!!(null===(e=this.props.searchTerm)||void 0===e?void 0:e.trim())}get hasFocus(){return this.state.nodes.focus.treeFocused}get hasNoSelection(){return 0===this.state.nodes.selection.ids.size}get hasOneSelection(){return 1===this.state.nodes.selection.ids.size}get hasMultipleSelections(){return this.state.nodes.selection.ids.size>1}isSelected(e){return!!e&&this.state.nodes.selection.ids.has(e)}isOpen(e){var t,n,r;if(!e)return!1;if(e===rl)return!0;const o=null===(t=this.props.openByDefault)||void 0===t||t;return this.isFiltered?null===(n=this.state.nodes.open.filtered[e])||void 0===n||n:null!==(r=this.state.nodes.open.unfiltered[e])&&void 0!==r?r:o}isEditable(e){var t;return null===(t=!Lc(e,this.props.disableEdit||(()=>!1)))||void 0===t||t}isDraggable(e){var t;return null===(t=!Lc(e,this.props.disableDrag||(()=>!1)))||void 0===t||t}isDragging(e){const t=hh(e);return!!t&&this.state.nodes.drag.id===t}isFocused(e){return this.hasFocus&&this.state.nodes.focus.id===e}isMatch(e){return this.matchFn(e)}willReceiveDrop(e){const t=hh(e);if(!t)return!1;const{destinationParentId:n,destinationIndex:r}=this.state.nodes.drag;return t===n&&null===r}onFocus(){const e=this.focusedNode||this.firstNode;e&&this.dispatch(tl(e.id))}onBlur(){this.dispatch({type:"TREE_BLUR"})}onItemsRendered(e){this.visibleStartIndex=e.visibleStartIndex,this.visibleStopIndex=e.visibleStopIndex}get renderContainer(){return this.props.renderContainer||sh}get renderRow(){return this.props.renderRow||Yc}get renderNode(){return this.props.children||Xc}get renderDragPreview(){return this.props.renderDragPreview||yl}get renderCursor(){return this.props.renderCursor||Jc}}function gh(e){return`Minified Redux error #${e}; visit https://redux.js.org/Errors?code=${e} for the full message or use the non-minified dev environment for full errors. `}var mh=(()=>"function"==typeof Symbol&&Symbol.observable||"@@observable")(),vh=()=>Math.random().toString(36).substring(7).split("").join("."),yh={INIT:`@@redux/INIT${vh()}`,REPLACE:`@@redux/REPLACE${vh()}`,PROBE_UNKNOWN_ACTION:()=>`@@redux/PROBE_UNKNOWN_ACTION${vh()}`};function bh(e,t,n){if("function"!=typeof e)throw new Error(gh(2));if("function"==typeof t&&"function"==typeof n||"function"==typeof n&&"function"==typeof arguments[3])throw new Error(gh(0));if("function"==typeof t&&void 0===n&&(n=t,t=void 0),void 0!==n){if("function"!=typeof n)throw new Error(gh(1));return n(bh)(e,t)}let r=e,o=t,i=new Map,s=i,a=0,c=!1;function l(){s===i&&(s=new Map,i.forEach(((e,t)=>{s.set(t,e)})))}function u(){if(c)throw new Error(gh(3));return o}function d(e){if("function"!=typeof e)throw new Error(gh(4));if(c)throw new Error(gh(5));let t=!0;l();const n=a++;return s.set(n,e),function(){if(t){if(c)throw new Error(gh(6));t=!1,l(),s.delete(n),i=null}}}function f(e){if(!function(e){if("object"!=typeof e||null===e)return!1;let t=e;for(;null!==Object.getPrototypeOf(t);)t=Object.getPrototypeOf(t);return Object.getPrototypeOf(e)===t||null===Object.getPrototypeOf(e)}(e))throw new Error(gh(7));if(void 0===e.type)throw new Error(gh(8));if("string"!=typeof e.type)throw new Error(gh(17));if(c)throw new Error(gh(9));try{c=!0,o=r(o,e)}finally{c=!1}return(i=s).forEach((e=>{e()})),e}f({type:yh.INIT});return{dispatch:f,subscribe:d,getState:u,replaceReducer:function(e){if("function"!=typeof e)throw new Error(gh(10));r=e,f({type:yh.REPLACE})},[mh]:function(){const e=d;return{subscribe(t){if("object"!=typeof t||null===t)throw new Error(gh(11));function n(){const e=t;e.next&&e.next(u())}n();return{unsubscribe:e(n)}},[mh](){return this}}}}}function wh(e){const t=Object.keys(e),n={};for(let s=0;s<t.length;s++){const r=t[s];"function"==typeof e[r]&&(n[r]=e[r])}const r=Object.keys(n);let o;try{!function(e){Object.keys(e).forEach((t=>{const n=e[t];if(void 0===n(void 0,{type:yh.INIT}))throw new Error(gh(12));if(void 0===n(void 0,{type:yh.PROBE_UNKNOWN_ACTION()}))throw new Error(gh(13))}))}(n)}catch(i){o=i}return function(e={},t){if(o)throw o;let i=!1;const s={};for(let o=0;o<r.length;o++){const a=r[o],c=n[a],l=e[a],u=c(l,t);if(void 0===u)throw t&&t.type,new Error(gh(14));s[a]=u,i=i||u!==l}return i=i||r.length!==Object.keys(e).length,i?s:e}}const xh=wh({nodes:wh({focus:function(e={id:null,treeFocused:!1},t){return"FOCUS"===t.type?Object.assign(Object.assign({},e),{id:t.id,treeFocused:!0}):"TREE_BLUR"===t.type?Object.assign(Object.assign({},e),{treeFocused:!1}):e},edit:function(e={id:null},t){return"EDIT"===t.type?Object.assign(Object.assign({},e),{id:t.id}):e},open:function(e={filtered:{},unfiltered:{}},t){return t.type.startsWith("VISIBILITY")?t.filtered?Object.assign(Object.assign({},e),{filtered:sl(e.filtered,t)}):Object.assign(Object.assign({},e),{unfiltered:sl(e.unfiltered,t)}):e},selection:function(e=al().nodes.selection,t){const n=e.ids;switch(t.type){case"SELECTION_CLEAR":return Object.assign(Object.assign({},e),{ids:new Set});case"SELECTION_ONLY":return Object.assign(Object.assign({},e),{ids:new Set([t.id])});case"SELECTION_ADD":return 0===t.ids.length?e:(t.ids.forEach((e=>n.add(e))),Object.assign(Object.assign({},e),{ids:new Set(n)}));case"SELECTION_REMOVE":return 0===t.ids.length?e:(t.ids.forEach((e=>n.delete(e))),Object.assign(Object.assign({},e),{ids:new Set(n)}));case"SELECTION_SET":return Object.assign(Object.assign({},e),{ids:t.ids,mostRecent:t.mostRecent,anchor:t.anchor});case"SELECTION_MOST_RECENT":return Object.assign(Object.assign({},e),{mostRecent:t.id});case"SELECTION_ANCHOR":return Object.assign(Object.assign({},e),{anchor:t.id});default:return e}},drag:function(e=al().nodes.drag,t){switch(t.type){case"DND_DRAG_START":return Object.assign(Object.assign({},e),{id:t.id,selectedIds:t.dragIds});case"DND_DRAG_END":return Object.assign(Object.assign({},e),{id:null,destinationParentId:null,destinationIndex:null,selectedIds:[]});case"DND_HOVERING":return t.parentId!==e.destinationParentId||t.index!=e.destinationIndex?Object.assign(Object.assign({},e),{destinationParentId:t.parentId,destinationIndex:t.index}):e;default:return e}}}),dnd:function(e=al().dnd,t){switch(t.type){case"DND_CURSOR":return Object.assign(Object.assign({},e),{cursor:t.cursor});case"DND_DRAG_START":return Object.assign(Object.assign({},e),{dragId:t.id,dragIds:t.dragIds});case"DND_DRAG_END":return al().dnd;case"DND_HOVERING":return Object.assign(Object.assign({},e),{parentId:t.parentId,index:t.index});default:return e}}}),Th=al();function Dh({treeProps:e,imperativeHandle:t,children:n}){const r=h.useRef(null),o=h.useRef(null),i=h.useRef(bh(xh,al(e))),s=Ze.useSyncExternalStore(i.current.subscribe,i.current.getState,(()=>Th)),c=h.useMemo((()=>new ph(i.current,e,r,o)),[]),l=h.useRef(0);return h.useMemo((()=>{l.current+=1,c.update(e)}),[...Object.values(e),s.nodes.open]),h.useImperativeHandle(t,(()=>c)),h.useEffect((()=>{c.props.selection?c.select(c.props.selection,{focus:!1}):c.deselectAll()}),[c.props.selection]),h.useEffect((()=>{c.props.searchTerm||i.current.dispatch(il.clear(!0))}),[c.props.searchTerm]),a.jsx(bc.Provider,{value:c,children:a.jsx(Dc.Provider,{value:l.current,children:a.jsx(xc.Provider,{value:s.nodes,children:a.jsx(Tc.Provider,{value:s.dnd,children:a.jsx(md,Object.assign({backend:Xf,options:{rootElement:c.props.dndRootElement||void 0}},e.dndManager&&{manager:e.dndManager},{children:n}))})})})})}function Sh(e){return function(){const e=wc(),[,t]=pf((()=>({accept:"NODE",canDrop:(t,n)=>!!n.isOver({shallow:!0})&&e.canDrop(),hover:(t,n)=>{if(!n.isOver({shallow:!0}))return;const r=n.getClientOffset();if(!e.listEl.current||!r)return;const{cursor:o,drop:i}=nh({element:e.listEl.current,offset:r,indent:e.indent,node:null,prevNode:e.visibleNodes[e.visibleNodes.length-1],nextNode:null});i&&e.dispatch(pl.hovering(i.parentId,i.index)),n.canDrop()?o&&e.showCursor(o):e.hideCursor()}})),[e]);t(e.listEl)}(),e.children}function Oh(){const e=wc().props.renderContainer||sh;return a.jsx(a.Fragment,{children:a.jsx(e,{})})}function Rh(){const e=wc(),{offset:t,mouse:n,item:r,isDragging:o}=(i=e=>({offset:e.getSourceClientOffset(),mouse:e.getClientOffset(),item:e.getItem(),isDragging:e.isDragging()}),s=zd().getMonitor(),c=gf(Qd(s,i),2),l=c[0],u=c[1],h.useEffect((function(){return s.subscribeToOffsetChange(u)})),h.useEffect((function(){return s.subscribeToStateChange(u)})),l);var i,s,c,l,u;const d=e.props.renderDragPreview||yl;return a.jsx(d,{offset:t,mouse:n,id:(null==r?void 0:r.id)||null,dragIds:(null==r?void 0:r.dragIds)||[],isDragging:o})}class jh{constructor(e){this.root=function(e){const t=new Nh({id:"ROOT"},null);return t.children=e.map((e=>Ch(e,t))),t}(e)}get data(){var e,t;return null!==(t=null===(e=this.root.children)||void 0===e?void 0:e.map((e=>e.data)))&&void 0!==t?t:[]}create(e){const t=e.parentId?this.find(e.parentId):this.root;if(!t)return null;t.addChild(e.data,e.index)}move(e){const t=this.find(e.id),n=e.parentId?this.find(e.parentId):this.root;t&&n&&(n.addChild(t.data,e.index),t.drop())}update(e){const t=this.find(e.id);t&&t.update(e.changes)}drop(e){const t=this.find(e.id);t&&t.drop()}find(e,t=this.root){if(!t)return null;if(t.id===e)return t;if(t.children){for(let n of t.children){const t=this.find(e,n);if(t)return t}return null}return null}}function Ch(e,t){const n=new Nh(e,t);return e.children&&(n.children=e.children.map((e=>Ch(e,n)))),n}class Nh{constructor(e,t){this.data=e,this.parent=t,this.id=e.id}hasParent(){return!!this.parent}get childIndex(){return this.hasParent()?this.parent.children.indexOf(this):-1}addChild(e,t){var n,r;const o=Ch(e,this);this.children=null!==(n=this.children)&&void 0!==n?n:[],this.children.splice(t,0,o),this.data.children=null!==(r=this.data.children)&&void 0!==r?r:[],this.data.children.splice(t,0,e)}removeChild(e){var t,n;null===(t=this.children)||void 0===t||t.splice(e,1),null===(n=this.data.children)||void 0===n||n.splice(e,1)}update(e){if(this.hasParent()){const t=this.childIndex;this.parent.addChild(Object.assign(Object.assign({},this.data),e),t),this.drop()}}drop(){this.hasParent()&&this.parent.removeChild(this.childIndex)}}let kh=0;function Ph(e){if(e.initialData&&e.data)throw new Error("React Arborist Tree => Provide either a data or initialData prop, but not both.");if(e.initialData&&(e.onCreate||e.onDelete||e.onMove||e.onRename))throw new Error("React Arborist Tree => You passed the initialData prop along with a data handler.\nUse the data prop if you want to provide your own handlers.");if(e.initialData){const[t,n]=function(e){const[t,n]=h.useState(e),r=h.useMemo((()=>new jh(t)),[t]);return[t,{onMove:e=>{for(const t of e.dragIds)r.move({id:t,parentId:e.parentId,index:e.index});n(r.data)},onRename:({name:e,id:t})=>{r.update({id:t,changes:{name:e}}),n(r.data)},onCreate:({parentId:e,index:t,type:o})=>{const i={id:"simple-tree-id-"+kh++,name:""};return"internal"===o&&(i.children=[]),r.create({parentId:e,index:t,data:i}),n(r.data),i},onDelete:e=>{e.ids.forEach((e=>r.drop({id:e}))),n(r.data)}}]}(e.initialData);return Object.assign(Object.assign(Object.assign({},e),n),{data:t})}return e}const Eh=h.forwardRef((function(e,t){const n=Ph(e);return a.jsxs(Dh,{treeProps:n,imperativeHandle:t,children:[a.jsx(Sh,{children:a.jsx(Oh,{})}),a.jsx(Rh,{})]})}));export{mt as B,gc as C,ct as G,Za as K,Na as L,yc as M,dt as P,vt as S,Eh as T,Ga as a,vc as b,$a as c,Ka as d,Oa as e,_a as f,za as g,Ia as h,jt as i,Ca as j,Xa as k,Wa as l,Ja as m,Ya as n,nc as o,Ot as p,Rt as q,st as r,pc as s,jh as t,wt as u};
